Package: RoBSA
Type: Package
Title: Robust Bayesian Survival Analysis
Version: 1.0.4
Maintainer: František Bartoš <f.bartos96@gmail.com>
Authors@R: c( 
    person("František", "Bartoš", role = c("aut", "cre"),
            email = "f.bartos96@gmail.com", comment = c(ORCID = "0000-0002-0018-5573")),
    person("Julia M.", "Haaf", role = "ths",
            comment = c(ORCID = "0000-0001-5122-706X")),
    person("Matthew", "Denwood", role="cph", 
           comment="Original copyright holder of some modified code where indicated."),
    person("Martyn", "Plummer", role="cph", 
           comment="Original copyright holder of some modified code where indicated.")
    )
Description: A framework for estimating ensembles of parametric survival models
    with different parametric families. The RoBSA framework uses Bayesian 
    model-averaging to combine the competing parametric survival models into 
    a model ensemble, weights the posterior parameter distributions based on 
    posterior model probabilities and uses Bayes factors to test for the 
    presence or absence of the individual predictors or preference for a 
    parametric family (Bartoš, Aust & Haaf, 2022, <doi:10.1186/s12874-022-01676-9>).
    The user can define a wide range of informative priors for all parameters 
    of interest. The package provides convenient functions for summary, visualizations, 
    fit diagnostics, and prior distribution calibration.
URL: https://fbartos.github.io/RoBSA/
BugReports: https://github.com/FBartos/RoBSA/issues
License: GPL-3
Encoding: UTF-8
RoxygenNote: 7.2.3
SystemRequirements: JAGS >= 4.3.1 (https://mcmc-jags.sourceforge.io/)
Depends: R (>= 4.0.0)
Imports: BayesTools (>= 0.3.0), survival, rjags, runjags, scales, coda,
        stats, graphics, rlang, Rdpack
Suggests: parallel, ggplot2, flexsurv, testthat, vdiffr, knitr,
        rmarkdown, covr
RdMacros: Rdpack
NeedsCompilation: yes
Packaged: 2026-05-07 09:46:23 UTC; fbart
Author: František Bartoš [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0018-5573>),
  Julia M. Haaf [ths] (ORCID: <https://orcid.org/0000-0001-5122-706X>),
  Matthew Denwood [cph] (Original copyright holder of some modified code
    where indicated.),
  Martyn Plummer [cph] (Original copyright holder of some modified code
    where indicated.)
Repository: CRAN
Date/Publication: 2026-05-07 12:40:41 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-07 16:14:08 UTC; unix
Archs: RoBSA.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(plot,RoBSA)
S3method(predict,RoBSA)
S3method(print,RoBSA)
S3method(print,summary.RoBSA)
S3method(summary,RoBSA)
S3method(update,RoBSA)
export(RoBSA)
export(RoBSA.get_option)
export(RoBSA.options)
export(calibrate_meta_analytic)
export(calibrate_quartiles)
export(check_RoBSA)
export(check_setup)
export(contr.independent)
export(contr.meandif)
export(contr.orthonormal)
export(diagnostics)
export(diagnostics_autocorrelation)
export(diagnostics_density)
export(diagnostics_trace)
export(exp_aft_density)
export(exp_aft_hazard)
export(exp_aft_log_density)
export(exp_aft_log_hazard)
export(exp_aft_log_survival)
export(exp_aft_mean)
export(exp_aft_p)
export(exp_aft_q)
export(exp_aft_r)
export(exp_aft_sd)
export(exp_aft_survival)
export(extract_flexsurv)
export(gamma_aft_density)
export(gamma_aft_hazard)
export(gamma_aft_log_density)
export(gamma_aft_log_hazard)
export(gamma_aft_log_survival)
export(gamma_aft_mean)
export(gamma_aft_p)
export(gamma_aft_q)
export(gamma_aft_r)
export(gamma_aft_sd)
export(gamma_aft_survival)
export(get_default_prior_aux)
export(get_default_prior_beta_alt)
export(get_default_prior_beta_null)
export(get_default_prior_factor_null)
export(get_default_prior_intercept)
export(is.RoBSA)
export(llogis_aft_density)
export(llogis_aft_hazard)
export(llogis_aft_log_density)
export(llogis_aft_log_hazard)
export(llogis_aft_log_survival)
export(llogis_aft_mean)
export(llogis_aft_p)
export(llogis_aft_q)
export(llogis_aft_r)
export(llogis_aft_sd)
export(llogis_aft_survival)
export(lnorm_aft_density)
export(lnorm_aft_hazard)
export(lnorm_aft_log_density)
export(lnorm_aft_log_hazard)
export(lnorm_aft_log_survival)
export(lnorm_aft_mean)
export(lnorm_aft_p)
export(lnorm_aft_q)
export(lnorm_aft_r)
export(lnorm_aft_sd)
export(lnorm_aft_survival)
export(plot_density)
export(plot_hazard)
export(plot_models)
export(plot_prediction)
export(plot_survival)
export(prior)
export(prior_factor)
export(prior_informed)
export(prior_informed_medicine_names)
export(prior_none)
export(set_autofit_control)
export(set_convergence_checks)
export(weibull_aft_density)
export(weibull_aft_hazard)
export(weibull_aft_log_density)
export(weibull_aft_log_hazard)
export(weibull_aft_log_survival)
export(weibull_aft_mean)
export(weibull_aft_p)
export(weibull_aft_q)
export(weibull_aft_r)
export(weibull_aft_sd)
export(weibull_aft_survival)
importFrom(Rdpack,reprompt)
importFrom(rlang,.data)
importFrom(survival,Surv)

README
================

<!-- README.md is generated from README.Rmd. Please edit that file -->
<!-- badges: start -->

[![R-CRAN-check](https://github.com/FBartos/RoBSA/workflows/R-CMD-check/badge.svg)](https://github.com/FBartos/RoBSA/actions)
[![R-tests](https://github.com/FBartos/RoBSA/workflows/R-CMD-tests/badge.svg)](https://github.com/FBartos/RoBSA/actions)
[![Codecov test
coverage](https://codecov.io/gh/FBartos/RoBSA/branch/master/graph/badge.svg)](https://app.codecov.io/gh/FBartos/RoBSA?branch=master)
[![CRAN
status](https://www.r-pkg.org/badges/version/RoBSA)](https://CRAN.R-project.org/package=RoBSA)
<!-- badges: end -->

# Robust Bayesian Survival Analysis (RoBSA)

This package estimates an ensemble of parametric survival models with
different parametric families and uses Bayesian model averaging to
combine them. The RoBSA ensemble uses Bayes factors to test for the
presence or absence of effects of the individual predictors and
evaluates the support for each parametric family. The resulting
model-averaged parameter estimates are based on posterior model
probabilities. The user can define a wide range of prior distributions
for the effect size, intercepts, and auxiliary parameters. The package
provides convenient functions for summary, visualizations, and fit
diagnostics.

See Bartoš et al. (2022) (<https://doi.org/10.1186/s12874-022-01676-9>)
introducing the methodology.

## Installation

The package requires [JAGS 4.3.1](https://mcmc-jags.sourceforge.io/) to
be installed. The development version of the package can be installed
from GitHub:

``` r
devtools::install_github("FBartos/RoBSA")
```

## Example

To illustrate the package’s functionality, we use the `veteran` data set
from the `survival` package containing 137 survival times of patients
from a randomized trial of two treatment regimens for lung cancer. We
start by loading the package and data set.

``` r
library("RoBSA")
#> Loading required namespace: runjags
data(cancer, package = "survival")
head(veteran)
#>   trt celltype time status karno diagtime age prior
#> 1   1 squamous   72      1    60        7  69     0
#> 2   1 squamous  411      1    70        5  64    10
#> 3   1 squamous  228      1    60        3  38     0
#> 4   1 squamous  126      1    60        9  63    10
#> 5   1 squamous  118      1    70       11  65    10
#> 6   1 squamous   10      1    20        5  49     0
```

Before we fit the RoBSA ensemble with five accelerated failure times
parametric families (exponential, Weibull, log-normal, log-logistic, and
gamma), we must specify prior distributions for the intercepts and
auxiliary parameters (governing the scales and shapes) of the competing
parametric families. Here, we assume that we would expect the median
survival type in the standard treatment group to be 5 years with an
interquartile range of 10 years. We set the standard deviation of the
prior distributions to 0.5, which provides us with a satisfactory degree
of uncertainty about the parameter values. We obtain the corresponding
prior distributions by using the `calibrate_quartiles()` function.
Subsequently, we print the list containing priors for the intercepts and
auxiliary parameters.

``` r
priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
priors
#> $intercept
#> $intercept$`exp-aft`
#> Normal(1.98, 0.5)
#> $intercept$`weibull-aft`
#> Normal(2.06, 0.5)
#> $intercept$`lnorm-aft`
#> Normal(1.61, 0.5)
#> $intercept$`llogis-aft`
#> Normal(1.61, 0.5)
#> $intercept$`gamma-aft`
#> Normal(2.46, 0.5)
#> 
#> $aux
#> $aux$`weibull-aft`
#> Lognormal(-0.36, 0.57)
#> $aux$`lnorm-aft`
#> Lognormal(0.2, 0.37)
#> $aux$`llogis-aft`
#> Lognormal(0.15, 0.39)
#> $aux$`gamma-aft`
#> Lognormal(-0.53, 0.63)
```

We create a new data.frame object containing a fit ready data set. We
(1) transform the survival times to years, (2) code the treatment
variable as a factor and set `"standard"` as the default level (to test
for effect of the `"new"` treatment effect; so the intercept of the
model corresponds to the standard treatment and the treatment estimate
to the improvement in the test treatment), and (3) scale the Karnofsky
performance score (karno) to range from 0-1 (i.e., the coefficient
estimate corresponds to the biggest possible difference).

``` r
df <- data.frame(
  time         = veteran$time / 12,
  status       = veteran$status,
  treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
  karno_scaled = veteran$karno / 100 
)
head(df)
#>         time status treatment karno_scaled
#> 1  6.0000000      1  standard          0.6
#> 2 34.2500000      1  standard          0.7
#> 3 19.0000000      1  standard          0.6
#> 4 10.5000000      1  standard          0.6
#> 5  9.8333333      1  standard          0.7
#> 6  0.8333333      1  standard          0.2
```

### Hypothesis Testing

We proceed by specifying a RoBSA model intended to test an informed
hypothesis of the presence of the treatment effect centered around the
log(AF) of 0.3 with the sd of 0.1,5 via the informed prior distribution
$\text{Normal}_+(0.30, 0.15)$ on the treatment effect. To set a prior
distribution on factor, we use the `prior_factor()` function. This
allows us to specify the type of contrast we want to use. Here, we use
`"treatment"` contrast to estimate differences from the default level
(we could also use `orthonormal` to test for a difference of the
individual levels from the grand mean). Furthermore, we adjust for the
Karnofsky performance score by setting a wider centered standard normal
prior distribution $\text{Normal}(0, 1)$. To test only for the presence
of the treatment effect and include the covariate in all models, we set
`test_predictors = "treatment"`. Finally, we pass the appropriate prior
distributions to the `priors`, `prior_intercept`, `prior_aux` arguments,
set `seed = 1` for reproducibility, and use `parallel = TRUE` for
speeding up the computation.

``` r
fit.test <- RoBSA(
    Surv(time, status) ~ treatment + karno_scaled,
    data   = df,
    priors = list(
        treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
                                    truncation = list(0, Inf), contrast = "treatment"),
        karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
    ),
    test_predictors = "treatment",
    prior_intercept = priors[["intercept"]],
    prior_aux       = priors[["aux"]],
    parallel = TRUE, seed = 1
) 
```

The `summary()` functions provides the main summary of the fitted model.

``` r
summary(fit.test)
#> Call:
#> RoBSA(formula = Surv(time, status) ~ treatment + karno_scaled, 
#>     data = df, priors = list(treatment = prior_factor("normal", 
#>         parameters = list(mean = 0.3, sd = 0.15), truncation = list(0, 
#>             Inf), contrast = "treatment"), karno_scaled = prior("normal", 
#>         parameters = list(mean = 0, sd = 1))), test_predictors = "treatment", 
#>     prior_intercept = priors[["intercept"]], prior_aux = priors[["aux"]], 
#>     parallel = TRUE, seed = 1)
#> 
#> Robust Bayesian survival analysis
#> Distributions summary:
#>             Models Prior prob. Post. prob. Inclusion BF
#> exp-aft       2/10       0.200       0.321        1.894
#> weibull-aft   2/10       0.200       0.026        0.105
#> lnorm-aft     2/10       0.200       0.149        0.699
#> llogis-aft    2/10       0.200       0.499        3.991
#> gamma-aft     2/10       0.200       0.005        0.019
#> 
#> Components summary:
#>           Models Prior prob. Post. prob. Inclusion BF
#> treatment   5/10       0.500       0.129        0.148
#> 
#> Model-averaged estimates:
#>                 Mean Median 0.025 0.975
#> treatment[new] 0.019  0.000 0.000 0.224
#> karno_scaled   2.499  2.508 1.647 3.336
```

In the first table, we see that most posterior model probability is
retained by the log-logistic (0.499), exponential (0.321), and
log-normal (0.149) family, with the inclusion Bayes factors quantifying
the change from prior to posterior model probabilities.

The second table then summarizes information about hypothesis tests of
the model components. Here, in the RoBSA ensemble intended for testing,
we are interested in the inclusion Bayes factor for the treatment
effect. We find Bayes factor 0.148, indicating that there is moderate
evidence in favor of no treatment effect (1/0.148 = 6.76) in comparison
to our informed hypothesis of a positive treatment effect.

### Parameter Estimation

Now we try to estimate the model-averaged estimate of the difference
between the two treatments. We change the prior distribution from the
informed positive treatment effect to a neutral standard normal prior
distribution allowing for both positive and negative treatment effect.
We further set `test_predictors` to `""` in order to omit models
assuming zero treatment effect.

``` r
fit.est <- RoBSA(
    Surv(time, status) ~ treatment + karno_scaled,
    data   = df,
    priors = list(
        treatment    = prior_factor("normal", parameters = list(mean = 0, sd = 1),
                                    contrast = "treatment"),
        karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
    ),
    test_predictors = "",
    prior_intercept = priors[["intercept"]],
    prior_aux       = priors[["aux"]],
    parallel = TRUE, seed = 1
) 
```

``` r
summary(fit.est)
#> Call:
#> RoBSA(formula = Surv(time, status) ~ treatment + karno_scaled, 
#>     data = df, priors = list(treatment = prior_factor("normal", 
#>         parameters = list(mean = 0, sd = 1), contrast = "treatment"), 
#>         karno_scaled = prior("normal", parameters = list(mean = 0, 
#>             sd = 1))), test_predictors = "", prior_intercept = priors[["intercept"]], 
#>     prior_aux = priors[["aux"]], parallel = TRUE, seed = 1)
#> 
#> Robust Bayesian survival analysis
#> Distributions summary:
#>             Models Prior prob. Post. prob. Inclusion BF
#> exp-aft        1/5       0.200       0.254        1.360
#> weibull-aft    1/5       0.200       0.021        0.086
#> lnorm-aft      1/5       0.200       0.206        1.036
#> llogis-aft     1/5       0.200       0.515        4.252
#> gamma-aft      1/5       0.200       0.004        0.016
#> 
#> Model-averaged estimates:
#>                  Mean Median  0.025 0.975
#> treatment[new] -0.179 -0.176 -0.561 0.193
#> karno_scaled    2.539  2.544  1.715 3.336
```

We again use the `summary()` function to obtain information about the
fitted model. We find out that the experimental treatment led to notably
shorter survival times with the mean model-averaged log(AF) = -0.19, 95%
CI\[-0.56, -0.19\]. Furthermore, we observe an enormous effect of the
scaled Karnofsky performance score, showing that moving from 0 to 1
increases the survival times with the mean model-averaged log(AF) =
2.54, 95% CI \[1.72, 3.34\].

### MCMC Diagnostics

To assess the convergence of the MCMC model, we can use the
`diagnostics = TRUE` argument in the `summary()` function. The resulting
table shows the maximum MCMC error, min effective sample size and
maximum R-hat for each of the models.

``` r
summary(fit.est, type = "diagnostics")
#> Call:
#> RoBSA(formula = Surv(time, status) ~ treatment + karno_scaled, 
#>     data = df, priors = list(treatment = prior_factor("normal", 
#>         parameters = list(mean = 0, sd = 1), contrast = "treatment"), 
#>         karno_scaled = prior("normal", parameters = list(mean = 0, 
#>             sd = 1))), test_predictors = "", prior_intercept = priors[["intercept"]], 
#>     prior_aux = priors[["aux"]], parallel = TRUE, seed = 1)
#> 
#> Robust Bayesian survival analysis
#> Diagnostics overview:
#>  Model Distribution          Prior treatment         Prior karno_scaled
#>      1      exp-aft treatment contrast: Normal(0, 1)       Normal(0, 1)
#>      2  weibull-aft treatment contrast: Normal(0, 1)       Normal(0, 1)
#>      3    lnorm-aft treatment contrast: Normal(0, 1)       Normal(0, 1)
#>      4   llogis-aft treatment contrast: Normal(0, 1)       Normal(0, 1)
#>      5    gamma-aft treatment contrast: Normal(0, 1)       Normal(0, 1)
#>  max[error(MCMC)] max[error(MCMC)/SD] min(ESS) max(R-hat)
#>           0.01473               0.037      712      1.002
#>           0.01706               0.040      625      1.002
#>           0.01486               0.037      745      1.001
#>           0.01486               0.038      694      1.003
#>           0.01500               0.041      589      1.008
```

We find that while the R-hat and max MCMC error is satisfactory for all
models from the RoBSA estimation ensemble, we might wish for a larger
ESS. To achieve that, we would simply increase the number of sampling
MCMC iterations by using the `iter` argument in the `RoBSA()` function.
Furthermore detailed diagnostics such as trace plot, autocorrelation
plots, and density plot for each parameter/model are provided via the
`diagnostics_trace()`, `diagnostics_autocorrelation()`, and
`diagnostics_density()` functions.

### Visualizing Predicted Survival

We can also visualize the model ensemble predictions for survival or
hazard. We visualize the model-averaged survival for each treatment
group via the `plot_survival()` and using the `predictor = "treatment"`
argument. By default, the remaining predictors are set to their
mean/default level, but predictions for different values can be obtained
by specifying the `covariates_data` argument.

``` r
plot_survival(fit.est, predictor = "treatment")
```

<img src="man/figures/README-predicted_surival-1.png" width="80%" style="display: block; margin: auto;" />

In alignment with the previous summary, we see that the model-averaged
survival for the new treatment group is below the estimated treatment of
the standard treatment group. Similarly, we could also visualize the
model-averaged hazard `plot_hazard()` or obtain numerical estimates for
the survival, hazard, density, mean survival, and standard deviation of
the survival at different levels of detail via the `predict()` function.

### References

<div id="refs" class="references csl-bib-body hanging-indent"
line-spacing="2">

<div id="ref-bartos2021informed" class="csl-entry">

Bartoš, F., Aust, F., & Haaf, J. M. (2022). Informed Bayesian survival
analysis. *BMC Medical Research Methodology*.
<https://doi.org/10.1186/s12874-022-01676-9>

</div>

</div>

### functions for creating model objects
.check_and_list_priors <- function(priors, distributions, data, test_predictors,
                                   default_prior_beta_null, default_prior_beta_alt,
                                   default_prior_factor_null, default_prior_factor_alt,
                                   default_prior_intercept, default_prior_aux){

  predictors      <- attr(data[["predictors"]],"terms")
  predictors_type <- attr(data[["predictors"]],"terms_type")

  # check the input
  if(!BayesTools::is.prior.simple(default_prior_beta_null) || BayesTools::is.prior.factor(default_prior_beta_null))
    stop("The default prior for predictors (null) is not a valid prior distribution.", call. = FALSE)
  if(!BayesTools::is.prior.simple(default_prior_beta_alt) || BayesTools::is.prior.factor(default_prior_beta_alt))
    stop("The default prior for predictors (alt) is not a valid prior distribution.", call. = FALSE)
  if(!BayesTools::is.prior.factor(default_prior_factor_null) & !BayesTools::is.prior.point(default_prior_factor_null))
    stop("The default prior for factors (null) is not a valid prior distribution.", call. = FALSE)
  if(!BayesTools::is.prior.factor(default_prior_factor_alt) & !BayesTools::is.prior.point(default_prior_factor_alt))
    stop("The default prior for factors (alt) is not a valid prior distribution.", call. = FALSE)
  if(!all(sapply(default_prior_intercept[distributions], function(p) BayesTools::is.prior.simple(p) && !BayesTools::is.prior.factor(p))))
    stop("The default prior for intercepts are not a valid prior distribution.", call. = FALSE)
  if(!all(sapply(default_prior_aux[distributions[.has_aux(distributions)]], function(p) BayesTools::is.prior.simple(p) && !BayesTools::is.prior.factor(p))))
    stop("The default prior for auxilary parameters are not a valid prior distribution.", call. = FALSE)

  # check for reserved words
  if(any(names(priors) %in% .reserved_words()))
    stop(paste0("The following prior names are internally reserved keywords and cannot be used: ",
                paste0(" '", names(priors)[names(priors) %in% .reserved_words()], "' ", collapse = ", ")), call. = FALSE)

  # completely the prior distribution specification
  if(is.null(priors) && (!is.null(test_predictors) && length(test_predictors) == 1 && isFALSE(test_predictors))){
    # default estimation if no priors and test_predictors is false
    test_predictors <- NULL

  }else if(is.null(priors) && is.null(test_predictors)){
    # complete default - tests all predictors with default priors
    test_predictors <- predictors

  }else if(!is.null(priors)){
    # find whether user specified some parameter priors, if not - tests all predictors with default priors
    predictors_prior_info <- unlist(sapply(predictors, function(p){
      if(is.null(priors[[p]])){
        return("no-priors-are-specified")
      }else if(BayesTools::is.prior(priors[[p]])){
        return("one-prior-is-specified")
      }else if(length(priors[[p]]) == 2 && all(names(priors[[p]]) %in% c("null", "alt"))){
        if(all(sapply(priors[[p]], BayesTools::is.prior))){
          return(p)
        }else{
          stop(paste0("The prior distribution for '",p,"' is specified incorrectly."))
        }
      }else{
        stop(paste0("The prior distribution for '",p,"' is specified incorrectly."))
      }
    }))
    if(isFALSE(test_predictors)){
      test_predictors <- NULL
    }else if(is.null(test_predictors)){
      test_predictors <- predictors[!predictors %in% c("one-prior-is-specified", "no-priors-are-specified")]
    }
  }

  # some additional checks for the remaining specifiable parameters
  if(!is.null(priors)){
    if(!is.null(priors[["aux"]])){
      sapply(distributions, function(d){
        if(!is.null(priors[["aux"]][[d]])){
          if(!BayesTools::is.prior(priors[["aux"]][[d]])){
            stop(paste0("The prior distribution for the auxilary parameter of '",d,"' distribution is specified incorrectly."))
          }
        }
      })
    }
    if(!is.null(priors[["intercept"]])){
      sapply(distributions, function(d){
        if(!is.null(priors[["intercept"]][[d]])){
          if(!BayesTools::is.prior(priors[["intercept"]][[d]])){
            stop(paste0("The prior distribution for the intercept parameter of '",d,"' distribution is specified incorrectly."))
          }
        }
      })
    }
  }


  if(is.null(priors)){

    priors <- list()

    to_test <- predictors[predictors %in% test_predictors]
    no_test <- predictors[!predictors %in% test_predictors]

    for(i in seq_along(to_test)){
      priors[[to_test[i]]] <- list(
        null = if(predictors_type[to_test[i]] == "factor") default_prior_factor_null else default_prior_beta_null,
        alt  = if(predictors_type[to_test[i]] == "factor") default_prior_factor_alt  else default_prior_beta_alt
      )
    }
    for(i in seq_along(no_test)){
      priors[[no_test[i]]] <- list(
        alt  = if(predictors_type[no_test[i]] == "factor") default_prior_factor_alt else default_prior_beta_alt
      )
    }

    priors[["intercept"]] <- default_prior_intercept[distributions]
    priors[["aux"]]       <- default_prior_aux[distributions]

  }else{

    if(any(!names(priors) %in% c(predictors, "intercept"  ,"aux")))
      stop(paste0("The following priors do not corresponds to any predictor or additional parameter: '", paste(names(priors)[!names(priors) %in% c(predictors, "intercept"  ,"aux")], collapse = "', '", sep = ""), "'"))


    to_test <- predictors[predictors %in% test_predictors]
    no_test <- predictors[!predictors %in% test_predictors]

    for(i in seq_along(to_test)){
      if(is.null(priors[[to_test[i]]])){
        priors[[to_test[i]]] <- list(
          null = if(predictors_type[to_test[i]] == "factor") default_prior_factor_null else default_prior_beta_null,
          alt  = if(predictors_type[to_test[i]] == "factor") default_prior_factor_alt  else default_prior_beta_alt
        )
      }else if(BayesTools::is.prior(priors[[to_test[i]]])){
        priors[[to_test[i]]] <- list(
          null  = if(predictors_type[to_test[i]] == "factor") default_prior_factor_null else default_prior_beta_null,
          alt   = priors[[to_test[i]]]
        )
        # fix possible missalignment of default and custom prior distribution
        if(BayesTools::is.prior.treatment(priors[[to_test[i]]][["alt"]]) && !BayesTools::is.prior.treatment(priors[[to_test[i]]][["null"]])){
          priors[[to_test[i]]][["null"]] <- .fix_default_prior_factor_null(priors[[to_test[i]]][["null"]], contrast = "treatment")
        }else if(BayesTools::is.prior.orthonormal(priors[[to_test[i]]][["alt"]]) && !BayesTools::is.prior.orthonormal(priors[[to_test[i]]][["null"]])){
          priors[[to_test[i]]][["null"]] <- .fix_default_prior_factor_null(priors[[to_test[i]]][["null"]], contrast = "orthonormal")
        }else if(BayesTools::is.prior.meandif(priors[[to_test[i]]][["alt"]]) && !BayesTools::is.prior.meandif(priors[[to_test[i]]][["null"]])){
          priors[[to_test[i]]][["null"]] <- .fix_default_prior_factor_null(priors[[to_test[i]]][["null"]], contrast = "meandif")
        }
      }else if(length(priors[[to_test[i]]]) == 2 && all(names(priors[[to_test[i]]]) %in% c("null", "alt"))){
        priors[[to_test[i]]] <- list(
          null  = priors[[to_test[i]]][["null"]],
          alt   = priors[[to_test[i]]][["alt"]]
        )
        if((BayesTools::is.prior.treatment(priors[[to_test[i]]][["alt"]])   && !BayesTools::is.prior.treatment(priors[[to_test[i]]][["null"]]))   |
           (BayesTools::is.prior.orthonormal(priors[[to_test[i]]][["alt"]]) && !BayesTools::is.prior.orthonormal(priors[[to_test[i]]][["null"]])) |
           (BayesTools::is.prior.meandif(priors[[to_test[i]]][["alt"]])     && !BayesTools::is.prior.meandif(priors[[to_test[i]]][["null"]])))
          stop(paste0("The null and alternative prior distribution for '", to_test[i], "' factor must have matching contrasts."))
      }else{
        stop(paste0("The predictor '", to_test[i], "' is supposed to be used for testing and the prior distributions are not specified properly."))
      }
    }

    for(i in seq_along(no_test)){
      if(is.null(priors[[no_test[i]]])){
        priors[[no_test[i]]] <- list(
          alt  = if(predictors_type[no_test[i]] == "factor") default_prior_factor_alt  else default_prior_beta_alt
        )
      }else{
        if(BayesTools::is.prior(priors[[no_test[i]]])){
          priors[[no_test[i]]] <- list(
            alt  = priors[[no_test[i]]]
          )
        }else{
          # should be stopped before
          stop(paste0("The predictor '", no_test[i], "' is supposed to be used for testing and the prior distributions are not specified properly"))
        }
      }
    }
    if(is.null(priors[["intercept"]])){
      priors[["intercept"]] <- default_prior_intercept[distributions]
    }else{
      for(d in distributions){
        if(is.null(priors[["intercept"]][[d]])){
          priors[["intercept"]][[d]] <- default_prior_intercept[d]
        }
      }
    }
    if(is.null(priors[["aux"]])){
      priors[["aux"]] <- default_prior_aux[distributions]
    }else{
      for(d in distributions){
        if(is.null(priors[["aux"]][[d]])){
          priors[["aux"]][[d]] <- default_prior_aux[d]
        }
      }
    }
  }

  priors$terms       <- priors[predictors]
  priors[predictors] <- NULL

  attr(priors, "distributions")  <- distributions
  attr(priors, "terms")          <- predictors
  attr(priors, "terms_test")     <- if(length(test_predictors) == 1 && test_predictors == "") NULL else test_predictors

  return(priors)
}
.prepare_models <- function(priors, distributions, distributions_weights){

  BayesTools::check_char(distributions, "distributions", allow_values = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"), check_length = FALSE)
  BayesTools::check_real(distributions_weights, "distributions_weights", lower = 0, check_length = length(distributions))

  ### create grid of the models
  grid <- list(
    distribution = distributions
  )

  no_test <- attr(priors, "terms")[!attr(priors, "terms") %in% attr(priors, "terms_test")]
  to_test <- attr(priors, "terms")[ attr(priors, "terms") %in% attr(priors, "terms_test")]

  for(i in seq_along(no_test)){
    grid[[no_test[i]]] <- "alt"
  }
  for(i in seq_along(to_test)){
    grid[[to_test[i]]] <- c("null", "alt")
  }

  grid       <- do.call(expand.grid, grid)
  grid$order <- 1:nrow(grid)
  grid       <- merge(grid, data.frame(cbind(distribution = distributions, prior_weights = distributions_weights)), by = "distribution", all.x = TRUE)
  grid       <- grid[order(grid$order),]
  grid$order <- NULL

  if(nrow(grid) > 50)
    warning("More than 50 models are about to be estimated based on the prior specification.", immediate. = TRUE)

  ### create empty models objects for fitting
  models <- lapply(1:nrow(grid), function(i).create_model(grid[i,], priors))

  return(models)
}
.create_model   <- function(grid_row, priors){

  distribution  <- as.character(grid_row[,"distribution"])
  prior_weights <- as.numeric(grid_row[,"prior_weights"])
  terms         <- attr(priors, "terms")

  ### priors
  model_priors <- list()
  model_priors[["intercept"]]  <- priors[["intercept"]][[distribution]]
  if(.has_aux(distribution)){
    model_priors[["aux"]]      <- priors[["aux"]][[distribution]]
  }
  model_priors[["terms"]] <- list()
  terms_test              <- NULL
  for(i in seq_along(terms)){
    model_priors[["terms"]][[terms[i]]]  <- priors[["terms"]][[terms[i]]][[grid_row[,terms[i]]]]
    prior_weights                        <- prior_weights * priors[["terms"]][[terms[i]]][[grid_row[,terms[i]]]]$prior_weights
    if(grid_row[,terms[i]] == "alt"){
      terms_test <- c(terms_test, terms[i])
    }
  }

  model <- list(
    priors       = model_priors,
    distribution = distribution,
    terms        = terms,
    terms_test   = terms_test,
    prior_weights     = prior_weights,
    prior_weights_set = prior_weights
  )
  class(model) <- "RoBSA.model"

  return(model)
}

.prepare_data <- function(formula, data, rescale_data){

  if(!is.language(formula))
    stop("The 'formula' is not specidied as a formula.")
  if(!is.null(data) && !is.data.frame(data))
    stop("'data' must be an object of type data.frame.")
  BayesTools::check_bool(rescale_data, "rescale_data")

  ### obtain the survival object
  lhs <- formula[[2]]
  if(lhs[[1]] != "Surv")
    stop("The left hand side of the formula must be a Surv() object.")

  # get the data matrix
  surv_outcome <- eval(lhs, envir = data)

  if(anyNA(surv_outcome))
    stop("NA values are not allowed in the Surv() object.")

  ### obtain the predictors part
  rhs <- formula[c(1,3)]
  model_frame <- stats::model.frame(rhs, data = data)

  # change characters into factors
  for(i in seq_along(attr(attr(model_frame, "terms"), "dataClasses"))){
    if(attr(attr(model_frame, "terms"), "dataClasses")[[i]] == "character"){
      model_frame[,names(attr(attr(model_frame, "terms"), "dataClasses"))[i]] <-
        as.factor(model_frame[,names(attr(attr(model_frame, "terms"), "dataClasses"))[i]])
      attr(attr(model_frame, "terms"), "dataClasses")[[i]] <- "factor"
    }
  }


  # prepare the data holders
  n_event <- n_cens_r <- n_cens_l <- n_cens_i  <- 0
  t_event <- t_cens_r <- t_cens_l <- t_cens_il <- t_cens_ir <- NULL
  status  <- NULL

  ### sort and order the data
  # always in the following order event, cens_r, cens_l, cens_i
  # (so a single formula with common parameters can be used for model building)
  if(attr(surv_outcome, "type") == "right"){

    n_event  <- sum(surv_outcome[,2] == 1)
    n_cens_r <- sum(surv_outcome[,2] == 0)

    t_event  <- surv_outcome[surv_outcome[,2] == 1,1]
    t_cens_r <- surv_outcome[surv_outcome[,2] == 0,1]

    model_frame <- model_frame[c(
      c(1:nrow(model_frame))[surv_outcome[,2] == 1],
      c(1:nrow(model_frame))[surv_outcome[,2] == 0]
    ),,drop=FALSE]

  }else if(attr(surv_outcome, "type") == "left"){

    n_event  <- sum(surv_outcome[,2] == 1)
    n_cens_l <- sum(surv_outcome[,2] == 0)

    t_event  <- surv_outcome[surv_outcome[,2] == 1,1]
    t_cens_l <- surv_outcome[surv_outcome[,2] == 0,1]

    model_frame <- model_frame[c(
      c(1:nrow(model_frame))[surv_outcome[,2] == 1],
      c(1:nrow(model_frame))[surv_outcome[,2] == 0]
    ),,drop=FALSE]

  }else if(attr(surv_outcome, "type") == "interval"){

    n_event  <- sum(surv_outcome[,3] == 1)
    n_cens_r <- sum(surv_outcome[,3] == 0)
    n_cens_l <- sum(surv_outcome[,3] == 2)
    n_cens_i <- sum(surv_outcome[,3] == 3)

    t_event   <- surv_outcome[surv_outcome[,3] == 1,1]
    t_cens_r  <- surv_outcome[surv_outcome[,3] == 0,1]
    t_cens_l  <- surv_outcome[surv_outcome[,3] == 2,1]
    t_cens_il <- surv_outcome[surv_outcome[,3] == 3,1]
    t_cens_ir <- surv_outcome[surv_outcome[,3] == 3,2]

    model_frame <- model_frame[c(
      c(1:nrow(model_frame))[surv_outcome[,2] == 1],
      c(1:nrow(model_frame))[surv_outcome[,2] == 0],
      c(1:nrow(model_frame))[surv_outcome[,2] == 2],
      c(1:nrow(model_frame))[surv_outcome[,2] == 3]
    ),,drop=FALSE]

  }else{
    stop(paste0("The ", attr(surv_outcome, "type"), " survival type is not implemented."))
  }


  data_survival <- list(
    n_event   = n_event,
    n_cens_r  = n_cens_r,
    n_cens_l  = n_cens_l,
    n_cens_i  = n_cens_i,
    t_event   = t_event,
    t_cens_r  = t_cens_r,
    t_cens_l  = t_cens_l,
    t_cens_il = t_cens_il,
    t_cens_ir = t_cens_ir
  )

  class(data_survival) <- "RoBSA.survdata"
  attr(data_survival, "type") <- c(
    if(n_event > 0)  "event",
    if(n_cens_r > 0) "cens_r",
    if(n_cens_l > 0) "cens_l",
    if(n_cens_i > 0) "cens_i")


  model_frame     <- as.list(model_frame)
  if(length(model_frame) == 0){
    data_predictors <- list()
  }else{
    data_predictors <- model_frame[1:length(model_frame)]
  }
  attr(data_predictors, "variables")  <- attr(attr(model_frame, "terms"), "term.labels")[attr(attr(model_frame, "terms"), "order") == 1]
  attr(data_predictors, "terms")      <- attr(attr(model_frame, "terms"), "term.labels")
  attr(data_predictors, "terms_type") <- attr(attr(model_frame, "terms"), "dataClasses")


  # add additional information about the predictors
  data_predictors_info <- list()
  to_warn              <- NULL
  for(i in seq_along(data_predictors)){
    if(attr(data_predictors, "terms_type")[i] == "numeric"){

      data_predictors_info[[names(data_predictors)[i]]] <- list(
        type = "continuous",
        mean = mean(data_predictors[[names(data_predictors)[i]]]),
        sd   = stats::sd(data_predictors[[names(data_predictors)[i]]])
      )

      if(rescale_data){
        data_predictors[[names(data_predictors)[i]]] <- .pred_scale(data_predictors[[names(data_predictors)[i]]], data_predictors_info[[names(data_predictors)[i]]])
      }else if(RoBSA.get_option("check_scaling") && (abs(mean(data_predictors[[names(data_predictors)[i]]])) > 0.01 | abs(1 - stats::sd(data_predictors[[names(data_predictors)[i]]])) > 0.01)){
        to_warn <- c(to_warn, names(data_predictors)[i])
      }

    }else if(attr(data_predictors, "terms_type")[i] == "factor"){

      data_predictors_info[[names(data_predictors)[i]]] <- list(
        type    = "factor",
        default = levels(data_predictors[[names(data_predictors)[i]]])[1],
        levels  = levels(data_predictors[[names(data_predictors)[i]]])
      )

    }
  }
  attr(data_predictors, "variables_info") <- data_predictors_info

  # create the output object
  output <- list(
    survival   = data_survival,
    predictors = data_predictors
  )

  # throw warnings and errors
  if(length(to_warn) > 0){
    scaling_warning <- paste0("The continuous predictors ", paste0("'", to_warn, "'", collapse = ", "), " are not scaled. Note that extra care need to be taken when specifying prior distributions for unscaled predictors.")
    warning(scaling_warning, immediate. = TRUE, call. = FALSE)
    warning("You can suppress this and following warnings via 'RoBSA.options(check_scaling = FALSE)'. To automatically rescale predictors set 'rescale_data = TRUE'.", immediate. = TRUE, call. = FALSE)
    attr(output, "warnings") <- scaling_warning
  }

  # post processing checks
  if(any(!attr(data_survival, "type") %in% c("event", "cens_r")))
    stop("Only right censored observations are supported.")

  # check for reserved words
  if(any(attr(data_predictors, "terms") %in% .reserved_words()))
    stop(paste0("The following variable names are internally reserved keywords and cannot be used: ",
                paste0(" '", attr(data_predictors, "terms")[attr(data_predictors, "terms") %in% .reserved_words()], "' ", collapse = ", ")))

  return(output)
}


.pred_scale    <- function(x, predictor_info){
  (x - predictor_info[["mean"]]) / predictor_info[["sd"]]
}
.pred_unscale  <- function(x, predictor_info){
  x * predictor_info[["sd"]] + predictor_info[["mean"]]
}

#' @title Check fitted RoBSA object for errors and warnings
#'
#' @description Checks fitted RoBSA object
#' for warnings and errors and prints them to the
#' console.
#'
#' @param fit a fitted RoBSA object.
#'
#'
#' @return \code{check_RoBSA} returns a vector of error and
#' warning messages.
#'
#' @export
check_RoBSA <- function(fit){
  .print_errors_and_warnings(fit, max_print = Inf)
}


.remove_model_posteriors   <- function(object){
  for(i in seq_along(object[["models"]])){
    if(inherits(object$models[[i]][["fit"]], "runjags")){
      object$models[[i]]$fit[["mcmc"]] <- NULL
    }
  }
  return(object)
}
.remove_model_margliks     <- function(object){
  for(i in seq_along(object[["models"]])){
    if(inherits(object$models[[i]][["marglik"]], "bridge")){
      object$models[[i]]$marglik[["q11"]] <- NULL
      object$models[[i]]$marglik[["q12"]] <- NULL
      object$models[[i]]$marglik[["q21"]] <- NULL
      object$models[[i]]$marglik[["q22"]] <- NULL
    }
  }
  return(object)
}
.print_errors_and_warnings <- function(object, max_print = 5){

  errors_and_warnings <- .collect_errors_and_warnings(object, max_print = max_print)

  for(i in seq_along(errors_and_warnings))
    warning(errors_and_warnings[i], immediate. = TRUE, call. = FALSE)

  return(invisible(errors_and_warnings))
}
.shorten_warnings    <- function(warnings, n_warnings = 5){
  if(is.null(warnings)){
    return(NULL)
  }else if(length(warnings) <= n_warnings){
    return(warnings)
  }else{
    return(c(warnings[1:n_warnings], paste0("There were another ", length(warnings) - n_warnings - 1, " warnings. To see all warnings call 'check_RoBSA(fit)'.")))
  }
}
.shorten_errors      <- function(errors, n_errors = 5){
  if(is.null(errors)){
    return(NULL)
  }else if(length(errors) <= n_errors){
    return(errors)
  }else{
    return(c(errors[1:n_errors], paste0("There were another ", length(errors) - n_errors - 1, " errors. To see all errors call 'check_RoBSA(fit)'.")))
  }
}
.convergence_warning <- function(object){
  if(any(!.get_model_convergence(object))){
    return(paste0(sum(!.get_model_convergence(object)), ifelse(sum(!.get_model_convergence(object)) == 1, " model", " models"), " failed to converge."))
  }else{
    return(NULL)
  }
}
.collect_errors_and_warnings <- function(object, max_print = 5){

  short_warnings <- .shorten_warnings(object$add_info[["warnings"]], max_print)
  short_errors   <- .shorten_errors(object$add_info[["errors"]],     max_print)
  conv_warning   <- .convergence_warning(object)

  return(c(short_warnings, short_errors, conv_warning))
}
.get_model_convergence       <- function(object){
  return(sapply(object[["models"]], function(model) if(is.null(model[["converged"]])) FALSE else model[["converged"]]))
}
.get_model_warnings          <- function(object){
  return(unlist(sapply(seq_along(object[["models"]]), function(i){
    if(!is.null(object[["models"]][[i]][["warnings"]])){
      paste0("Model (", i, "): ", object[["models"]][[i]][["warnings"]])
    }
  })))
}
.get_model_errors            <- function(object){
  return(unlist(sapply(seq_along(object[["models"]]), function(i){
    if(!is.null(object[["models"]][[i]][["errors"]])){
      paste0("Model (", i, "): ", object[["models"]][[i]][["errors"]])
    }
  })))
}


.has_aux         <- function(distributions){
  distributions != "exp-aft"
}
.intercept_name  <- function(distribution){
  switch(
    distribution,
    "exp-aft"     = "rate",
    "weibull-aft" = "scale",
    "lnorm-aft"   = "meanlog",
    "llogis-aft"  = "scale",
    "gamma-aft"   = "rate"
  )
}
.aux_name        <- function(distribution){
  switch(
    distribution,
    "exp-aft"     = NULL,
    "weibull-aft" = "shape",
    "lnorm-aft"   = "sdlog",
    "llogis-aft"  = "shape",
    "gamma-aft"   = "shape"
  )
}
.intercept_transformation <- function(distribution){
  switch(
    distribution,
    "exp-aft"     = function(x)exp(-x),
    "weibull-aft" = function(x)exp(x),
    "lnorm-aft"   = function(x)x,
    "llogis-aft"  = function(x)exp(x),
    "gamma-aft"   = function(x)exp(-x)
  )
}
.aux_transformation       <- function(distribution){
  switch(
    distribution,
    "exp-aft"     = NULL,
    "weibull-aft" = function(x)x,
    "lnorm-aft"   = function(x)x,
    "llogis-aft"  = function(x)x,
    "gamma-aft"   = function(x)x
  )
}
.is_zero_spike            <- function(prior){
  if(prior$distribution == "point"){
    if(prior$parameters$location == 0){
      return(TRUE)
    }
  }
  return(FALSE)
}
.is_model_constant        <- function(priors){
  # checks whether there is at least one non-nill prior
  return(all(sapply(priors, function(prior) BayesTools::is.prior.point(prior) | BayesTools::is.prior.none(prior))))
}

.extract_aux_samples <- function(fit){
  return(as.numeric(suppressWarnings(coda::as.mcmc(fit))[,"aux"]))
}

.BayesTools_parameter_name <- function(parameter){
  return(BayesTools::JAGS_parameter_names(parameter, formula_parameter = "mu"))
}
.output_parameter_names    <- function(parameter){
  return(BayesTools::format_parameter_names(parameter, formula_parameters = "mu", formula_prefix = FALSE))
}


.reserved_words <- function() c("aux", "intercept", "terms", "time", "mu")

context("(0) Basic tests for CRAN")
# These are just a very rudimentary tests that don't require time or saved files.
# The full range of tests is run locally.


test_that("Basic functionality works", {

  RoBSA:::.load_RoBSA_module()

  # fit a default model
  df <- data.frame(
    time = c(0.5, 1.0, 0.2, 0.8, 0.1, 0.1, 0.8, 0.4, 0.2, 0.8),
    cens = c(  0,   1,   1,   0,   0,   1,   0,   0,   1,   1),
    x_c  = c(0.2, 0.8, 0.1, 0.3, 0.5, 0.4, 0.2, 0.1, 0.8, 0.9),
    x_d  = c("A", "A", "A", "B", "B", "B", "A", "A", "B", "B")
  )

  fit <- suppressWarnings(RoBSA(Surv(time, cens) ~ x_c + x_d, data = df, chains = 1, burnin = 50, sample = 100, autofit = FALSE, seed = 1))

  expect_equal(TRUE, is.RoBSA(fit))

  expect_equal(
    capture_output_lines(fit, print = TRUE, width = 150),
    c("Call:"                                                                ,
      "RoBSA(formula = Surv(time, cens) ~ x_c + x_d, data = df, chains = 1, ",
      "    sample = 100, burnin = 50, autofit = FALSE, seed = 1)"            ,
      ""                                                                     ,
      "Estimates:"                                                           ,
      "       x_c     x_d[B] "                                               ,
      "-0.1492768 -0.2029159 "                                               ))

  expect_equal(
    capture_output_lines(summary(fit), print = TRUE, width = 150),
    c("Call:"                                                                                                                                                              ,
      "RoBSA(formula = Surv(time, cens) ~ x_c + x_d, data = df, chains = 1, "                                                                                              ,
      "    sample = 100, burnin = 50, autofit = FALSE, seed = 1)"                                                                                                          ,
      ""                                                                                                                                                                   ,
      "Robust Bayesian survival analysis"                                                                                                                                  ,
      "Distributions summary:"                                                                                                                                             ,
      "            Models Prior prob. Post. prob. Inclusion BF"                                                                                                            ,
      "exp-aft       4/20       0.200       0.362        2.272"                                                                                                            ,
      "weibull-aft   4/20       0.200       0.153        0.721"                                                                                                            ,
      "lnorm-aft     4/20       0.200       0.149        0.702"                                                                                                            ,
      "llogis-aft    4/20       0.200       0.142        0.662"                                                                                                            ,
      "gamma-aft     4/20       0.200       0.194        0.961"                                                                                                            ,
      ""                                                                                                                                                                   ,
      "Components summary:"                                                                                                                                                ,
      "    Models Prior prob. Post. prob. Inclusion BF"                                                                                                                    ,
      "x_c  10/20       0.500       0.470        0.888"                                                                                                                    ,
      "x_d  10/20       0.500       0.447        0.807"                                                                                                                    ,
      ""                                                                                                                                                                   ,
      "Model-averaged estimates:"                                                                                                                                          ,
      "         Mean Median  0.025 0.975"                                                                                                                                  ,
      "x_c    -0.149  0.000 -1.596 0.944"                                                                                                                                  ,
      "x_d[B] -0.203  0.000 -1.568 0.692"                                                                                                                                  ,
      "\033[0;31mThe continuous predictors 'x_c' are not scaled. Note that extra care need to be taken when specifying prior distributions for unscaled predictors.\033[0m",
      "\033[0;31mModel (1): ESS 84 is lower than the set target (500).\033[0m"                                                                                             ,
      "\033[0;31mModel (2): ESS 37 is lower than the set target (500).\033[0m"                                                                                             ,
      "\033[0;31mModel (3): ESS 33 is lower than the set target (500).\033[0m"                                                                                             ,
      "\033[0;31mModel (4): ESS 38 is lower than the set target (500).\033[0m"                                                                                             ,
      "\033[0;31mThere were another 15 warnings. To see all warnings call 'check_RoBSA(fit)'.\033[0m"))

})

#' @title Prints a fitted RoBSA object
#'
#' @param x a fitted RoBSA object.
#' @param ... additional arguments.
#'
#'
#' @return \code{print.RoBSA} invisibly returns the print statement.
#'
#' @seealso [RoBSA()]
#' @export
print.RoBSA <- function(x, ...){
  cat("Call:\n")
  print(x$call)
  cat("\nEstimates:\n")
  print(stats::coef(x))
}


#' @title Summarize fitted RoBSA object
#'
#' @description \code{summary.RoBSA} creates a numerical
#' summary of the RoBSA object.
#'
#' @param object a fitted RoBSA object.
#' @param type whether to show the overall RoBSA results (\code{"ensemble"}),
#' an overview of the individual models (\code{"models"}), or detailed summary
#' for the individual models (\code{"individual"}).
#' @param conditional show the conditional estimates (assuming that the
#' alternative is true). Defaults to \code{FALSE}. Only available for
#' \code{type == "conditional"}.
#' @param exp whether exponents of the regression estimates should be also presented
#' @param probs quantiles of the posterior samples to be displayed.
#' Defaults to \code{c(.025, .50, .975)}
#' @param logBF show log of the BFs. Defaults to \code{FALSE}.
#' @param BF01 show BF in support of the null hypotheses. Defaults to
#' \code{FALSE}.
#' @param transform_factors Whether factors with orthonormal prior
#' distributions should be transformed to differences from the grand mean. Defaults
#' to \code{TRUE}.
#' @param ... additional arguments
#' @inheritParams BayesTools::BayesTools_ensemble_tables
#'
#'
#' @return summary of a RoBSA object
#' @examples \dontrun{
#' # (execution of the example takes several minutes)
#' # example from the README (more details and explanation therein)
#' data(cancer, package = "survival")
#' priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
#' df <- data.frame(
#'   time         = veteran$time / 12,
#'   status       = veteran$status,
#'   treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
#'   karno_scaled = veteran$karno / 100
#' )
#' RoBSA.options(check_scaling = FALSE)
#' fit <- RoBSA(
#'   Surv(time, status) ~ treatment + karno_scaled,
#'   data   = df,
#'   priors = list(
#'     treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
#'                                 truncation = list(0, Inf), contrast = "treatment"),
#'     karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
#'   ),
#'   test_predictors = "treatment",
#'   prior_intercept = priors[["intercept"]],
#'   prior_aux       = priors[["aux"]],
#'   parallel = TRUE, seed = 1
#' )
#'
#' # summary can provide many details about the model
#' summary(fit)
#'
#' # note that the summary function contains additional arguments
#' # that allow to obtain a specific output, i.e, the conditional estimates
#' # (assuming that the non-null models are true) can be obtained
#' summary(fit, conditional = TRUE)
#'
#' # overview of the models and their prior and posterior probability, marginal likelihood,
#' # and inclusion Bayes factor:
#' summary(fit, type = "models")
#'
#' # and the model diagnostics overview, containing maximum R-hat and minimum ESS across parameters
#' # but see '?diagnostics' for diagnostics plots for individual model parameters
#' summary(fit, type = "diagnostics")
#'
#' # summary of individual models and their parameters can be further obtained by
#' summary(fit, type = "individual")
#'
#' }
#'
#' @note See [diagnostics()] for visual convergence checks of the individual models.
#'
#'
#' @return \code{summary.RoBSA} returns a list of tables of class 'BayesTools_table'.
#'
#' @seealso [RoBSA()], [diagnostics()], [check_RoBSA()]
#' @export
summary.RoBSA       <- function(object, type = "ensemble", conditional = FALSE,
                                exp = FALSE, parameters = FALSE, probs = c(.025, .975), logBF = FALSE, BF01 = FALSE,
                                transform_factors = TRUE, short_name = FALSE, remove_spike_0 = FALSE, ...){

  BayesTools::check_bool(conditional, "conditional")
  BayesTools::check_char(type, "type")
  BayesTools::check_bool(exp,  "exp")
  BayesTools::check_bool(parameters,  "parameters")
  BayesTools::check_real(probs, "probs", allow_NULL = TRUE, check_length = 0)
  BayesTools::check_bool(BF01,  "BF01")
  BayesTools::check_bool(logBF, "logBF")
  BayesTools::check_bool(transform_factors, "transform_factors")
  BayesTools::check_bool(short_name, "short_name")
  BayesTools::check_bool(remove_spike_0, "remove_spike_0")

  # print diagnostics if all models fail to converge
  if(!any(.get_model_convergence(object))){
    if(substr(type,1,1) != "d")
      warning("All models failed to converge. Model diagnostics were printed instead.")
    type        <- "diagnostics"
  }

  if(substr(type,1,1) == "e"){

    # obtain components overview
    components_distributions <- BayesTools::ensemble_inference_table(
      inference  = object$RoBSA[["inference_distributions"]],
      parameters = names(object$RoBSA[["inference_distributions"]]),
      logBF      = logBF,
      BF01       = BF01,
      title      = "Distributions summary:"
    )

    if(length(object$RoBSA[["inference"]]) > 0){
      for(i in seq_along(object$RoBSA[["inference"]])){
        attr(object$RoBSA[["inference"]][[i]], "parameter_name") <- gsub("(mu) ", "", attr(object$RoBSA[["inference"]][[i]], "parameter_name"), fixed = TRUE)
      }
      components <- BayesTools::ensemble_inference_table(
        inference      = object$RoBSA[["inference"]],
        parameters     = names(object$RoBSA[["inference"]]),
        logBF          = logBF,
        BF01           = BF01,
        title          = "Components summary:",
      )
    }else{
      components <- NULL
    }

    # obtain estimates tables
    if(length(object$RoBSA[["posteriors"]]) > 0){
      estimates <- BayesTools::ensemble_estimates_table(
        samples               = object$RoBSA[["posteriors"]],
        parameters            = names(object$RoBSA[["posteriors"]]),
        probs                 = probs,
        title                 = "Model-averaged estimates:",
        warnings              = .collect_errors_and_warnings(object),
        transform_factors = transform_factors,
        formula_prefix        = FALSE
      )
      if(exp){
        estimates <- .table_add_exp(estimates)
      }
    }else{
      estimates                    <- data.frame(matrix(nrow = 0, ncol = length(probs) + 2))
      colnames(estimates)          <- c("Mean", "Median", probs)
      class(estimates)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(estimates))
      attr(estimates, "type")      <- rep("estimate", ncol(estimates))
      attr(estimates, "rownames")  <- TRUE
      attr(estimates, "title")     <- "Conditional estimates:"
      attr(estimates, "warnings")  <- .collect_errors_and_warnings(object)
    }


    # deal with possibly empty table in case of no alternative models
    if(is.null(object$RoBSA[["posteriors_conditional"]])){
      estimates_conditional                    <- data.frame(matrix(nrow = 0, ncol = length(probs) + 2))
      colnames(estimates_conditional)          <- c("Mean", "Median", probs)
      class(estimates_conditional)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(estimates_conditional))
      attr(estimates_conditional, "type")      <- rep("estimate", ncol(estimates_conditional))
      attr(estimates_conditional, "rownames")  <- TRUE
      attr(estimates_conditional, "title")     <- "Conditional estimates:"
      attr(estimates_conditional, "warnings")  <- .collect_errors_and_warnings(object)
    }else{
      estimates_conditional <- BayesTools::ensemble_estimates_table(
        samples        = object$RoBSA[["posteriors_conditional"]],
        parameters     = names(object$RoBSA[["posteriors_conditional"]]),
        probs          = probs,
        title          = "Conditional estimates:",
        warnings       = .collect_errors_and_warnings(object),
        formula_prefix = FALSE
      )
      if(exp){
        estimates_conditional <- .table_add_exp(estimates_conditional)
      }
    }

    estimates_intercept <- BayesTools::ensemble_estimates_table(
      samples    = object$RoBSA[["posteriors_intercept"]],
      parameters = names(object$RoBSA[["posteriors_intercept"]]),
      probs      = probs,
      title      = "Distribution estimates (intercept):"
    )
    if(!is.null(names(object$RoBSA[["posteriors_aux"]]))){
      estimates_aux <- BayesTools::ensemble_estimates_table(
        samples    = object$RoBSA[["posteriors_aux"]],
        parameters = names(object$RoBSA[["posteriors_aux"]]),
        probs      = probs,
        title      = "Distribution estimates (auxiliary):"
      )
    }else{
      estimates_aux                    <- data.frame(matrix(nrow = 0, ncol = length(probs) + 2))
      colnames(estimates_aux)          <- c("Mean", "Median", probs)
      class(estimates_aux)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(estimates_aux))
      attr(estimates_aux, "type")      <- rep("estimate", ncol(estimates_aux))
      attr(estimates_aux, "rownames")  <- TRUE
      attr(estimates_aux, "title")     <- "Distribution estimates (auxiliary):"
    }


    #
    # if(parameters){
    #   intercept <- cbind(intercept, "Parameter" = sapply(rownames(intercept), .intercept_name), do.call(rbind, lapply(rownames(intercept), function(distribution){
    #     do.call(.intercept_transformation(distribution), list(intercept[distribution,]))
    #   })))
    # }

    ### return results
    output <- list(
      call                     = object[["call"]],
      title                    = "Robust Bayesian survival analysis",
      components_distributions = components_distributions,
      components               = components,
      estimates                = estimates
    )

    if(conditional){
      output$estimates_conditional <- estimates_conditional
    }

    if(parameters){
      output$estimates_intercept <- estimates_intercept
      output$estimates_aux       <- estimates_aux
    }


    class(output) <- "summary.RoBSA"
    attr(output, "type") <- "ensemble"

    return(output)

  }else if(substr(type,1,1) == "m"){


    if(parameters){
      components <- list("Intercept" = "mu_intercept", "Auxiliary" = "aux")
    }else{
      components <- list()
    }

    for(i in seq_along(object$add_info[["predictors"]])){
      components[[object$add_info[["predictors"]][i]]] <- .BayesTools_parameter_name(object$add_info[["predictors"]][i])
    }

    # trick for dealing with empty models
    if(length(components) == 0){
      summary <- BayesTools::ensemble_summary_table(
        models         = object[["models"]],
        parameters     = "mu_intercept",
        title          = "Models overview:",
        footnotes      = NULL,
        warnings       = .collect_errors_and_warnings(object),
        short_name     = short_name,
        remove_spike_0 = remove_spike_0
      )
      summary <- BayesTools::remove_column(summary, 2)
    }else{
      summary <- BayesTools::ensemble_summary_table(
        models         = object[["models"]],
        parameters     = components,
        title          = "Models overview:",
        footnotes      = NULL,
        warnings       = .collect_errors_and_warnings(object),
        short_name     = short_name,
        remove_spike_0 = remove_spike_0
      )
    }

    # add distribution column to the summary
    summary <- BayesTools::add_column(
      summary,
      column_title    = "Distribution",
      column_values   = sapply(object[["models"]], function(m) m[["distribution"]]),
      column_position = 2,
      column_type     = "string")


    output <- list(
      call       = object[["call"]],
      title      = "Robust Bayesian survival analysis",
      summary    = summary
    )

    class(output) <- "summary.RoBSA"
    attr(output, "type") <- "models"

    return(output)

  }else if(substr(type,1,1) == "d"){


    if(parameters){
      components <- list("Intercept" = "mu_intercept", "Auxiliary" = "aux")
    }else{
      components <- list()
    }

    for(i in seq_along(object$add_info[["predictors"]])){
      components[[object$add_info[["predictors"]][i]]] <- .BayesTools_parameter_name(object$add_info[["predictors"]][i])
    }


    # trick for dealing with empty models
    if(length(components) == 0){
      diagnostics <- BayesTools::ensemble_diagnostics_table(
        models         = object[["models"]],
        parameters     = "mu_intercept",
        title          = "Diagnostics overview:",
        footnotes      = NULL,
        warnings       = .collect_errors_and_warnings(object),
        short_name     = short_name,
        remove_spike_0 = remove_spike_0
      )
      diagnostics <- BayesTools::remove_column(diagnostics, 2)
    }else{
      diagnostics <- BayesTools::ensemble_diagnostics_table(
        models         = object[["models"]],
        parameters     = components,
        title          = "Diagnostics overview:",
        footnotes      = NULL,
        warnings       = .collect_errors_and_warnings(object),
        short_name     = short_name,
        remove_spike_0 = remove_spike_0
      )
    }


    # add distribution column to the summary
    diagnostics <- BayesTools::add_column(
      diagnostics,
      column_title    = "Distribution",
      column_values   = sapply(object[["models"]], function(m) m[["distribution"]]),
      column_position = 2,
      column_type     = "string")


    output <- list(
      call        = object[["call"]],
      title       = "Robust Bayesian survival analysis",
      diagnostics = diagnostics
    )

    class(output) <- "summary.RoBSA"
    attr(output, "type") <- "diagnostics"

    return(output)

  }else if(substr(type, 1, 1) == "i"){

    output <- list(
      call       = object[["call"]],
      title      = "Robust Bayesian meta-analysis",
      models     = list()
    )

    for(i in seq_along(object[["models"]])){

      summary  <- BayesTools::model_summary_table(
        model                 = object[["models"]][[i]],
        short_name            = short_name,
        remove_spike_0        = remove_spike_0,
        formula_prefix        = FALSE
      )

      estimates <- object[["models"]][[i]][["fit_summary"]]
      attr(estimates, "warnings")  <- object[["models"]][[i]][["warnings"]]
      attr(estimates, "title")     <- "Parameter estimates:"
      if(exp){
        estimates <- .table_add_exp(estimates)
      }

      output[["models"]][[i]] <- list(
        summary   = summary,
        estimates = estimates
      )
    }

    class(output) <- "summary.RoBSA"
    attr(output, "type") <- "individual"

    return(output)

  }else{
    stop(paste0("Unknown summary type: '", type, "'."))
  }
}


.table_add_exp <- function(table){
  for(i in which(attr(table, "type") == "estimate")){
    table <- BayesTools::add_column(
      table,
      column_title    = paste0("exp(", colnames(table)[i], ")"),
      column_values   = exp(table[,i]),
      column_position = max(which(attr(table, "type") == "estimate")) + 1,
      column_type     = "estimate")
  }
  table[rownames(table) == "aux", grep("exp", colnames(table))] <- NA
  return(table)
}


#' @title Prints summary object for RoBSA method
#'
#' @param x a summary of a RoBSA object
#' @param ... additional arguments
#'
#' @return \code{print.summary.RoBSA} invisibly returns the print statement.
#'
#' @seealso [RoBSA()]
#' @export
print.summary.RoBSA <- function(x, ...){

  cat("Call:\n")
  print(x[["call"]])

  cat("\n")
  cat(x[["title"]])


  if(attr(x, "type") == "ensemble"){

    cat("\n")
    print(x[["components_distributions"]])

    if(!is.null(x[["components"]])){
      cat("\n")
      print(x[["components"]])
    }

    cat("\n")
    print(x[["estimates"]])

    if(!is.null(x[["estimates_conditional"]])){
      cat("\n")
      print(x[["estimates_conditional"]])
    }

    if(!is.null(x[["estimates_intercept"]])){
      cat("\n")
      print(x[["estimates_intercept"]])
    }

    if(!is.null(x[["estimates_aux"]])){
      cat("\n")
      print(x[["estimates_aux"]])
    }

    return(invisible())

  }else if(attr(x, "type") == "models"){

    cat("\n")
    print(x[["summary"]])

    return(invisible())

  }else if(attr(x, "type") == "diagnostics"){

    cat("\n")
    print(x[["diagnostics"]])

    return(invisible())

  }else if(attr(x, "type") == "individual"){

    for(i in seq_along(x[["models"]])){

      if(i > 1){
        cat("\n")
      }
      print(x[["models"]][[i]][["summary"]])

      cat("\n")
      print(x[["models"]][[i]][["estimates"]])
    }

    return(invisible())
  }
}

#' @title Reports whether x is a RoBSA object
#'
#' @param x an object to test
#'
#' @return is.RoBSA returns a boolean.
#' @export is.RoBSA
is.RoBSA            <- function(x){
  inherits(x, "RoBSA")
}

# adapted from the runjags package version 2.2.0
.onLoad <- function(libname, pkgname){

  requireNamespace("runjags")

  RoBSA.private$RoBSA_version <- utils::packageDescription(pkgname, fields = 'Version')

  # Get and save the library location, getting rid of any trailing / caused by r_arch being empty:
  module_location <- gsub('/$','', file.path(libname, pkgname, 'libs', if(.Platform$r_arch!="") .Platform$r_arch else ""))
  if(!file.exists(file.path(module_location, paste('RoBSA', .Platform$dynlib.ext, sep='')))){
    module_location <- NULL
    warning('The RoBSA module could not be loaded.', call. = FALSE)
  }else{
    rjags::load.module("RoBSA", path = module_location, quiet = TRUE)
    if(!"RoBSA" %in% rjags::list.modules()){
      warning('The RoBSA module could not be loaded.', call. = FALSE)
    }
  }

  RoBSA.private$module_location <- module_location
  RoBSA.private$lib_name        <- libname

  setopts <- mget('.RoBSA.options', envir=.GlobalEnv, ifnotfound = list(.RoBSA.options = NULL))[[1]]
  if(!is.null(setopts)){
    if(!is.list(setopts)){
      warning('Ignoring invalid (non-list) specification for .RoBSA.options on loading the RoBSA package', call.=FALSE)
    }else{
      newopts <- do.call('RoBSA.options', args = setopts)
    }
  }

  .check_BayesTools()

}

.onAttach <- function(libname, pkgname){

  # nothing to do

}

.onUnload <- function(libpath){

  # tricking the dyn.library unload
  if(!is.null(RoBSA.private$lib_name)){
    library.dynam("RoBSA", "RoBSA", RoBSA.private$lib_name)
  }

  # Just in case it is not always safe to try and access an element of an env that is in the process of being deleted (when R quits):
  if(!is.null(RoBSA.private$module_location)){
    rjags::unload.module("RoBSA")
  }
}

.load_RoBSA_module <- function(pkgname = "RoBSA"){

  if(is.null(RoBSA.private$module_location) || (!is.null(RoBSA.private$module_location) && RoBSA.private$module_location == "")){
    libnames         <- .libPaths()
    module_locations <- sapply(libnames, function(libname) gsub('/$','', file.path(libname, pkgname, 'libs', if(.Platform$r_arch!="") .Platform$r_arch else "")))
    sapply(module_locations, function(module_location) rjags::load.module("RoBSA", path = module_location))
  }else{
    rjags::load.module("RoBSA", path = RoBSA.private$module_location)
  }

}

# some general functions
.distributions <- c(
  "exp-aft",
  "weibull-aft",
  "lnorm-aft",
  "llogis-aft",
  "gamma-aft"
)

.get_marginal_distribution_function <- function(distribution, type){
  eval(parse(text = paste0(".", gsub("-", "_", distribution))))[[paste0("log_", switch(
    type,
    "event"   = "density",
    "cens_r"  = "survival",
    "cens_l"  = stop("not implemented"),
    "cens_ir" = stop("not implemented"),
    "cens_il" = stop("not implemented")
  ))]]
}
.get_prediction_distribution_function <- function(distribution, type){
  eval(parse(text = paste0(".", gsub("-", "_", distribution))))[[type]]
}


#' @title Exponential AFT parametric family.
#'
#' @description (log) density, hazard, and survival
#' functions for AFT exponential parametric family.
#'
#' @param t vector of survival times
#' @param eta linear predictor
#' @param q vector of quantiles
#' @param p vector of probabilities
#' @param n number of observations
#'
#'
#' @return \code{exp_aft_density}, \code{exp_aft_hazard}, and
#' \code{exp_aft_survival} return the density, hazard, and survival
#' of the specified survival distribution. The \code{exp_aft_log_density},
#' \code{exp_aft_log_hazard}, \code{exp_aft_log_survival} return log of
#' the corresponding qualities. \code{exp_aft_mean} and \code{exp_aft_sd}
#' return the mean and standard deviation of the specified survival distribution.
#' \code{exp_aft_r}, \code{exp_aft_q}, and \code{exp_aft_p} return a random
#' generation, quantiles, and cumulative probabilities of the specified
#' survival distribution.
#'
#'
#' @export exp_aft_log_density
#' @export exp_aft_log_hazard
#' @export exp_aft_log_survival
#' @export exp_aft_density
#' @export exp_aft_hazard
#' @export exp_aft_survival
#' @export exp_aft_mean
#' @export exp_aft_sd
#' @export exp_aft_r
#' @export exp_aft_q
#' @export exp_aft_p
#' @name exp-aft
NULL

# corresponds to stats::dexp(x, rate = 1/exp(eta))
.exp_aft <- list(
  log_density   = function(t, eta) .exp_aft$log_hazard(t, eta) + .exp_aft$log_survival(t, eta),
  log_hazard    = function(t, eta) -eta,
  log_survival  = function(t, eta) -t * exp(-eta),
  density       = function(t, eta) exp(.exp_aft$log_density(t, eta)),
  hazard        = function(t, eta) exp(.exp_aft$log_hazard(t, eta)),
  survival      = function(t, eta) exp(.exp_aft$log_survival(t, eta)),
  mean          = function(eta)    exp(eta),
  sd            = function(eta)    exp(eta),
  r             = function(n, eta) stats::rexp(n, rate = 1/exp(eta)),
  q             = function(p, eta) stats::qexp(p, rate = 1/exp(eta)),
  p             = function(q, eta) stats::pexp(q, rate = 1/exp(eta))
)

#' @rdname exp-aft
exp_aft_log_density  <- function(t, eta){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.exp_aft$log_density(t, eta))
}
#' @rdname exp-aft
exp_aft_log_hazard   <- function(t, eta){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.exp_aft$log_hazard(t, eta))
}
#' @rdname exp-aft
exp_aft_log_survival <- function(t, eta){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.exp_aft$log_survival(t, eta))
}
#' @rdname exp-aft
exp_aft_density      <- function(t, eta){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.exp_aft$density(t, eta))
}
#' @rdname exp-aft
exp_aft_hazard       <- function(t, eta){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.exp_aft$hazard(t, eta))
}
#' @rdname exp-aft
exp_aft_survival     <- function(t, eta){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.exp_aft$survival(t, eta))
}
#' @rdname exp-aft
exp_aft_mean <- function(eta){
  return(.exp_aft$mean(eta))
}
#' @rdname exp-aft
exp_aft_sd   <- function(eta){
  return(.exp_aft$sd(eta))
}
#' @rdname exp-aft
exp_aft_r    <- function(n, eta){
  BayesTools::check_int(n, "n", lower = 0)
  return(.exp_aft$r(n, eta))
}
#' @rdname exp-aft
exp_aft_q    <- function(p, eta){
  BayesTools::check_real(p, "p", lower = 0, upper = 1)
  return(.exp_aft$q(p, eta))
}
#' @rdname exp-aft
exp_aft_p    <- function(q, eta){
  BayesTools::check_real(q, "q", lower = 0)
  return(.exp_aft$p(q, eta))
}


#' @title Weibull AFT parametric family.
#'
#' @description (log) density, hazard, and survival
#' functions for AFT Weibull parametric family.
#'
#' @param t vector of survival times
#' @param eta linear predictor
#' @param shape auxiliary parameter
#' @param q vector of quantiles
#' @param p vector of probabilities
#' @param n number of observations
#'
#'
#' @return \code{weibull_aft_density}, \code{weibull_aft_hazard}, and
#' \code{weibull_aft_survival} return the density, hazard, and survival
#' of the specified survival distribution. The \code{weibull_aft_log_density},
#' \code{weibull_aft_log_hazard}, \code{weibull_aft_log_survival} return log of
#' the corresponding qualities. \code{weibull_aft_mean} and \code{weibull_aft_sd}
#' return the mean and standard deviation of the specified survival distribution.
#' \code{weibull_aft_r}, \code{weibull_aft_q}, and \code{weibull_aft_p} return a random
#' generation, quantiles, and cumulative probabilities of the specified
#' survival distribution.
#'
#'
#' @export weibull_aft_log_density
#' @export weibull_aft_log_hazard
#' @export weibull_aft_log_survival
#' @export weibull_aft_density
#' @export weibull_aft_hazard
#' @export weibull_aft_survival
#' @export weibull_aft_mean
#' @export weibull_aft_sd
#' @export weibull_aft_r
#' @export weibull_aft_q
#' @export weibull_aft_p
#' @name weibull-aft
NULL

# corresponds to stats::dweibull(x, shape = shape, scale = exp(eta))
.weibull_aft <- list(
  log_density   = function(t, eta, shape) .weibull_aft$log_hazard(t, eta, shape) + .weibull_aft$log_survival(t, eta, shape),
  log_hazard    = function(t, eta, shape) log(shape) + (shape - 1) * log(t) - shape * eta,
  log_survival  = function(t, eta, shape) -t^shape * exp(-shape * eta),
  density       = function(t, eta, shape) exp(.weibull_aft$log_density(t, eta, shape)),
  hazard        = function(t, eta, shape) exp(.weibull_aft$log_hazard(t, eta, shape)),
  survival      = function(t, eta, shape) exp(.weibull_aft$log_survival(t, eta, shape)),
  mean          = function(eta, shape)    exp(eta)*gamma(1+1/shape),
  sd            = function(eta, shape)    sqrt(exp(eta)^2 * (gamma(1+2/shape) - gamma(1+1/shape)^2)),
  r             = function(n, eta, shape) stats::rweibull(n, shape = shape, scale = exp(eta)),
  q             = function(p, eta, shape) stats::qweibull(p, shape = shape, scale = exp(eta)),
  p             = function(q, eta, shape) stats::pweibull(q, shape = shape, scale = exp(eta))
)

#' @rdname weibull-aft
weibull_aft_log_density  <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.weibull_aft$log_density(t, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_log_hazard   <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.weibull_aft$log_hazard(t, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_log_survival <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.weibull_aft$log_survival(t, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_density      <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.weibull_aft$density(t, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_hazard       <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.weibull_aft$hazard(t, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_survival     <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.weibull_aft$survival(t, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_mean <- function(eta, shape){
  return(.weibull_aft$mean(eta, shape))
}
#' @rdname weibull-aft
weibull_aft_sd   <- function(eta, shape){
  return(.weibull_aft$sd(eta, shape))
}
#' @rdname weibull-aft
weibull_aft_r    <- function(n, eta, shape){
  BayesTools::check_int(n, "n", lower = 0)
  return(.weibull_aft$r(n, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_q    <- function(p, eta, shape){
  BayesTools::check_real(p, "p", lower = 0, upper = 1)
  return(.weibull_aft$q(p, eta, shape))
}
#' @rdname weibull-aft
weibull_aft_p    <- function(q, eta, shape){
  BayesTools::check_real(q, "q", lower = 0)
  return(.weibull_aft$p(q, eta, shape))
}


#' @title Log-normal AFT parametric family.
#'
#' @description (log) density, hazard, and survival
#' functions for AFT log-normal parametric family.
#'
#' @param t vector of survival times
#' @param eta linear predictor
#' @param sd auxiliary parameter
#' @param q vector of quantiles
#' @param p vector of probabilities
#' @param n number of observations
#'
#'
#' @return \code{lnorm_aft_density}, \code{lnorm_aft_hazard}, and
#' \code{lnorm_aft_survival} return the density, hazard, and survival
#' of the specified survival distribution. The \code{lnorm_aft_log_density},
#' \code{lnorm_aft_log_hazard}, \code{lnorm_aft_log_survival} return log of
#' the corresponding qualities. \code{lnorm_aft_mean} and \code{lnorm_aft_sd}
#' return the mean and standard deviation of the specified survival distribution.
#' \code{lnorm_aft_r}, \code{lnorm_aft_q}, and \code{lnorm_aft_p} return a random
#' generation, quantiles, and cumulative probabilities of the specified
#' survival distribution.
#'
#'
#' @export lnorm_aft_log_density
#' @export lnorm_aft_log_hazard
#' @export lnorm_aft_log_survival
#' @export lnorm_aft_density
#' @export lnorm_aft_hazard
#' @export lnorm_aft_survival
#' @export lnorm_aft_mean
#' @export lnorm_aft_sd
#' @export lnorm_aft_r
#' @export lnorm_aft_q
#' @export lnorm_aft_p
#' @name lnorm-aft
NULL

# corresponds to stats::dlnorm(x, meanlog = eta, sdlog = sd)
.lnorm_aft <- list(
  log_density   = function(t, eta, sd) stats::dlnorm(t, meanlog = eta, sdlog = sd, log = TRUE),
  log_hazard    = function(t, eta, sd) .lnorm_aft$log_density(t, eta, sd) - .lnorm_aft$log_survival(t, eta, sd),
  log_survival  = function(t, eta, sd) stats::plnorm(t, meanlog = eta, sdlog = sd, lower.tail = FALSE, log.p = TRUE),
  density       = function(t, eta, sd) exp(.lnorm_aft$log_density(t, eta, sd)),
  hazard        = function(t, eta, sd) exp(.lnorm_aft$log_hazard(t, eta, sd)),
  survival      = function(t, eta, sd) exp(.lnorm_aft$log_survival(t, eta, sd)),
  mean          = function(eta, sd)    exp(eta + sd^2/2),
  sd            = function(eta, sd)    sqrt( (exp(sd^2)-1) * exp(2*eta+sd^2) ),
  r             = function(n, eta, sd) stats::rlnorm(n, meanlog = eta, sdlog = sd),
  q             = function(p, eta, sd) stats::qlnorm(p, meanlog = eta, sdlog = sd),
  p             = function(q, eta, sd) stats::plnorm(q, meanlog = eta, sdlog = sd)
)

#' @rdname lnorm-aft
lnorm_aft_log_density  <- function(t, eta, sd){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.lnorm_aft$log_density(t, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_log_hazard   <- function(t, eta, sd){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.lnorm_aft$log_hazard(t, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_log_survival <- function(t, eta, sd){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.lnorm_aft$log_survival(t, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_density      <- function(t, eta, sd){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.lnorm_aft$density(t, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_hazard       <- function(t, eta, sd){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.lnorm_aft$hazard(t, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_survival     <- function(t, eta, sd){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.lnorm_aft$survival(t, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_mean <- function(eta, sd){
  return(.lnorm_aft$mean(eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_sd   <- function(eta, sd){
  return(.lnorm_aft$sd(eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_r    <- function(n, eta, sd){
  BayesTools::check_int(n, "n", lower = 0)
  return(.lnorm_aft$r(n, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_q    <- function(p, eta, sd){
  BayesTools::check_real(p, "p", lower = 0, upper = 1)
  return(.lnorm_aft$q(p, eta, sd))
}
#' @rdname lnorm-aft
lnorm_aft_p    <- function(q, eta, sd){
  BayesTools::check_real(q, "q", lower = 0)
  return(.lnorm_aft$p(q, eta, sd))
}

#' @title Log-logistic AFT parametric family.
#'
#' @description (log) density, hazard, and survival
#' functions for AFT log-logistic parametric family.
#'
#' @param t vector of survival times
#' @param eta linear predictor
#' @param shape auxiliary parameter
#' @param q vector of quantiles
#' @param p vector of probabilities
#' @param n number of observations
#'
#'
#' @return \code{llogis_aft_density}, \code{llogis_aft_hazard}, and
#' \code{llogis_aft_survival} return the density, hazard, and survival
#' of the specified survival distribution. The \code{llogis_aft_log_density},
#' \code{llogis_aft_log_hazard}, \code{llogis_aft_log_survival} return log of
#' the corresponding qualities. \code{llogis_aft_mean} and \code{llogis_aft_sd}
#' return the mean and standard deviation of the specified survival distribution.
#' \code{llogis_aft_r}, \code{llogis_aft_q}, and \code{llogis_aft_p} return a random
#' generation, quantiles, and cumulative probabilities of the specified
#' survival distribution.
#'
#'
#' @export llogis_aft_log_density
#' @export llogis_aft_log_hazard
#' @export llogis_aft_log_survival
#' @export llogis_aft_density
#' @export llogis_aft_hazard
#' @export llogis_aft_survival
#' @export llogis_aft_mean
#' @export llogis_aft_sd
#' @export llogis_aft_r
#' @export llogis_aft_q
#' @export llogis_aft_p
#' @name llogis-aft
NULL

# corresponds to flexsurv::dllogis(x, shape = shape, scale = exp(eta))
.llogis_aft <- list(
  log_density   = function(t, eta, shape) log(shape) - eta + (shape - 1) * (log(t) - eta) - 2 * log(1 + (t / exp(eta))^shape),
  log_hazard    = function(t, eta, shape) .llogis_aft$log_density(t, eta, shape) - .llogis_aft$log_survival(t, eta, shape),
  log_survival  = function(t, eta, shape) - log(1 + (t / exp(eta))^shape),
  density       = function(t, eta, shape) exp(.llogis_aft$log_density(t, eta, shape)),
  hazard        = function(t, eta, shape) exp(.llogis_aft$log_hazard(t, eta, shape)),
  survival      = function(t, eta, shape) exp(.llogis_aft$log_survival(t, eta, shape)),
  mean          = function(eta, shape)    ifelse(shape > 1, (exp(eta) * pi / shape) / sin (pi / shape), NA),
  sd            = function(eta, shape)    ifelse(shape > 2, sqrt(exp(eta)^2 * (2*pi/shape / sin(2*pi/shape) - (pi/shape)^2 / sin(pi/shape)^2)), NA),
  r             = function(n, eta, shape) flexsurv::rllogis(n, shape = shape, scale = exp(eta)),
  q             = function(p, eta, shape) flexsurv::qllogis(p, shape = shape, scale = exp(eta)),
  p             = function(q, eta, shape) flexsurv::pllogis(q, shape = shape, scale = exp(eta))
)

#' @rdname llogis-aft
llogis_aft_log_density  <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.llogis_aft$log_density(t, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_log_hazard   <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.llogis_aft$log_hazard(t, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_log_survival <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.llogis_aft$log_survival(t, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_density      <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.llogis_aft$density(t, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_hazard       <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.llogis_aft$hazard(t, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_survival     <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.llogis_aft$survival(t, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_mean <- function(eta, shape){
  if(any(shape <= 1))
    warning("llogis-aft: mean is undefinied for shape <= 1")
  return(.llogis_aft$mean(eta, shape))
}
#' @rdname llogis-aft
llogis_aft_sd   <- function(eta, shape){
  if(any(shape <= 2))
    warning("llogis-aft: sd is undefinied for shape <= 2")
  return(.llogis_aft$sd(eta, shape))
}
#' @rdname llogis-aft
llogis_aft_r    <- function(n, eta, shape){
  BayesTools::check_int(n, "n", lower = 0)
  return(.llogis_aft$r(n, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_q    <- function(p, eta, shape){
  BayesTools::check_real(p, "p", lower = 0, upper = 1)
  return(.llogis_aft$q(p, eta, shape))
}
#' @rdname llogis-aft
llogis_aft_p    <- function(q, eta, shape){
  BayesTools::check_real(q, "q", lower = 0)
  return(.llogis_aft$p(q, eta, shape))
}


#' @title Gamma AFT parametric family.
#'
#' @description (log) density, hazard, and survival
#' functions for AFT gamma parametric family.
#'
#' @param t vector of survival times
#' @param eta linear predictor
#' @param shape auxiliary parameter
#' @param q vector of quantiles
#' @param p vector of probabilities
#' @param n number of observations
#'
#'
#' @return \code{gamma_aft_density}, \code{gamma_aft_hazard}, and
#' \code{gamma_aft_survival} return the density, hazard, and survival
#' of the specified survival distribution. The \code{gamma_aft_log_density},
#' \code{gamma_aft_log_hazard}, \code{gamma_aft_log_survival} return log of
#' the corresponding qualities. \code{gamma_aft_mean} and \code{gamma_aft_sd}
#' return the mean and standard deviation of the specified survival distribution.
#' \code{gamma_aft_r}, \code{gamma_aft_q}, and \code{gamma_aft_p} return a random
#' generation, quantiles, and cumulative probabilities of the specified
#' survival distribution.
#'
#'
#' @export gamma_aft_log_density
#' @export gamma_aft_log_hazard
#' @export gamma_aft_log_survival
#' @export gamma_aft_density
#' @export gamma_aft_hazard
#' @export gamma_aft_survival
#' @export gamma_aft_mean
#' @export gamma_aft_sd
#' @export gamma_aft_r
#' @export gamma_aft_q
#' @export gamma_aft_p
#' @name gamma-aft
NULL

# corresponds to stats::dgamma(x, shape = shape, scale = 1/ exp(-eta))
.gamma_aft <- list(
  log_density   = function(t, eta, shape) stats::dgamma(t, shape = shape, scale = 1/ exp(-eta), log = TRUE),
  log_hazard    = function(t, eta, shape) .gamma_aft$log_density(t, eta, shape) - .gamma_aft$log_survival(t, eta, shape),
  log_survival  = function(t, eta, shape) stats::pgamma(t, shape = shape, scale = 1/ exp(-eta), lower.tail = FALSE, log.p = TRUE),
  density       = function(t, eta, shape) exp(.gamma_aft$log_density(t, eta, shape)),
  hazard        = function(t, eta, shape) exp(.gamma_aft$log_hazard(t, eta, shape)),
  survival      = function(t, eta, shape) exp(.gamma_aft$log_survival(t, eta, shape)),
  mean          = function(eta, shape)    shape / exp(-eta),
  sd            = function(eta, shape)    sqrt(shape / exp(-eta)^2),
  r             = function(n, eta, shape) stats::rgamma(n, shape = shape, scale = 1/exp(-eta)),
  q             = function(p, eta, shape) stats::qgamma(p, shape = shape, scale = 1/exp(-eta)),
  p             = function(q, eta, shape) stats::pgamma(q, shape = shape, scale = 1/exp(-eta))
)

#' @rdname gamma-aft
gamma_aft_log_density  <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.gamma_aft$log_density(t, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_log_hazard   <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.gamma_aft$log_hazard(t, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_log_survival <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.gamma_aft$log_survival(t, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_density      <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.gamma_aft$density(t, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_hazard       <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.gamma_aft$hazard(t, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_survival     <- function(t, eta, shape){
  BayesTools::check_real(t, "t", lower = 0, allow_bound = TRUE, check_length = FALSE)
  return(.gamma_aft$survival(t, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_mean <- function(eta, shape){
  return(.gamma_aft$mean(eta, shape))
}
#' @rdname gamma-aft
gamma_aft_sd   <- function(eta, shape){
  return(.gamma_aft$sd(eta, shape))
}
#' @rdname gamma-aft
gamma_aft_r    <- function(n, eta, shape){
  BayesTools::check_int(n, "n", lower = 0)
  return(.gamma_aft$r(n, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_q    <- function(p, eta, shape){
  BayesTools::check_real(p, "p", lower = 0, upper = 1)
  return(.gamma_aft$q(p, eta, shape))
}
#' @rdname gamma-aft
gamma_aft_p    <- function(q, eta, shape){
  BayesTools::check_real(q, "q", lower = 0)
  return(.gamma_aft$p(q, eta, shape))
}


# some helper functions ----
lnorm_mean     <- function(meanlog, sdlog){
  exp(meanlog + sdlog^2/2)
}
lnorm_sd       <- function(meanlog, sdlog){
  sqrt( (exp(sdlog^2)-1) * exp(2*meanlog+sdlog^2) )
}
lnorm_meanlog  <- function(mean, sd){
  log(mean) - log(-(-mean^2-sd^2)/mean^2)/2
}
lnorm_sdlog    <- function(mean, sd){
  sqrt(log(-(-mean^2-sd^2)/mean^2))
}
dlnorm_mean_sd <- function(x, mean, sd){
  stats::dlnorm(x, lnorm_meanlog(mean, sd), lnorm_sdlog(mean, sd))
}
rlnorm_mean_sd <- function(n, mean, sd){
  stats::rlnorm(n, lnorm_meanlog(mean, sd), lnorm_sdlog(mean, sd))
}

#' @title Predict method for RoBSA objects.
#'
#' @description Predicts survival/hazard/density/mean/sd for a given
#' RoBSA object. Either predicts values for each row of a fully specified
#' \code{new_data} data.frame, or for all levels of a given \code{predictor}
#' at the mean of continuous covariate values and default factor levels or
#' covariate values specified as \code{covariates_data} data.frame.
#'
#' @param object a fitted RoBSA object
#' @param time a vector of time values at which the survival/hazard/density
#' will be predicted (for each passed data point)
#' @param new_data a data.frame containing fully specified predictors for which
#' predictions should be made
#' @param predictor an alternative input to \code{new_data} that automatically
#' generates predictions for each level of the predictor across all either across
#' levels of covariates specified by \code{covariates_data} or at the default values
#' of other predictors
#' @param covariates_data a supplementary input to \code{predictor} that specifies
#' levels of covariates for which predictions should be made
#' @param type what type of prediction should be created
#' @param summarize whether the predictions should be aggregated as mean and sd.
#' Otherwise, prediction for for posterior samples is returned.
#' @param averaged whether predictions should be combined with Bayesian model-averaging
#' or whether predictions for each individual model should be returned.
#' @param conditional whether only models assuming presence of the specified
#' \code{predictor} should be used
#' @param samples number of posterior samples to be evaluated
#' @param ... additional arguments (unused)
#'
#' @return a list with predictions (or a list of lists in case that predictions for each
#' individual model are requested \code{averaged = FALSE})
#' @export
predict.RoBSA <- function(object, time = NULL, new_data = NULL, predictor = NULL, covariates_data = NULL,
                          type = c("survival", "hazard", "density", "mean", "sd"),
                          summarize = TRUE, averaged = TRUE, conditional = FALSE, samples = 10000, ...){

  predictors_all  <- attr(object$data$predictors, "variables")
  predictors_info <- attr(object$data$predictors, "variables_info")
  models          <- object[["models"]]

  BayesTools::check_real(time, "time", allow_NULL = TRUE, lower = 0, check_length = FALSE)
  BayesTools::check_char(type, "type", allow_values = c("survival", "hazard", "density", "mean", "sd"))
  if(type %in% c("survival", "hazard", "density") && is.null(time))
    stop("Time variable 'time' must be provided.")
  BayesTools::check_char(predictor, "predictor", allow_NULL = TRUE, allow_values = predictors_all)
  BayesTools::check_bool(summarize, "summarize")
  BayesTools::check_bool(averaged, "averaged")
  BayesTools::check_bool(conditional, "conditional")
  BayesTools::check_int(samples, "samples", lower = 0)

  # make sure that predictions based on the same model are always the same
  set.seed(object[["add_info"]][["seed"]])

  # check that the new data are correctly specified
  if(!is.null(new_data) && (is.null(predictor) && is.null(covariates_data)) ){

    if(!is.data.frame(new_data))
      stop("'new_data' must be a data.frame")
    if(!all(predictors_all %in% colnames(new_data)))
      stop("All predictors must be provided.")

  }else if(is.null(new_data)){

    if(!is.null(covariates_data) && !is.data.frame(covariates_data))
      stop("'covariates_data' must be a data.frame")

    missing_predictors <- predictors_all[!predictors_all %in% c(colnames(covariates_data), predictor)]

    # add the predictor if specified
    if(!is.null(predictor)){

      predictor_data  <- data.frame(
        if(predictors_info[[predictor]][["type"]] == "factor") predictors_info[[predictor]][["levels"]]
        else if(predictors_info[[predictor]][["type"]] == "continuous") predictors_info[[predictor]][["mean"]]
      )

      if(!is.null(covariates_data)){
        new_data        <- cbind(
          do.call(rbind, lapply(1:nrow(covariates_data), function(i) predictor_data)),
          do.call(rbind, lapply(1:nrow(predictor_data), function(i) covariates_data))
        )
      }else{
        new_data        <- predictor_data
      }

      colnames(new_data)[1] <- predictor

    }

    # construct the missing predictors (i.e., the first level of factors and mean of covariates)
    for(i in seq_along(missing_predictors)){
      new_data <- cbind(
        new_data,
        if(predictors_info[[missing_predictors[i]]][["type"]] == "factor") predictors_info[[missing_predictors[i]]][["default"]]
        else if(predictors_info[[missing_predictors[i]]][["type"]] == "continuous") predictors_info[[missing_predictors[i]]][["mean"]])
      colnames(new_data)[ncol(new_data)] <- missing_predictors[i]
    }

  }else{
    stop("Either no data or only the 'new_data' or 'parameter' (and 'covariates_data') need to be specified.")
  }


  # rescale the new data if the original input was re-scaled
  if(object[["add_info"]][["rescale_data"]]){
    for(i in seq_along(predictors_all)){
      if(predictors_info[[predictors_all[i]]][["type"]] == "continuous"){
        new_data[,predictors_all[i]] <- .pred_scale(new_data[,predictors_all[i]], predictors_info[[predictors_all[i]]])
      }
    }
  }


  # select conditional models (if the predictor is to be tested)
  if(conditional){
    if(!predictor %in% object$add_info[["predictors_test"]])
      stop("Conditional models cannot be selected since the given predictor was not tested (i.e., it was included in all models). ")

    is_null_models <- attr(object$RoBSA$inference_conditional[[.BayesTools_parameter_name(predictor)]], "is_null")
    models         <- models[!is_null_models]
  }


  # pre-specify number of samples for model-averaging
  if(averaged){
    # get re-standardized model weights in case a condition models were selected
    model_weights <- sapply(models, function(model) model[["inference"]][["post_prob"]])
    model_weights <- model_weights / sum(model_weights)

    samples <- round(model_weights * samples)
  }else{
    samples <- rep(samples, length(models))
  }


  # obtain evaluated posterior distributions from each model
  model_parameters <- lapply(seq_along(models), function(i){

    # the samples
    posteriors <- list(
        mu = BayesTools::JAGS_evaluate_formula(
          fit        = models[[i]][["fit"]],
          formula    = object[["formula"]],
          parameter  = "mu",
          data       = new_data,
          prior_list = attr(models[[i]][["fit"]], "prior_list")
    ))
    if(.has_aux(models[[i]][["distribution"]])){
      posteriors$aux = .extract_aux_samples(models[[i]][["fit"]])
    }

    # subset them to have equal amount across models
    indx <- sample(ncol(posteriors[["mu"]]), size = samples[i], replace = TRUE)
    posteriors[["mu"]] <- posteriors[["mu"]][,indx, drop = FALSE]
    if(.has_aux(models[[i]][["distribution"]])){
      posteriors[["aux"]] <- posteriors[["aux"]][indx]
    }

    attr(posteriors, "model")        <- models[[i]][["inference"]][["m_number"]]
    attr(posteriors, "distribution") <- models[[i]][["distribution"]]

    return(posteriors)
  })


  # deal with different types of predictions separately
  if(type %in% c("mean", "sd")){

    model_predictions <- lapply(model_parameters, function(parameters){

      # get the corresponding function
      prediction_function <- .get_prediction_distribution_function(attr(parameters, "distribution"), type)

      # make predictions for each data point and merge into a data.frame
      out <- do.call(rbind, lapply(1:nrow(new_data), function(i){
        args <- list(eta = parameters[["mu"]][i,])
        if(.has_aux(attr(parameters, "distribution"))){
          args <- c(args, list(parameters[["aux"]]))
        }
        return(do.call(prediction_function, args))
      }))

      attr(out, "data")         <- new_data
      attr(out, "model")        <- attr(parameters, "model")
      attr(out, "distribution") <- attr(parameters, "distribution")
      attr(out, "outcome")      <- type
      return(out)
    })


    # return the properly aggregated results
    if(!averaged && !summarize){

      attr(model_predictions, "data")    <- new_data
      attr(model_predictions, "outcome") <- type
      return(model_predictions)

    }else if(!averaged && summarize){

      model_predictions <- lapply(model_predictions, function(predictions){

        if(anyNA(predictions))
          warning("Some of the model returned undefined predictions. The returned summary ommits all NA samples and might be biased.", immediate. = TRUE)

        out <- data.frame(
          mean   = apply(predictions, 1, mean,      na.rm = TRUE),
          sd     = apply(predictions, 1, stats::sd, na.rm = TRUE),
          lCI    = apply(predictions, 1, stats::quantile, probs = 0.025, na.rm = TRUE),
          median = apply(predictions, 1, stats::quantile, probs = 0.500, na.rm = TRUE),
          uCI    = apply(predictions, 1, stats::quantile, probs = 0.975, na.rm = TRUE)
        )

        attr(out, "data")         <- new_data
        attr(out, "model")        <- attr(predictions, "model")
        attr(out, "distribution") <- attr(predictions, "distribution")
        attr(out, "outcome")      <- type
        return(out)
      })

      attr(model_predictions, "data")    <- new_data
      attr(model_predictions, "outcome") <- type
      return(model_predictions)

    }else{

      # average the predictions
      data_predictions <- do.call(cbind, lapply(seq_along(model_predictions), function(i){
        if(samples[i] > 0){
         return(model_predictions[[i]][,1:samples[i]])
        }else{
          return(matrix(nrow = nrow(model_predictions[[i]]), ncol = 0))
        }
      }))
      colnames(data_predictions) <- NULL

      if(summarize){

        if(anyNA(data_predictions))
          warning("Some of the model returned undefined predictions. The returned summary ommits all NA samples and might be biased.", immediate. = TRUE)

        data_predictions <- data.frame(
          mean   = apply(data_predictions, 1, mean,      na.rm = TRUE),
          sd     = apply(data_predictions, 1, stats::sd, na.rm = TRUE),
          lCI    = apply(data_predictions, 1, stats::quantile, probs = 0.025, na.rm = TRUE),
          median = apply(data_predictions, 1, stats::quantile, probs = 0.500, na.rm = TRUE),
          uCI    = apply(data_predictions, 1, stats::quantile, probs = 0.975, na.rm = TRUE)
        )
      }

      attr(data_predictions, "data")    <- new_data
      attr(data_predictions, "outcome") <- type
      return(data_predictions)
    }
  }

  if(type %in% c("survival", "hazard", "density")){

    model_predictions <- lapply(model_parameters, function(parameters){

      # get the corresponding function
      prediction_function <- .get_prediction_distribution_function(attr(parameters, "distribution"), type)

      # at each data point, make predictions for all times for each posterior sample
      out <- lapply(1:nrow(new_data), function(i){
        out <- do.call(rbind, lapply(seq_along(time), function(t){
          args <- list(
            t   = time[t],
            eta = if(is.matrix(parameters[["mu"]])) parameters[["mu"]][i,] else parameters[["mu"]][i]
          )
          if(.has_aux(attr(parameters, "distribution"))){
            args <- c(args, list(parameters[["aux"]]))
          }
          return(do.call(prediction_function, args))
        }))

        attr(out, "time") <- time
        return(out)
      })

      attr(out, "data")         <- new_data
      attr(out, "model")        <- attr(parameters, "model")
      attr(out, "distribution") <- attr(parameters, "distribution")
      attr(out, "outcome")      <- type
      return(out)
    })


    # return the properly aggregated results
    if(!averaged && !summarize){

      attr(model_predictions, "data")    <- new_data
      attr(model_predictions, "time")    <- time
      attr(model_predictions, "outcome") <- type
      return(model_predictions)

    }else if(!averaged && summarize){

      model_predictions <- lapply(model_predictions, function(predictions){

        if(anyNA(predictions))
          warning("Some of the model returned undefined predictions. The returned summary ommits all NA samples and might be biased.", immediate. = TRUE)

        out <- lapply(seq_along(predictions), function(i){
          out <- data.frame(
            time   = time,
            mean   = apply(predictions[[i]], 1, mean,      na.rm = TRUE),
            sd     = apply(predictions[[i]], 1, stats::sd, na.rm = TRUE),
            lCI    = apply(predictions[[i]], 1, stats::quantile, probs = 0.025, na.rm = TRUE),
            median = apply(predictions[[i]], 1, stats::quantile, probs = 0.500, na.rm = TRUE),
            uCI    = apply(predictions[[i]], 1, stats::quantile, probs = 0.975, na.rm = TRUE)
          )
        })

        attr(out, "data")         <- attr(predictions,"data")
        attr(out, "model")        <- attr(predictions,"model")
        attr(out, "distribution") <- attr(predictions,"distribution")
        return(out)
      })

      attr(model_predictions, "data")    <- new_data
      attr(model_predictions, "time")    <- time
      attr(model_predictions, "outcome") <- type
      return(model_predictions)

    }else{

      # average the predictions
      data_predictions <- lapply(1:nrow(new_data), function(j){

        out <- do.call(cbind, lapply(seq_along(model_predictions), function(i){
          if(samples[i] > 0){
            return(model_predictions[[i]][[j]][,1:samples[i]])
          }else{
            return(matrix(nrow = nrow(model_predictions[[i]][[j]]), ncol = 0))
          }
        }))

        attr(out, "time") <- time
        return(out)
      })


      if(summarize){
        data_predictions <- lapply(data_predictions, function(predictions){

          if(anyNA(predictions))
            warning("Some of the model returned undefined predictions. The returned summary ommits all NA samples and might be biased.", immediate. = TRUE)

          return(data.frame(
            time   = time,
            mean   = apply(predictions, 1, mean,      na.rm = TRUE),
            sd     = apply(predictions, 1, stats::sd, na.rm = TRUE),
            lCI    = apply(predictions, 1, stats::quantile, probs = 0.025, na.rm = TRUE),
            median = apply(predictions, 1, stats::quantile, probs = 0.500, na.rm = TRUE),
            uCI    = apply(predictions, 1, stats::quantile, probs = 0.975, na.rm = TRUE)
          ))
        })
      }

      attr(data_predictions, "data")    <- new_data
      attr(data_predictions, "time")    <- time
      attr(data_predictions, "outcome") <- type
      return(data_predictions)
    }
  }
}

#' @name prior
#' @inherit BayesTools::prior
#' @export
prior <- BayesTools::prior

#' @name prior_none
#' @inherit BayesTools::prior_none
#' @export
prior_none <- BayesTools::prior_none

#' @name prior_factor
#' @inherit BayesTools::prior_factor
#' @export
prior_factor <- function(distribution, parameters, truncation = list(lower = -Inf, upper = Inf), prior_weights = 1, contrast = "orthonormal"){
  BayesTools::prior_factor(distribution = distribution, parameters = parameters, truncation = truncation, prior_weights = prior_weights, contrast = contrast)
}

#' @name prior_informed
#' @inherit BayesTools::prior_informed
#' @details Further details can be found in \insertCite{erp2017estimates;textual}{BayesTools},
#' \insertCite{gronau2017bayesian;textual}{BayesTools}, and
#' \insertCite{bartos2021bayesian;textual}{BayesTools}.
#' @references
#' \insertAllCited{}
#' @export
prior_informed <- BayesTools::prior_informed

#' @name prior_informed_medicine_names
#' @title Names of medical subfields from the Cochrane database of systematic reviews
#' @description Contain names identifying the individual subfields
#' from the Cochrane database of systematic reviews. The individual
#' elements correspond to valid name arguments for the [prior_informed()]
#' function.
#' @export
prior_informed_medicine_names <- BayesTools::prior_informed_medicine_names

# needs to be completely re-exported because of some odd roxygen issues
#' @title BayesTools Contrast Matrices
#'
#' @description BayesTools provides several contrast matrix functions for Bayesian factor analysis.
#' These functions create different types of contrast matrices that can be used with factor
#' variables in Bayesian models.
#'
#' @details
#' The package includes the following contrast functions:
#' \describe{
#'   \item{\code{contr.orthonormal}}{Return a matrix of orthonormal contrasts.
#'     Code is based on \code{stanova::contr.bayes} and corresponding to description
#'     by \insertCite{rouder2012default;textual}{BayesTools}. Returns a matrix with n rows and
#'     k columns, with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
#'   \item{\code{contr.meandif}}{Return a matrix of mean difference contrasts.
#'     This is an adjustment to the \code{contr.orthonormal} that ascertains that the prior
#'     distributions on difference between the grand mean and factor level are identical independent
#'     of the number of factor levels (which does not hold for the orthonormal contrast). Furthermore,
#'     the contrast is re-scaled so the specified prior distribution exactly corresponds to the prior
#'     distribution on difference between each factor level and the grand mean -- this is approximately
#'     twice the scale of \code{contr.orthonormal}. Returns a matrix with n rows and k columns,
#'     with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
#'   \item{\code{contr.independent}}{Return a matrix of independent contrasts -- a level for each term.
#'     Returns a matrix with n rows and k columns, with k = n if \code{contrasts = TRUE} and k = n
#'     if \code{contrasts = FALSE}.}
#' }
#'
#' @param n a vector of levels for a factor, or the number of levels
#' @param contrasts logical indicating whether contrasts should be computed
#'
#' @examples
#' # Orthonormal contrasts
#' contr.orthonormal(c(1, 2))
#' contr.orthonormal(c(1, 2, 3))
#'
#' # Mean difference contrasts
#' contr.meandif(c(1, 2))
#' contr.meandif(c(1, 2, 3))
#'
#' # Independent contrasts
#' contr.independent(c(1, 2))
#' contr.independent(c(1, 2, 3))
#'
#' @references
#' \insertAllCited{}
#'
#' @aliases contr.orthonormal contr.meandif contr.independent
#' @name contr.BayesTools
NULL

#' @rdname contr.BayesTools
#' @export
contr.orthonormal <- BayesTools::contr.orthonormal

#' @rdname contr.BayesTools
#' @export
contr.meandif <- BayesTools::contr.meandif

#' @rdname contr.BayesTools
#' @export
contr.independent <- BayesTools::contr.independent

#' @title RoBSA: Robust Bayesian survival analysis
#'
#' @description Bayesian model-averaged parametric survival analysis
#' with ability to specify informed prior distributions and draw inference with
#' inclusion Bayes factors. See \insertCite{bartos2021informed;textual}{RoBSA}
#' for more details about the methodology.
#'
#'
#' @name RoBSA-package
#' @author František Bartoš \email{f.bartos96@@gmail.com}
#' @keywords package
#' @aliases RoBSA-package RoBSA_package RoBSA.package
#' @docType package
#' @section
#' User guide: See \insertCite{bartos2021informed;textual}{RoBSA},
#' for details regarding the RoBSA methodology.
#'
#' @references \insertAllCited{}
#' @importFrom survival Surv
#' @importFrom Rdpack reprompt
#' @importFrom rlang .data
"_PACKAGE"

#' @title Visualizes MCMC diagnostics for a fitted RoBSA object
#'
#' @description \code{diagnostics} creates visual
#' checks of individual models convergence. Numerical
#' overview of individual models can be obtained by
#' \code{summary(object, type = "diagnostics")},
#' or even more detailed information by
#' \code{summary(object, type = "individual")}.
#'
#' @param fit a fitted RoBSA object
#' @param parameter a parameter to be plotted.
#' @param type type of MCMC diagnostic to be plotted.
#' Options are \code{"trace"} for the chains' trace plots,
#' \code{"autocorrelation"} for autocorrelation of the
#' chains, and \code{"densities"} for the overlaying
#' densities of the individual chains. Can be abbreviated to
#' first letters.
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting. Defaults to
#' \code{"base"}.
#' @param show_models MCMC diagnostics of which models should be
#' plotted. Defaults to \code{NULL} which plots MCMC diagnostics
#' for a specified parameter for every model that is part of the
#' ensemble.
#' @param title whether the model number should be displayed in title.
#' Defaults to \code{TRUE} when more than one model is selected.
#' @param lags number of lags to be shown for
#' \code{type = "autocorrelation"}. Defaults to \code{30}.
#' @param ... additional arguments to be passed to the plotting functions.
#'
#'
#' @return \code{diagnostics} returns either \code{NULL} if \code{plot_type = "base"}
#' or an object/list of objects (depending on the number of parameters to be plotted)
#' of class 'ggplot2' if \code{plot_type = "ggplot2"}.
#'
#' @seealso [RoBSA()], [summary.RoBSA()]
#'
#' @name diagnostics
#' @aliases diagnostics_autocorrelation diagnostics_trace diagnostics_density
#' @export diagnostics
#' @export diagnostics_density
#' @export diagnostics_autocorrelation
#' @export diagnostics_trace

#' @rdname diagnostics
diagnostics <- function(fit, parameter = NULL, type, plot_type = "base", show_models = NULL,
                        lags = 30, title = is.null(show_models) | length(show_models) > 1, ...){

  # check settings
  if(!is.RoBSA(fit))
    stop("Diagnostics are available only for RoBSA models.")
  if(fit$add_info[["save"]] == "min")
    stop("Diagnostics cannot be produced because individual model posteriors were not save during the fitting process. Set 'save' parameter to 'all' while fitting the model (see ?RoBSA for more details).")
  BayesTools::check_char(parameter, "parameter", allow_NULL = TRUE)
  BayesTools::check_char(type, "type", allow_values = c("density", "trace", "autocorrelation"))
  BayesTools::check_bool(title, "title", allow_NULL = TRUE)

  # verify whether given parameter exists and set a default parameter for plotting
  if(is.null(parameter)){
    parameter <- fit$add_info[["predictors"]][1]
  }else if(!parameter %in% c("intercept", "aux", fit$add_info[["predictors"]])){
    stop(paste0("The passed parameter does not correspond to any of the specified predictors: ", paste0("'", fit$add_info[["predictors"]], "'", collapse = ", ")))
  }


  # obtain model indexes to plot
  models_ind <- 1:length(fit[["models"]])
  if(!is.null(show_models)){
    models_ind <- models_ind[show_models]
  }

  # a message with info about multiple plots
  if(plot_type == "base" & length(models_ind) > 1)
    message("Multiple plots will be produced. See '?layout' for help with setting multiple plots.")


  dots  <- .set_dots_diagnostics(..., type = type, chains = fit[["fit_control"]][["chains"]])
  plots <- list()

  for(i in models_ind){

    if(!parameter %in% c("intercept", "aux", fit$models[[i]][["terms_test"]])){

      plots[[i]] <- NULL

    }else{

      # get the parameter name
      args                          <- dots
      args$fit                      <- fit$models[[i]][["fit"]]
      args$parameter                <- if(parameter == "aux") "aux" else .BayesTools_parameter_name(parameter)
      args$type                     <- type
      args$plot_type                <- plot_type
      args$lags                     <- lags
      args$transformations          <- NULL
      args$transform_factors    <- TRUE
      args$short_name               <- FALSE
      args$parameter_names          <- FALSE
      args$formula_prefix           <- FALSE

      if(!is.null(title) && title){
        args$main <- paste0("Model ", i)
      }

      plots[[i]] <- do.call(BayesTools::JAGS_diagnostics, args)
    }
  }



  # return the plots
  if(plot_type == "base"){
    return(invisible(plots))
  }else if(plot_type == "ggplot"){
    if(length(plots) == 1){
      plots <- plots[[1]]
    }
    return(plots)
  }
}

#' @rdname diagnostics
diagnostics_autocorrelation <- function(fit, parameter = NULL, plot_type = "base", show_models = NULL,
                                        lags = 30, title = is.null(show_models) | length(show_models) > 1, ...){
  diagnostics(fit = fit, parameter = parameter, type = "autocorrelation", plot_type = plot_type, show_models = show_models, lags = lags, title = title, ...)
}

#' @rdname diagnostics
diagnostics_trace           <- function(fit, parameter = NULL, plot_type = "base", show_models = NULL,
                                        title = is.null(show_models) | length(show_models) > 1, ...){
  diagnostics(fit = fit, parameter = parameter, type = "trace", plot_type = plot_type, show_models = show_models, title = title, ...)
}

#' @rdname diagnostics
diagnostics_density         <- function(fit, parameter = NULL, plot_type = "base", show_models = NULL,
                                        title = is.null(show_models) | length(show_models) > 1, ...){
  diagnostics(fit = fit, parameter = parameter, type = "density", plot_type = plot_type, show_models = show_models, title = title, ...)
}

.set_dots_diagnostics  <- function(..., type, chains){

  dots <- list(...)
  if(is.null(dots[["col"]])){
    dots[["col"]]      <- if(type == "autocorrelation") "black" else rev(scales::viridis_pal()(chains))
  }

  return(dots)
}

.ensemble_inference    <- function(object){

  # use only converged models with prior weights > 0 for inference about parameters
  prior_weights <- sapply(object[["models"]], function(model) model[["prior_weights"]])
  models        <- object[["models"]][.get_model_convergence(object) & prior_weights > 0]

  model_predictors      <- lapply(models, function(model) model[["terms"]])
  model_predictors_test <- lapply(models, function(model) model[["terms_test"]])
  model_distributions   <- sapply(models, function(model) model[["distribution"]])

  distributions   <- object$add_info[["distributions"]]
  predictors      <- object$add_info[["predictors"]]
  predictors_test <- object$add_info[["predictors_test"]]

  # define inference options
  components      <- NULL
  parameters      <- NULL
  components_null <- list()
  parameters_null <- list()

  components_distributions      <- NULL
  components_distributions_null <- list()

  # distributions
  for(i in seq_along(distributions)){
    components_distributions                          <- c(components_distributions, distributions[i])
    components_distributions_null[[distributions[i]]] <- model_distributions != distributions[i]
  }

  # predictors
  for(i in seq_along(predictors_test)){
    components <- c(components, .BayesTools_parameter_name(predictors_test[i]))
    components_null[[.BayesTools_parameter_name(predictors_test[i])]] <-
      sapply(model_predictors_test, function(x) if(length(x) == 0) TRUE else !(predictors_test[i] %in% x))
  }

  for(i in seq_along(predictors)){
    parameters <- c(parameters, .BayesTools_parameter_name(predictors[i]))
    parameters_null[[.BayesTools_parameter_name(predictors[i])]] <-
      sapply(model_predictors_test, function(x) if(length(x) == 0) TRUE else !(predictors[i] %in% x))
  }


  ### get models inference
  if(is.null(components)){
    inference <- NULL
  }else{
    inference <- BayesTools::ensemble_inference(
      model_list   = models,
      parameters   = components,
      is_null_list = components_null,
      conditional  = FALSE
    )
  }
  # deal with the possibility of only null models models
  if(all(sapply(parameters_null, all))){
    inference_conditional <- NULL
  }else{
    inference_conditional <- BayesTools::ensemble_inference(
      model_list   = models,
      parameters   = parameters[!sapply(parameters_null, all)],
      is_null_list = parameters_null[!sapply(parameters_null, all)],
      conditional  = TRUE
    )
  }
  inference_distributions <- BayesTools::ensemble_inference(
    model_list   = models,
    parameters   = components_distributions[!sapply(components_distributions_null, all)],
    is_null_list = components_distributions_null[!sapply(components_distributions_null, all)],
    conditional  = FALSE
  )
  inference_distributions_conditional <- BayesTools::ensemble_inference(
    model_list   = models,
    parameters   = components_distributions[!sapply(components_distributions_null, all)],
    is_null_list = components_distributions_null[!sapply(components_distributions_null, all)],
    conditional  = TRUE
  )


  ### get model-averaged posteriors
  if(is.null(parameters)){
    posteriors <- NULL
  }else{
    posteriors <- BayesTools::mix_posteriors(
      model_list   = models,
      parameters   = parameters,
      is_null_list = parameters_null,
      seed         = object$add_info[["seed"]],
      conditional  = FALSE
    )
  }

  # deal with the possibility of only null models models
  if(all(sapply(parameters_null, all))){
    posteriors_conditional <- NULL
  }else{
    posteriors_conditional <- BayesTools::mix_posteriors(
      model_list   = models,
      parameters   = parameters[!sapply(parameters_null, all)],
      is_null_list = parameters_null[!sapply(parameters_null, all)],
      seed         = object$add_info[["seed"]],
      conditional  = TRUE
    )
  }

  posteriors_intercept <- .mix_posteriors_intercept(
    model_list          = models,
    distributions       = distributions,
    model_distributions = model_distributions,
    seed                = object$add_info[["seed"]]
  )
  posteriors_aux       <- .mix_posteriors_aux(
    model_list          = models,
    distributions       = distributions,
    model_distributions = model_distributions,
    seed                = object$add_info[["seed"]]
  )

  # return the results
  output <- list(
    inference                           = inference,
    inference_conditional               = inference_conditional,
    inference_distributions             = inference_distributions,
    inference_distributions_conditional = inference_distributions_conditional,

    posteriors             = posteriors,
    posteriors_conditional = posteriors_conditional,
    posteriors_intercept   = posteriors_intercept,
    posteriors_aux         = posteriors_aux
  )
  return(output)
}

.mix_posteriors_intercept <- function(model_list, distributions, model_distributions, seed){

  posteriors_intercept <- list()

  for(i in seq_along(distributions)){
    if(all(model_distributions != distributions[i])){
      posteriors_intercept[distributions[i]] <- NULL
    }else{
      posteriors_intercept[distributions[i]] <- BayesTools::mix_posteriors(
        model_list   = model_list,
        parameters   = "mu_intercept",
        is_null_list = list("mu_intercept" = c(model_distributions != distributions[i])),
        seed         = seed,
        conditional  = TRUE
      )
    }
  }

  return(posteriors_intercept)
}
.mix_posteriors_aux       <- function(model_list, distributions, model_distributions, seed){

  posteriors_aux <- list()

  for(i in seq_along(distributions)){
    if(.has_aux(distributions[i]) & !all(model_distributions != distributions[i])){
      posteriors_aux[distributions[i]] <- BayesTools::mix_posteriors(
        model_list   = model_list,
        parameters   = "aux",
        is_null_list = list("aux" = c(model_distributions != distributions[i])),
        seed         = seed,
        conditional  = TRUE
      )
    }else{
      posteriors_aux[distributions[i]] <- NULL
    }
  }

  return(posteriors_aux)
}

.compute_coeficients   <- function(RoBSA){

  return(do.call(c, unname(lapply(RoBSA$posteriors, function(posterior){
    if(inherits(posterior, "mixed_posteriors.factor")){
      out        <- apply(posterior, 2, mean)
      names(out) <- .output_parameter_names(names(out))
    }else{
      out        <- mean(posterior)
      names(out) <- .output_parameter_names(attr(posterior,"parameter"))
    }
    return(out)
  }))))
}

#' @title Extract parameter estimates from \code{flexsurv} object
#'
#' @description \code{extract_flexsurv} extracts estimates from a
#' \code{flexsurv} object in and transform them to match the
#' \code{RoBSA} output.
#'
#' @param fit an object fitted with the \code{flexsurv::flexsurvreg}
#' function
#'
#' @return \code{extract_flexsurv} return list of estimates lists for each
#' parameter.
#'
#' @export
extract_flexsurv <- function(fit){

  if(!inherits(fit, "flexsurvreg"))
    stop("'fit' must be a model fitted with the 'flexsurv::flexsurvreg' function.")

  distribution <- .flexsurv_distribution_name(fit$dlist$name)
  estimates    <- fit$res

  out <- list()


  out[["intercept"]] <- list(
      "mean" = .flexsurv_intercept_transform(distribution)(estimates[.flexsurv_intercept_name(distribution), "est"]),
      "lCI"  = .flexsurv_intercept_transform(distribution)(estimates[.flexsurv_intercept_name(distribution), "est"] -
                                                             1.96 * estimates[.flexsurv_intercept_name(distribution), "se"]),
      "uCI"  = .flexsurv_intercept_transform(distribution)(estimates[.flexsurv_intercept_name(distribution), "est"] +
                                                             1.96 * estimates[.flexsurv_intercept_name(distribution), "se"])
  )

  if(.has_aux(distribution)){
    out[["aux"]] <- list(
      "mean" = .flexsurv_aux_transform(distribution)(estimates[.flexsurv_aux_name(distribution), "est"]),
      "lCI"  = .flexsurv_aux_transform(distribution)(estimates[.flexsurv_aux_name(distribution), "est"] -
                                                       1.96 * estimates[.flexsurv_aux_name(distribution), "se"]),
      "uCI"  = .flexsurv_aux_transform(distribution)(estimates[.flexsurv_aux_name(distribution), "est"] +
                                                       1.96 * estimates[.flexsurv_aux_name(distribution), "se"])
    )
  }

  # omit parameter names
  estimates <- estimates[!rownames(estimates) %in% c(.flexsurv_intercept_name(distribution), .flexsurv_aux_name(distribution)),,drop=FALSE]
  if(nrow(estimates) > 0){
    out[["parameters"]]        <- lapply(rownames(estimates), function(est) list(
      "mean" = .flexsurv_est_transform(distribution)(estimates[est, "est"]),
      "lCI"  = .flexsurv_est_transform(distribution)(estimates[est, "est"] - 1.96 * estimates[est, "se"]),
      "uCI"  = .flexsurv_est_transform(distribution)(estimates[est, "est"] + 1.96 * estimates[est, "se"])
    ))
    names(out[["parameters"]]) <- rownames(estimates)
  }

  return(out)
}

.flexsurv_distribution_name    <- function(distribution){
  switch(
    distribution,
    "weibull.quiet" = "weibull-aft",
    "lnorm"         = "lnorm-aft",
    "llogis"        = "llogis-aft",
    "exp"           = "exp-aft",
    "gamma"         = "gamma-aft"
  )
}
.flexsurv_intercept_name       <- function(distribution){
  switch(
    distribution,
    "weibull-aft"   = "scale",
    "lnorm-aft"     = "meanlog",
    "llogis-aft"    = "scale",
    "exp-aft"       = "rate",
    "gamma-aft"     = "rate"
  )
}
.flexsurv_intercept_transform  <- function(distribution){
  switch(
    distribution,
    "weibull-aft"   = function(x) log(x),
    "lnorm-aft"     = function(x) x,
    "llogis-aft"    = function(x) log(x),
    "exp-aft"       = function(x) log(1/x),
    "gamma-aft"     = function(x) x
  )
}
.flexsurv_aux_name             <- function(distribution){
  switch(
    distribution,
    "weibull-aft"   = "shape",
    "lnorm-aft"     = "sdlog",
    "llogis-aft"    = "shape",
    "gamma-aft"     = "shape"
  )
}
.flexsurv_aux_transform        <- function(distribution){
  switch(
    distribution,
    "weibull-aft"   = function(x) x,
    "lnorm-aft"     = function(x) x,
    "llogis-aft"    = function(x) x,
    "gamma-aft"     = function(x) x
  )
}
.flexsurv_est_transform        <- function(distribution){
  switch(
    distribution,
    "exp-aft"       = function(x) -x,
    "weibull-aft"   = function(x) x,
    "lnorm-aft"     = function(x) x,
    "llogis-aft"    = function(x) x,
    "gamma-aft"     = function(x) -x
  )
}

#' @title Fit Robust Bayesian Survival Analysis
#'
#' @description \code{RoBSA} is used to estimate a robust Bayesian
#' survival analysis. The interface allows a complete customization of
#' the ensemble with different prior distributions for the null and
#' alternative hypothesis of each parameter.
#' (See README for an example.)
#'
#' @param formula formula for the survival model
#' @param data data frame containing the data
#' @param distributions distributions of parametric
#' survival models
#' @param test_predictors vector of predictor names
#' to be tested with Bayesian model-averaged testing.
#' Defaults to \code{NULL}, no parameters are tested.
#' @param priors names list of prior distributions for each
#' predictor. It allows users to specify both the null and alternative
#' hypothesis prior distributions by assigning a named list
#' (with \code{"null"} and \code{"alt"} object) to the predictor
#' @param distributions_weights prior odds for the competing
#' distributions
#' @param prior_beta_null default prior distribution for the
#' null hypotheses of continuous predictors
#' @param prior_beta_alt default prior distribution for the
#' alternative hypotheses of continuous predictors
#' @param prior_factor_null default prior distribution for the
#' null hypotheses of categorical predictors
#' @param prior_factor_alt default prior distribution for the
#' alternative hypotheses of categorical predictors
#' @param prior_intercept named list containing prior
#' distribution for the intercepts (with names corresponding
#' to the distributions)
#' @param prior_aux named list containing prior
#' distribution for the auxiliary parameters (with names corresponding
#' to the distributions)
#' @param chains a number of chains of the MCMC algorithm.
#' @param sample a number of sampling iterations of the MCMC algorithm.
#' Defaults to \code{5000}.
#' @param burnin a number of burnin iterations of the MCMC algorithm.
#' Defaults to \code{2000}.
#' @param adapt a number of adaptation iterations of the MCMC algorithm.
#' Defaults to \code{500}.
#' @param thin a thinning of the chains of the MCMC algorithm. Defaults to
#' \code{1}.
#' @param parallel whether the individual models should be fitted in parallel.
#' Defaults to \code{FALSE}. The implementation is not completely stable
#' and might cause a connection error.
#' @param autofit whether the model should be fitted until the convergence
#' criteria (specified in \code{autofit_control}) are satisfied. Defaults to
#' \code{TRUE}.
#' @param autofit_control allows to pass autofit control settings with the
#' [set_autofit_control()] function. See \code{?set_autofit_control} for
#' options and default settings.
#' @param convergence_checks automatic convergence checks to assess the fitted
#' models, passed with [set_convergence_checks()] function. See
#' \code{?set_convergence_checks} for options and default settings.
#' @param save whether all models posterior distributions should be kept
#' after obtaining a model-averaged result. Defaults to \code{"all"} which
#' does not remove anything. Set to \code{"min"} to significantly reduce
#' the size of final object, however, some model diagnostics and further
#' manipulation with the object will not be possible.
#' @param seed a seed to be set before model fitting, marginal likelihood
#' computation, and posterior mixing for reproducibility of results. Defaults
#' to \code{NULL} - no seed is set.
#' @param silent whether all print messages regarding the fitting process
#' should be suppressed. Defaults to \code{TRUE}. Note that \code{parallel = TRUE}
#' also suppresses all messages.
#' @param rescale_data whether continuous predictors should be rescaled prior to
#' estimating the model. Defaults to \code{FALSE}.
#' @param ... additional arguments.
#'
#' @examples \dontrun{
#' # (execution of the example takes several minutes)
#' # example from the README (more details and explanation therein)
#' data(cancer, package = "survival")
#' priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
#' df <- data.frame(
#'   time         = veteran$time / 12,
#'   status       = veteran$status,
#'   treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
#'   karno_scaled = veteran$karno / 100
#' )
#' RoBSA.options(check_scaling = FALSE)
#' fit <- RoBSA(
#'   Surv(time, status) ~ treatment + karno_scaled,
#'   data   = df,
#'   priors = list(
#'     treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
#'                                 truncation = list(0, Inf), contrast = "treatment"),
#'     karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
#'   ),
#'   test_predictors = "treatment",
#'   prior_intercept = priors[["intercept"]],
#'   prior_aux       = priors[["aux"]],
#'   parallel = TRUE, seed = 1
#' )
#' summary(fit)
#'
#' }
#'
#' @return \code{RoBSA} returns an object of class 'RoBSA'.
#'
#' @rdname RoBSA
#' @aliases RoBSA
#' @export
RoBSA <- function(
  formula, data, priors = NULL, test_predictors = NULL,

  distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
  distributions_weights = rep(1, length(distributions)),

  # default prior distribution
  prior_beta_null   = get_default_prior_beta_null(),
  prior_beta_alt    = get_default_prior_beta_alt(),
  prior_factor_null = get_default_prior_factor_null(),
  prior_factor_alt  = get_default_prior_factor_alt(),
  prior_intercept   = get_default_prior_intercept(),
  prior_aux         = get_default_prior_aux(),

  # MCMC fitting settings
  chains = 3, sample = 5000, burnin = 2000, adapt = 500, thin = 1, parallel = FALSE,
  autofit = TRUE, autofit_control = set_autofit_control(), convergence_checks = set_convergence_checks(),

  # additional settings
  save = "all", seed = NULL, silent = TRUE, rescale_data = FALSE, ...){

  dots         <- .RoBSA_collect_dots(...)
  object       <- NULL
  object$call  <- match.call()


  ### prepare & check the data
  object$data_input <- data
  object$data       <- .prepare_data(formula, data, rescale_data)
  object$formula    <- formula


  ### check MCMC settings
  object$fit_control        <- BayesTools::JAGS_check_and_list_fit_settings(chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin, autofit = autofit, parallel = parallel, cores = chains, silent = silent, seed = seed)
  object$autofit_control    <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_control = autofit_control)
  object$convergence_checks <- .check_and_list_convergence_checks(convergence_checks)


  ### prepare and check the settings
  object$priors  <- .check_and_list_priors(
    priors = priors, distributions = distributions, data = object[["data"]], test_predictors = test_predictors,
    default_prior_beta_null = prior_beta_null, default_prior_beta_alt = prior_beta_alt,
    default_prior_factor_null = prior_factor_null, default_prior_factor_alt = prior_factor_alt,
    default_prior_intercept = prior_intercept, default_prior_aux = prior_aux)
  object$models  <- .prepare_models(object$priors, distributions, distributions_weights)


  ### additional information
  object$add_info <- .check_and_list_add_info(
    distributions    = distributions,
    predictors       = attr(object[["priors"]], "terms"),
    predictors_test  = attr(object[["priors"]], "terms_test"),
    seed             = seed,
    save             = save,
    rescale_data     = rescale_data,
    warnings         = attr(object[["data"]], "warnings"),
    errors           = NULL
  )


  ### fit the models and compute marginal likelihoods
  if(!object$fit_control[["parallel"]]){

    if(dots[["is_JASP"]]){
      .JASP_progress_bar_start(length(object[["models"]]))
    }

    for(i in seq_along(object[["models"]])){
      object$models[[i]] <- .fit_RoBSA_model(object, i)
      if(dots[["is_JASP"]]){
        .JASP_progress_bar_tick()
      }
    }

  }else{

    fitting_order <- .fitting_priority(object[["models"]])

    cl <- parallel::makePSOCKcluster(floor(RoBSA.get_option("max_cores") / object$fit_control[["chains"]]))
    parallel::clusterEvalQ(cl, {library("RoBSA")})
    parallel::clusterExport(cl, "object", envir = environment())
    object$models <- parallel::parLapplyLB(cl, fitting_order, .fit_RoBSA_model, object = object)[order(fitting_order)]
    parallel::stopCluster(cl)

  }


  # create ensemble only if at least one model converged
  if(any(.get_model_convergence(object))){

    # TODO? balance probability of non-converged models?

    ### compute the model-space results
    object$models        <- BayesTools::models_inference(object[["models"]])
    object$RoBSA         <- .ensemble_inference(object)
    object$coefficients  <- .compute_coeficients(object[["RoBSA"]])
  }


  ### collect and print errors and warnings
  object$add_info[["errors"]]   <- c(object$add_info[["errors"]],   .get_model_errors(object))
  object$add_info[["warnings"]] <- c(object$add_info[["warnings"]], .get_model_warnings(object))
  .print_errors_and_warnings(object)


  ### remove model posteriors if asked to
  if(save == "min"){
    object <- .remove_model_posteriors(object)
    object <- .remove_model_margliks(object)
  }


  class(object) <- "RoBSA"
  return(object)
}



#' @title Updates a fitted RoBSA object
#'
#' @description \code{update.RoBSA} can be used to
#' \enumerate{
#'   \item{add an additional model to an existing \code{"RoBSA"} object by
#'    specifying the distribution, and either null or alternative priors
#'    for each parameter and prior weight of the model,}
#'   \item{change the prior weights of fitted models by specifying a vector
#'   \code{prior_weights} of the same length as the fitted models,}
#'   \item{refitting models that failed to converge with updated settings
#'   of control parameters,}
#'   \item{or changing the convergence criteria and recalculating the ensemble
#'   results by specifying new \code{control} argument and setting
#'   \code{refit_failed == FALSE}.}
#' }
#'
#' @param object a fitted RoBSA object
#' @param distribution a distribution of the new model.
#' @param model_weights either a single value specifying prior model weight
#' of a newly specified model using priors argument, or a vector of the
#' same length as already fitted models to update their prior weights.
#' @param refit_failed whether failed models should be refitted. Relevant only
#' if new priors or \code{prior_weights} are not supplied. Defaults to \code{TRUE}.
#' @inheritParams RoBSA
#' @param ... additional arguments.
#'
#' @details See [RoBSA()] for more details.
#'
#' @return \code{update.RoBSA} returns an object of class 'RoBSA'.
#'
#' @seealso [RoBSA()], [summary.RoBSA()], [prior()], [check_setup()]
#' @export
update.RoBSA <- function(
  object, refit_failed = TRUE,

  formula = NULL, priors = NULL, test_predictors = "", distribution = NULL, model_weights = 1,

  # default prior distribution
  prior_beta_null   = get_default_prior_beta_null(),
  prior_beta_alt    = get_default_prior_beta_alt(),
  prior_factor_null = get_default_prior_factor_null(),
  prior_factor_alt  = get_default_prior_factor_alt(),
  prior_intercept   = get_default_prior_intercept(),
  prior_aux         = get_default_prior_aux(),

  chains = NULL, adapt = NULL, burnin = NULL, sample = NULL, thin = NULL, autofit = NULL, parallel = NULL,
  autofit_control = NULL, convergence_checks = NULL,
  save = "all", seed = NULL, silent = TRUE, ...){


  if(object$add_info$save == "min")
    stop("Models cannot be updated because individual model posteriors were not save during the fitting process. Set 'save' parameter to 'all' in while fitting the model (see ?RoBSA for more details).")


  ### choose proper action based on the supplied input
  if(!is.null(distribution)){

    what_to_do <- "fit_new_model"
    BayesTools::check_real(model_weights, "model_weights", lower = 0)

    # use the old formula if new is not supplied (i.e., adding models with different priors / distributions)
    if(!is.null(formula)){
      object$data    <- .prepare_data(formula, object[["data_input"]], object$add_info[["rescale_data"]])
      object$formula <- formula
    }

    # use the old priors if new is not supplied
    if(!is.null(priors)){
      priors  <- .check_and_list_priors(
        priors = priors, distributions = distribution, data = object[["data"]], test_predictors = test_predictors,
        default_prior_beta_null = prior_beta_null, default_prior_beta_alt = prior_beta_alt,
        default_prior_factor_null = prior_factor_null, default_prior_factor_alt = prior_factor_alt,
        default_prior_intercept = prior_intercept, default_prior_aux = prior_aux)
    }else{
      priors  <- .check_and_list_priors(
        priors = object[["priors"]][["terms"]], distributions = distribution, data = object[["data"]], test_predictors = test_predictors,
        default_prior_beta_null = prior_beta_null, default_prior_beta_alt = prior_beta_alt,
        default_prior_factor_null = prior_factor_null, default_prior_factor_alt = prior_factor_alt,
        default_prior_intercept = prior_intercept, default_prior_aux = prior_aux)
    }


    # update add info
    object$add_info <- .update_add_info(
      old_add_info     = object[["add_info"]],
      distribution     = distribution,
      predictors       = attr(priors, "terms"),
      predictors_test  = attr(priors, "terms_test")
    )

    object$models[length(object$models) + 1]  <- list(.prepare_models(priors, distribution, model_weights)[[1]])

    object$models[[length(object$models)]]$model_weights     <- model_weights
    object$models[[length(object$models)]]$model_weights_set <- model_weights


  }else if(!is.null(model_weights)){

    what_to_do <- "update_model_weights"
    BayesTools::check_real(model_weights, "model_weights", check_length = length(object[["models"]]), lower = 0)

    for(i in 1:length(object$models)){
      object$models[[i]]$prior_weights     <- model_weights[i]
      object$models[[i]]$prior_weights_set <- model_weights[i]
    }

  }else if(refit_failed & any(!.get_model_convergence(object))){

    what_to_do <- "refit_failed_models"

  }else{

    what_to_do <- "update_settings"

  }


  ### update control settings if any change is specified
  object[["fit_control"]]        <- .update_fit_control(object[["fit_control"]], chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin, autofit = autofit, parallel = parallel, cores = chains, silent = silent, seed = seed)
  object[["autofit_control"]]    <- .update_autofit_control(object[["autofit_control"]], autofit_control)
  object[["convergence_checks"]] <- .update_convergence_checks(object[["convergence_checks"]], convergence_checks)


  ### do the stuff
  if(what_to_do == "fit_new_model"){

    object[["models"]][[length(object$models)]] <- .fit_RoBSA_model(object, length(object$models))

  }else if(what_to_do == "refit_failed_models"){

    for(i in c(1:length(object$models))[!.get_model_convergence(object)]){
      object[["models"]][[i]] <- .fit_RoBSA_model(object, i)
    }

  }


  # create ensemble only if at least one model converged
  if(any(.get_model_convergence(object))){

    # TODO? balance probability of non-converged models?

    ### compute the model-space results
    object$models        <- BayesTools::models_inference(object[["models"]])
    object$RoBSA         <- .ensemble_inference(object)
    object$coefficients  <- .compute_coeficients(object[["RoBSA"]])
  }


  ### collect and print errors and warnings
  object$add_info[["errors"]]   <- c(object$add_info[["errors"]],   .get_model_errors(object))
  object$add_info[["warnings"]] <- c(object$add_info[["warnings"]], .get_model_warnings(object))
  .print_errors_and_warnings(object)


  ### remove model posteriors if asked to
  if(save == "min"){
    object <- .remove_model_posteriors(object)
    object <- .remove_model_margliks(object)
  }


  class(object) <- "RoBSA"
  return(object)
}

#' @title Options for the RoBSA package
#'
#' @description A placeholder object and functions for the RoBSA package.
#' (adapted from the runjags R package).
#'
#' @param name the name of the option to get the current value of - for a list of
#' available options, see details below.
#' @param ... named option(s) to change - for a list of available options, see
#' details below.
#'
#' @return The current value of all available RoBSA options (after applying any
#' changes specified) is returned invisibly as a named list.
#'
#' @export RoBSA.options
#' @export RoBSA.get_option
#' @name RoBSA_options
#' @aliases RoBSA_options RoBSA.options RoBSA.get_option
NULL


#' @rdname RoBSA_options
RoBSA.options    <- function(...){

	opts <- list(...)

	for(i in seq_along(opts)){

	  if(!names(opts)[i] %in% names(RoBSA.private))
	    stop(paste("Unmatched or ambiguous option '", names(opts)[i], "'", sep=""))

	  if(names(opts)[i] == "distribution")
	    stop("List of default distributions cannot be changed.")

	  assign(names(opts)[i], opts[[i]] , envir = RoBSA.private)
	}

	return(invisible(RoBSA.private$options))
}

#' @rdname RoBSA_options
RoBSA.get_option <- function(name){

	if(length(name)!=1)
	  stop("Only 1 option can be retrieved at a time")

	if(!name %in% names(RoBSA.private))
	  stop(paste("Unmatched or ambiguous option '", name, "'", sep=""))

	# Use eval as some defaults are put in using 'expression' to avoid evaluating at load time:
	return(eval(RoBSA.private[[name]]))
}

# export the function directly to suppress import warnings
.runjags__findjags <- function() runjags::findjags()

# adapted from the runjags package version 2.2.0
RoBSA.private <- new.env()
# Use 'expression' for functions to avoid having to evaluate before the package is fully loaded:
assign("JAGS_path",       expression(.runjags__findjags()),           envir = RoBSA.private)
assign("RoBSA_version",   utils::packageVersion("RoBSA"),             envir = RoBSA.private)
assign("min_jags_major",  4,                                          envir = RoBSA.private)
assign("max_jags_major",  4,                                          envir = RoBSA.private)
assign("max_cores",       parallel::detectCores(logical = TRUE) - 1,  envir = RoBSA.private)
assign("check_scaling",   TRUE,                                       envir = RoBSA.private)
assign("distributions",   .distributions,                             envir = RoBSA.private)

# check proper BayesTools package version
.check_BayesTools <- function(){

  RoBSA.version      <- try(utils::packageVersion("RoBSA"))
  BayesTools.version <- try(utils::packageVersion("BayesTools"))

  if(inherits(RoBSA.version, "try-error") | inherits(BayesTools.version, "try-error")){
    return(invisible(FALSE))
  }

  if(is.null(RoBSA.version) | is.null(BayesTools.version)){
    return(invisible(FALSE))
  }

  BayesTools_required <- switch(
    paste0(RoBSA.version, collapse = "."),
    "1.0.0" = c("0.2.10", "0.2.13"),
    "1.0.1" = c("0.2.10", "0.2.13"),
    "1.0.2" = c("0.2.14", "0.2.23"),
    "1.0.3" = c("0.2.14", "0.2.23"),
    "1.0.4" = c("0.3.0", "999.999.999"),
    stop("New RoBSA version needs to be defined in '.check_BayesTools' function!")
  )

  min_OK <- sum(as.numeric(strsplit(BayesTools_required[1], ".", fixed = TRUE)[[1]]) * c(1e9, 1e6, 1e3)) <=
    sum(unlist(BayesTools.version) * c(1e9, 1e6, 1e3))
  max_OK <- sum(as.numeric(strsplit(BayesTools_required[2], ".", fixed = TRUE)[[1]]) * c(1e9, 1e6, 1e3)) >=
    sum(unlist(BayesTools.version) * c(1e9, 1e6, 1e3))

  if(min_OK && max_OK){
    return(invisible(TRUE))
  }else{
    warning(sprintf(
      "RoBSA version %1$s requires BayesTools version higher or equal %2$s and lower or equal %3$s.",
      paste0(RoBSA.version, collapse = "."),
      BayesTools_required[1],
      BayesTools_required[2]
    ), call.=FALSE)
    return(invisible(FALSE))
  }
}

#' @title Prints summary of \code{"RoBSA"} corresponding to the input
#'
#' @description \code{check_setup} prints summary of \code{"RoBSA"} ensemble
#' corresponding to the specified formula, data, and priors.
#' This function is useful for checking the ensemble configuration prior
#' to fitting all models.
#'
#' @inheritParams RoBSA
#' @param models should the models' details be printed.
#' @param silent do not print the results.
#'
#' @return \code{check_setup} invisibly returns list of summary tables.
#'
#' @seealso [RoBSA()]
#' @export
check_setup <- function(
  formula, data, priors = NULL, test_predictors = NULL,

  distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
  distributions_weights = rep(1, length(distributions)),

  # default prior distribution
  prior_beta_null   = get_default_prior_beta_null(),
  prior_beta_alt    = get_default_prior_beta_alt(),
  prior_factor_null = get_default_prior_factor_null(),
  prior_factor_alt  = get_default_prior_factor_alt(),
  prior_intercept   = get_default_prior_intercept(),
  prior_aux         = get_default_prior_aux(),

  # MCMC fitting settings
  chains = 3, sample = 5000, burnin = 2000, adapt = 500, thin = 1, parallel = FALSE,
  autofit = TRUE, autofit_control = set_autofit_control(), convergence_checks = set_convergence_checks(),

  # additional settings
  save = "all", seed = NULL, silent = FALSE, rescale_data = FALSE, models = FALSE, ...){

  BayesTools::check_bool(models, "models")

  # ----------------------------------------------------------------------------------------- #
  ### run the model generation from RoBSA()
  dots         <- .RoBSA_collect_dots(...)
  object       <- NULL
  object$call  <- match.call()


  ### prepare & check the data
  object$data    <- .prepare_data(formula, data, rescale_data)
  object$formula <- formula


  ### check MCMC settings
  object$fit_control        <- BayesTools::JAGS_check_and_list_fit_settings(chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin, autofit = autofit, parallel = parallel, cores = chains, silent = silent, seed = seed)
  object$autofit_control    <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_control = autofit_control)
  object$convergence_checks <- .check_and_list_convergence_checks(convergence_checks)


  ### prepare and check the settings
  object$priors  <- .check_and_list_priors(priors = priors, distributions = distributions, data = object[["data"]], test_predictors = test_predictors,
                                           default_prior_beta_null = prior_beta_null, default_prior_beta_alt = prior_beta_alt,
                                           default_prior_factor_null = prior_factor_null, default_prior_factor_alt = prior_factor_alt,
                                           default_prior_intercept = prior_intercept, default_prior_aux = prior_aux)
  object$models  <- .prepare_models(object$priors, distributions, distributions_weights)


  ### additional information
  object$add_info <- .check_and_list_add_info(
    distributions    = distributions,
    predictors       = attr(object[["priors"]], "terms"),
    predictors_test  = attr(object[["priors"]], "terms_test"),
    seed             = seed,
    save             = save,
    rescale_data     = rescale_data,
    warnings         = attr(object[["data"]], "warnings"),
    errors           = NULL
  )

  # ----------------------------------------------------------------------------------------- #
  ### run the model specification from .ensemble_inference()
  prior_weights <- sapply(object[["models"]], function(model) model[["prior_weights"]])

  model_predictors      <- lapply(object[["models"]], function(model) model[["terms"]])
  model_predictors_test <- lapply(object[["models"]], function(model) model[["terms_test"]])
  model_distributions   <- sapply(object[["models"]], function(model) model[["distribution"]])

  distributions   <- object$add_info[["distributions"]]
  predictors      <- object$add_info[["predictors"]]
  predictors_test <- object$add_info[["predictors_test"]]

  # define inference options
  components      <- NULL
  parameters      <- NULL
  components_null <- list()
  parameters_null <- list()

  components_distributions      <- NULL
  components_distributions_null <- list()

  # distributions
  for(i in seq_along(distributions)){
    components_distributions                          <- c(components_distributions, distributions[i])
    components_distributions_null[[distributions[i]]] <- model_distributions != distributions[i]
  }

  # predictors
  for(i in seq_along(predictors_test)){
    components <- c(components, .BayesTools_parameter_name(predictors_test[i]))
    components_null[[.BayesTools_parameter_name(predictors_test[i])]] <-
      sapply(model_predictors_test, function(x) if(length(predictors_test) == 0) FALSE else !(predictors_test[i] %in% x))
  }

  for(i in seq_along(predictors)){
    parameters <- c(parameters, .BayesTools_parameter_name(predictors[i]))
    parameters_null[[.BayesTools_parameter_name(predictors[i])]] <-
      sapply(model_predictors_test, function(x) if(length(predictors_test) == 0) FALSE else !(predictors_test[i] %in% x))
  }


  # ----------------------------------------------------------------------------------------- #
  ### create overview tables
  prior_prob <- prior_weights / sum(prior_weights)
  output     <- list()

  if(!models){

    components_distributions <- data.frame(
      "models"     = sapply(components_distributions_null, function(component) sum(!component)),
      "prior_prob" = sapply(components_distributions_null, function(component) sum(prior_prob[!component]))
    )
    rownames(components_distributions) <- distributions

    class(components_distributions)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(components_distributions))
    attr(components_distributions, "type")      <- c("n_models", "prior_prob")
    attr(components_distributions, "rownames")  <- TRUE
    attr(components_distributions, "n_models")  <- length(object$models)
    attr(components_distributions, "title")     <- "Distributions summary:"
    attr(components_distributions, "footnotes") <- NULL
    attr(components_distributions, "warnings")  <- NULL

    output$components_distributions <- components_distributions


    if(!is.null(components)){

      components <- data.frame(
        "models"     = sapply(components_null, function(component) sum(!component)),
        "prior_prob" = sapply(components_null, function(component) sum(prior_prob[!component]))
      )
      rownames(components) <- predictors_test

      class(components)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(components))
      attr(components, "type")      <- c("n_models", "prior_prob")
      attr(components, "rownames")  <- TRUE
      attr(components, "n_models")  <- length(object$models)
      attr(components, "title")     <- "Components summary:"
      attr(components, "footnotes") <- NULL
      attr(components, "warnings")  <- NULL

      output$components <- components

    }


    if(!silent){
      cat("Robust Bayesian survival analysis (set-up)\n")
      print(components_distributions, quote = FALSE, right = TRUE)

      if(!is.null(components)){
        cat("\n")
        print(components, quote = FALSE, right = TRUE)
      }
    }

  }else{

    summary <- cbind.data.frame(
      "Model"             = 1:length(object$models),
      "Distribution"      = sapply(object[["models"]], function(m) m[["distribution"]]),
      "Intercept"         = sapply(object[["models"]], function(m) print(m[["priors"]][["intercept"]], silent = TRUE)),
      "Auxiliary"         = sapply(object[["models"]], function(m) if(!.has_aux(m[["distribution"]])) print(BayesTools::prior_none(), silent = TRUE) else print(m[["priors"]][["aux"]], silent = TRUE))
    )

    for(i in seq_along(predictors)){
      summary <- cbind(summary, "xxx" = sapply(object[["models"]], function(m) print(m[["priors"]][["terms"]][[predictors[i]]], silent = TRUE)))
      colnames(summary)[length(colnames(summary))] <- predictors[i]
    }

    summary <- cbind(summary, "prior_prob" = prior_prob)

    class(summary)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(summary))
    attr(summary, "type")      <- c("integer", "string", rep("prior", 2 + length(predictors)), "prior_prob")
    attr(summary, "rownames")  <- FALSE
    attr(summary, "title")     <- "Models overview:"
    attr(summary, "footnotes") <- NULL
    attr(summary, "warnings")  <- NULL

    object$summary <- summary

    if(!silent){
      cat("Robust Bayesian survival analysis (set-up)\n")
      print(summary, quote = FALSE, right = TRUE)
    }
  }

  return(invisible(object))
}



#' @title Control MCMC fitting process
#'
#' @description Controls settings for the autofit
#' process of the MCMC JAGS sampler (specifies termination
#' criteria), and values for the convergence checks.
#'
#' @param max_Rhat maximum value of the R-hat diagnostic.
#' Defaults to \code{1.05}.
#' @param min_ESS minimum estimated sample size.
#' Defaults to \code{500}.
#' @param max_error maximum value of the MCMC error.
#' Defaults to \code{NULL}. Be aware that PEESE publication bias
#' adjustment can have estimates on different scale than the rest of
#' the output, resulting in relatively large max MCMC error.
#' @param max_SD_error maximum value of the proportion of MCMC error
#' of the estimated SD of the parameter.
#' Defaults to \code{NULL}.
#' @param max_time list with the time and unit specifying the maximum
#' autofitting process per model. Passed to \link[base]{difftime} function
#' (possible units are \code{"secs", "mins", "hours", "days", "weeks", "years"}).
#' Defaults to \code{list(time = 60, unit = "mins")}.
#' @param sample_extend number of samples to extend the fitting process if
#' the criteria are not satisfied.
#' Defaults to \code{1000}.
#' @param restarts number of fitting restarts attempted when the initial JAGS
#' model fitting fails. Defaults to \code{10}.
#' @param max_extend maximum number of autofit extensions per model.
#' Defaults to \code{10}.
#' @param check_indicators whether convergence checks should include latent
#' prior inclusion indicators. Defaults to \code{FALSE}.
#' @param remove_failed whether models not satisfying the convergence checks should
#' be removed from the inference. Defaults to \code{FALSE} - only a warning is raised.
#' @param balance_probability whether prior model probability should be balanced
#' across the combinations of models with the same H0/H1 for effect / heterogeneity / bias
#' in the case of non-convergence. Defaults to \code{TRUE}.
#'
#'
#' @return \code{set_autofit_control} returns a list of autofit control settings
#' and \code{set_convergence_checks} returns a list of convergence checks settings.
#'
#' @export set_autofit_control
#' @export set_convergence_checks
#' @name RoBSA_control
#' @aliases set_autofit_control, set_convergence_checks
#'
#' @seealso [RoBSA], [update.RoBSA]
NULL

#' @rdname RoBSA_control
set_autofit_control     <- function(max_Rhat = 1.05, min_ESS = 500, max_error = NULL, max_SD_error = NULL, max_time = list(time = 60, unit = "mins"),
                                    sample_extend = 1000, restarts = 10, max_extend = 10, check_indicators = FALSE){

  autofit_settings <- list(
    max_Rhat         = max_Rhat,
    min_ESS          = min_ESS,
    max_error        = max_error,
    max_SD_error     = max_SD_error,
    max_time         = max_time,
    sample_extend    = sample_extend,
    restarts         = restarts,
    max_extend       = max_extend,
    check_indicators = check_indicators
  )
  autofit_settings <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_settings, call = "Checking 'autofit_control':\n\t")

  return(autofit_settings)
}
#' @rdname RoBSA_control
set_convergence_checks  <- function(max_Rhat = 1.05, min_ESS = 500, max_error = NULL, max_SD_error = NULL,
                                    check_indicators = FALSE, remove_failed = FALSE, balance_probability = TRUE){

  convergence_checks <- list(
    max_Rhat            = max_Rhat,
    min_ESS             = min_ESS,
    max_error           = max_error,
    max_SD_error        = max_SD_error,
    check_indicators    = check_indicators,
    remove_failed       = remove_failed,
    balance_probability = balance_probability
  )
  # allows NULL arguments so it can be used in this way too
  convergence_checks <- .check_and_list_convergence_checks(convergence_checks)

  return(convergence_checks)
}


.update_fit_control     <- function(old_fit_control, chains, adapt, burnin, sample, thin, autofit, parallel, cores, silent, seed){

  if(is.null(chains)){
    chains <- old_fit_control[["chains"]]
  }
  if(is.null(adapt)){
    adapt  <- old_fit_control[["adapt"]]
  }
  if(is.null(burnin)){
    burnin <- old_fit_control[["burnin"]]
  }
  if(is.null(sample)){
    sample <- old_fit_control[["sample"]]
  }
  if(is.null(thin)){
    thin  <- old_fit_control[["thin"]]
  }
  if(is.null(autofit)){
    autofit  <- old_fit_control[["autofit"]]
  }
  if(is.null(parallel)){
    parallel <- old_fit_control[["parallel"]]
  }
  if(is.null(silent)){
    silent <- old_fit_control[["silent"]]
  }
  if(is.null(seed)){
    seed   <- old_fit_control[["seed"]]
  }

  new_fit_control <- BayesTools::JAGS_check_and_list_fit_settings(chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin, autofit = autofit, parallel = parallel, cores = chains, silent = silent, seed = seed)

  return(new_fit_control)
}
.update_autofit_control <- function(old_autofit_control, autofit_control){

  if(!is.null(autofit_control[["max_Rhat"]])){
    max_Rhat <- autofit_control[["max_Rhat"]]
  }else{
    max_Rhat <- old_autofit_control[["max_Rhat"]]
  }
  if(!is.null(autofit_control[["min_ESS"]])){
    min_ESS <- autofit_control[["min_ESS"]]
  }else{
    min_ESS <- old_autofit_control[["min_ESS"]]
  }
  if(!is.null(autofit_control[["max_error"]])){
    max_error <- autofit_control[["max_error"]]
  }else{
    max_error <- old_autofit_control[["max_error"]]
  }
  if(!is.null(autofit_control[["max_SD_error"]])){
    max_SD_error <- autofit_control[["max_SD_error"]]
  }else{
    max_SD_error <- old_autofit_control[["max_SD_error"]]
  }
  if(!is.null(autofit_control[["max_time"]])){
    max_time <- autofit_control[["max_time"]]
  }else{
    max_time <- old_autofit_control[["max_time"]]
  }
  if(!is.null(autofit_control[["sample_extend"]])){
    sample_extend <- autofit_control[["sample_extend"]]
  }else{
    sample_extend <- old_autofit_control[["sample_extend"]]
  }
  if(!is.null(autofit_control[["restarts"]])){
    restarts <- autofit_control[["restarts"]]
  }else{
    restarts <- old_autofit_control[["restarts"]]
  }
  if(!is.null(autofit_control[["max_extend"]])){
    max_extend <- autofit_control[["max_extend"]]
  }else{
    max_extend <- old_autofit_control[["max_extend"]]
  }
  if(!is.null(autofit_control[["check_indicators"]])){
    check_indicators <- autofit_control[["check_indicators"]]
  }else{
    check_indicators <- old_autofit_control[["check_indicators"]]
  }

  new_autofit_control <- set_autofit_control(
    max_Rhat         = max_Rhat,
    min_ESS          = min_ESS,
    max_error        = max_error,
    max_SD_error     = max_SD_error,
    max_time         = max_time,
    sample_extend    = sample_extend,
    restarts         = restarts,
    max_extend       = max_extend,
    check_indicators = check_indicators
  )
  new_autofit_control <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_control = new_autofit_control)

  return(new_autofit_control)
}
.update_convergence_checks <- function(old_convergence_checks, convergence_checks){

  if(!is.null(convergence_checks[["max_Rhat"]])){
    max_Rhat <- convergence_checks[["max_Rhat"]]
  }else{
    max_Rhat <- old_convergence_checks[["max_Rhat"]]
  }
  if(!is.null(convergence_checks[["min_ESS"]])){
    min_ESS <- convergence_checks[["min_ESS"]]
  }else{
    min_ESS <- old_convergence_checks[["min_ESS"]]
  }
  if(!is.null(convergence_checks[["max_error"]])){
    max_error <- convergence_checks[["max_error"]]
  }else{
    max_error <- old_convergence_checks[["max_error"]]
  }
  if(!is.null(convergence_checks[["max_SD_error"]])){
    max_SD_error <- convergence_checks[["max_SD_error"]]
  }else{
    max_SD_error <- old_convergence_checks[["max_SD_error"]]
  }
  if(!is.null(convergence_checks[["check_indicators"]])){
    check_indicators <- convergence_checks[["check_indicators"]]
  }else{
    check_indicators <- old_convergence_checks[["check_indicators"]]
  }
  if(!is.null(convergence_checks[["remove_failed"]])){
    remove_failed <- convergence_checks[["remove_failed"]]
  }else{
    remove_failed <- old_convergence_checks[["remove_failed"]]
  }
  if(!is.null(convergence_checks[["balance_probability"]])){
    balance_probability <- convergence_checks[["balance_probability"]]
  }else{
    balance_probability <- old_convergence_checks[["balance_probability"]]
  }

  new_convergence_checks <- set_convergence_checks(
    max_Rhat            = max_Rhat,
    min_ESS             = min_ESS,
    max_error           = max_error,
    max_SD_error        = max_SD_error,
    check_indicators    = check_indicators,
    remove_failed       = remove_failed,
    balance_probability = balance_probability
  )
  new_convergence_checks <- .check_and_list_convergence_checks(new_convergence_checks)
}


.check_and_list_convergence_checks <- function(convergence_checks){

  remove_failed       <- convergence_checks[["remove_failed"]]
  balance_probability <- convergence_checks[["balance_probability"]]
  convergence_checks["remove_failed"]       <- NULL
  convergence_checks["balance_probability"] <- NULL
  convergence_checks <- BayesTools::JAGS_check_and_list_autofit_settings(convergence_checks, skip_sample_extend = TRUE, call = "Checking 'convergence_checks':\n\t")

  BayesTools::check_bool(remove_failed,       "remove_failed",       call = "Checking 'convergence_checks':\n\t")
  BayesTools::check_bool(balance_probability, "balance_probability", call = "Checking 'convergence_checks':\n\t")
  convergence_checks[["remove_failed"]]       <- remove_failed
  convergence_checks[["balance_probability"]] <- balance_probability
  return(convergence_checks)
}
.check_and_list_add_info           <- function(distributions, predictors, predictors_test, seed, save, rescale_data, warnings, errors){

  BayesTools::check_char(distributions, "distributions", allow_values = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"), check_length = FALSE)
  BayesTools::check_char(predictors, "predictors", allow_NULL = TRUE, check_length = FALSE)
  BayesTools::check_char(predictors_test, "predictors_test", allow_NULL = TRUE, check_length = FALSE)
  BayesTools::check_real(seed, "seed", allow_NULL = TRUE)
  BayesTools::check_char(save, "save", allow_values = c("min", "all"))
  BayesTools::check_bool(rescale_data, "rescale_data")

  return(list(
    distributions    = distributions,
    predictors       = predictors,
    predictors_test  = predictors_test,
    seed             = seed,
    save             = save,
    rescale_data     = rescale_data,
    warnings         = warnings,
    errors           = errors
  ))
}

.update_add_info <- function(old_add_info, distribution, predictors, predictors_test){

  if(!distribution %in% old_add_info[["distributions"]]){
    old_add_info[["distributions"]] <- c(old_add_info[["distributions"]], distribution)
  }

  if(any(!predictors %in% old_add_info[["predictors"]])){
    old_add_info[["predictors"]] <- c(
      old_add_info[["predictors"]],
      predictors[!predictors %in% old_add_info[["predictors"]]])
  }

  if(any(!predictors_test %in% old_add_info[["predictors_test"]])){
    old_add_info[["predictors_test"]] <- c(
      old_add_info[["predictors_test"]],
      predictors_test[!predictors_test %in% old_add_info[["predictors_test"]]])
  }


  return(old_add_info)
}


#' @title Default prior distributions
#'
#' @description Functions for setting default prior distributions. Note that
#' these default prior distributions might (and probably won't) apply to your
#' specific data scenario.
#'
#' @return \code{get_default_prior_beta_null} and \code{get_default_prior_beta_alt}
#' return a prior distribution and \code{get_default_prior_intercept} and
#' \code{get_default_prior_aux} return a list of prior distributions.
#'
#' @export get_default_prior_beta_null
#' @export get_default_prior_beta_alt
#' @export get_default_prior_factor_null
#' @export get_default_prior_factor_null
#' @export get_default_prior_intercept
#' @export get_default_prior_aux
#' @name default_prior
#' @aliases get_default_prior_beta_null, get_default_prior_beta_alt,
#' get_default_prior_factor_null, get_default_prior_factor_alt,
#' get_default_prior_intercept, get_default_prior_aux
NULL

#' @rdname default_prior
get_default_prior_beta_null <- function(){
  prior("spike", list(location = 0))
}
#' @rdname default_prior
get_default_prior_beta_alt  <- function(){
  prior("normal", list(mean = 0, sd = 1))
}
#' @rdname default_prior
get_default_prior_factor_null  <- function(){
  prior_factor("spike", parameters = list(location = 0), contrast = "treatment")
}
#' @rdname default_prior
get_default_prior_factor_alt   <- function(){
  prior_factor("normal",  list(mean = 0, sd = 1), contrast = "treatment")
}
#' @rdname default_prior
get_default_prior_intercept <- function(){
  list(
    "exp-aft"     = prior("normal", list(mean = 0, sd = 5)),
    "weibull-aft" = prior("normal", list(mean = 0, sd = 5)),
    "lnorm-aft"   = prior("normal", list(mean = 0, sd = 5)),
    "llogis-aft"  = prior("normal", list(mean = 0, sd = 5)),
    "gamma-aft"   = prior("normal", list(mean = 0, sd = 5))
  )
}
#' @rdname default_prior
get_default_prior_aux       <- function(){
  list(
    "exp-aft"     = NULL,
    "weibull-aft" = prior("normal", list(mean = 0, sd = 1), list(0, Inf)),
    "lnorm-aft"   = prior("normal", list(mean = 0, sd = 1), list(0, Inf)),
    "llogis-aft"  = prior("normal", list(mean = 0, sd = 1), list(0, Inf)),
    "gamma-aft"   = prior("normal", list(mean = 0, sd = 1), list(0, Inf))
  )
}

.fix_default_prior_factor_null <- function(old_prior, contrast){
  prior_factor(distribution = old_prior[["distribution"]], parameters = old_prior[["parameters"]], truncation = old_prior[["truncation"]], prior_weights = old_prior[["prior_weights"]], contrast = contrast)
}

# some functions for the JASP implementation
.RoBSA_collect_dots      <- function(...){

  dots <- list(...)

  known_dots <- c("is_JASP")
  if(any(!names(dots) %in% known_dots))
    stop(paste0("Uknown arguments to 'RoBSA': ", paste("'", names(dots)[!names(dots) %in% known_dots], "'", collapse = ", "), "."), call. = FALSE)

  if(is.null(dots[["is_JASP"]])){
    dots[["is_JASP"]] <- FALSE
  }else{
    dots[["is_JASP"]] <- TRUE
  }

  return(dots)
}
.JASP_progress_bar_start <- function(n){
  eval(expr = parse(text = 'startProgressbar(n)'))
}
.JASP_progress_bar_tick  <- function(){
  eval(expr = parse(text = 'progressbarTick()'))
}

#' @title Calibrate prior distributions based on quartiles
#'
#' @description Calibrates prior distributions for parametric
#' survival analysis based on median survival and interquartile
#' range. Returns a list of prior distribution for the intercepts
#' and auxiliary parameters.
#'
#' @param median_t median survival
#' @param iq_range_t interquartile range of the survival
#' @param prior_sd pre-specified standard deviation of the
#' prior distributions (either a single value that is used
#' for both the intercept and auxiliary parameter or a vector
#' where the first value corresponds to the sd for the prior
#' distribution on the intercept and the second value to the
#' sd for the prior distribution on the auxiliary parameter)
#' @param distributions vector of parametric families for which
#' prior distributions ought to be calibrated
#' @param verbose whether debug information be printed
#' @param search_bounds1 search boundaries for the intercept
#' parameter
#' @param search_bounds2 search boundaries for the auxiliary
#' parameter
#'
#' @examples
#' priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
#'
#'
#' @return returns a list of prior distribution for the intercepts
#' and auxiliary parameters.
#'
#' @export
calibrate_quartiles      <- function(median_t, iq_range_t, prior_sd = 0.5, distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
                                     verbose = FALSE, search_bounds1 = c(-1e2, 1e2), search_bounds2 = c(0 + 1e-2, 1e2)){

  parameters <- list()
  priors     <- list()
  priors_int <- list()
  priors_aux <- list()

  # set the same sd for both parameters if unspecified
  if(length(prior_sd) == 1){
    prior_sd <- rep(prior_sd, 2)
  }

  for(distribution in distributions){

    parameters[[distribution]] <- .calibrate_quantiles_fun(median_t, iq_range_t, distribution, verbose, search_bounds1, search_bounds2)
    priors[[distribution]]     <- .make_prior(parameters[[distribution]], prior_sd, distribution)

    priors_int[[distribution]] <- priors[[distribution]][["intercept"]]
    if(.has_aux(distribution)){
      priors_aux[[distribution]] <- priors[[distribution]][["aux"]]
    }
  }

  return(list(
    intercept = priors_int,
    aux       = priors_aux
  ))
}
.calibrate_quantiles_fun <- function(median_t, iq_range_t, distribution, verbose, search_bounds1, search_bounds2){

  if(verbose){
    cat(paste0("Distribution: ", distribution, "\n"))
  }


  if(.has_aux(distribution)){

    parameters <- stats::optim(
      par   = c(0, 1),
      fn    = .solve_quantiles2,
      lower = c(search_bounds1[1], search_bounds2[1]),
      upper = c(search_bounds1[2], search_bounds2[2]),
      median_t     = median_t,
      iq_range_t   = iq_range_t,
      quant_fun    = eval(parse(text = paste0(".", gsub("-", "_", distribution))))[["q"]],
      method       = "L-BFGS-B",
      verbose      = verbose
    )

    if(abs(parameters[["par"]][1] - search_bounds1[1]) < .1 | abs(parameters[["par"]][1] - search_bounds1[2]) < .1)
      warning("The calibrated prior distribution is close to search boundary. Consider specifying a wider search range.")

    if(abs(parameters[["par"]][2] - search_bounds2[1]) < .1 | abs(parameters[["par"]][2] - search_bounds2[2]) < .1)
      warning("The calibrated prior distribution is close to search boundary. Consider specifying a wider search range.")

    parameters <- list(
      intercept = parameters[["par"]][1],
      aux       = parameters[["par"]][2]
    )

  }else{

    parameters <- stats::optim(
      par   = 0,
      fn    = .solve_quantiles1,
      lower = search_bounds1[1],
      upper = search_bounds1[2],
      median_t     = median_t,
      quant_fun    = eval(parse(text = paste0(".", gsub("-", "_", distribution))))[["q"]],
      method       = "L-BFGS-B",
      verbose      = verbose
    )

    if(abs(parameters[["par"]][1] - search_bounds1[1]) < .1 | abs(parameters[["par"]][1] - search_bounds1[2]) < .1)
      warning("The calibrated prior distribution is close to search boundary. Consider specifying a wider search range.")

    parameters <- list(
      intercept = parameters[["par"]][1]
    )

  }

  if(verbose){
    cat("Parameters: \n")
    print(parameters)
    cat("\n\n")
  }

  return(parameters)
}

.solve_quantiles1 <- function(x, median_t, quant_fun, verbose){

  intercept <- x[1]

  res_median_t <- do.call(quant_fun, list(.5, eta = intercept))

  MSE <- (res_median_t - median_t)^2

  if(verbose){
    print(data.frame(intercept, res_median_t, MSE))
  }

  return(MSE)
}
.solve_quantiles2 <- function(x, median_t, iq_range_t, quant_fun, verbose){

  intercept <- x[1]
  aux       <- x[2]

  res_median_t   <- do.call(quant_fun, list(.50, eta = intercept, aux))
  res_25q_t      <- do.call(quant_fun, list(.25, eta = intercept, aux))
  res_75q_t      <- do.call(quant_fun, list(.75, eta = intercept, aux))
  res_iq_range_t <- res_75q_t - res_25q_t

  MSE <- (res_median_t - median_t)^2 + (res_iq_range_t - iq_range_t)^2

  if(verbose){
    print(data.frame(intercept, aux, res_median_t, res_iq_range_t, MSE))
  }

  if(is.na(MSE)){
    MSE <- 1e100
  }

  return(MSE)
}

.make_prior   <- function(parameters, prior_sd, distribution){

  if(.has_aux(distribution)){

    priors <- list(
      intercept = prior("normal",    list(parameters[["intercept"]], prior_sd[1])),
      aux       = prior("lognormal", list(lnorm_meanlog(parameters[["aux"]], prior_sd[2]), lnorm_sdlog(parameters[["aux"]], prior_sd[2])))
    )

  }else{

    priors <- list(
      intercept = prior("normal",    list(parameters[["intercept"]], prior_sd[1]))
    )

  }

  return(priors)
}

#' @title Create meta-analytic predictive prior distributions
#'
#' @description Calibrates prior distributions for parametric
#' survival analysis based on historical data.
#' Returns a list of prior distribution for the intercepts
#' and auxiliary parameters.
#'
#' @param datasets list of data.frames containing the historical
#' data. Each data.frame must contain a column named \code{"time"}
#' with the survival times and a column named \code{"status"}
#' with the censoring status.
#' @param distributions vector of parametric families for which
#' prior distributions ought to be calibrated
#' @param prior_mu prior distribution for the the meta-analytic mean parameter
#' @param prior_tau prior distribution for the the meta-analytic heterogeneity parameter
#' @param ... additional parameters to be passed to the meta-analytic function.
#' See [BayesTools::JAGS_fit] for more details.
#'
#'
#' @return returns a list of prior distribution for the intercepts
#' and auxiliary parameters.
#'
#' @export
calibrate_meta_analytic  <- function(datasets, distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
                                     prior_mu  = prior("cauchy", parameters = list(location = 0, scale = 100)),
                                     prior_tau = prior("cauchy", parameters = list(location = 0, scale = 10), truncation = list(0, Inf)), ...){

  if(!is.list(datasets))
    stop("'datasets' must be a list")
  if(!all(sapply(datasets, is.data.frame)))
    stop("'datasets' must be a list of data.frames")
  if(!all(sapply(datasets, function(df) all(c("time", "status") %in% colnames(df)))))
    stop("'datasets' must be a list of data.frames that contains 'time' and 'status' columns describing the survival times and censoring")
  if(!BayesTools::is.prior(prior_mu) | !BayesTools::is.prior(prior_tau))
    stop("'prior_mu' and 'prior_tau' must be prior distributions")
  requireNamespace("flexsurv")

  # prepare output
  priors_int <- list()
  priors_aux <- list()

  # fit the survival models with flexsurv
  distributions <- unname(sapply(distributions, .RoBSA2flex_distribution))

  surv_fits <- list()
  for(df in seq_along(datasets)){
    surv_fits[[df]] <- list()
    for (d in distributions){
      surv_fits[[df]][[d]] <- suppressWarnings(flexsurv::flexsurvreg(Surv(time, status) ~ 1, data = datasets[[df]][datasets[[df]]$time != 0,], dist = d))
    }
  }

  coefs    <- sapply(distributions, function(fam) data.frame(do.call(rbind, sapply(surv_fits, function(df) stats::coef(df[[fam]]), simplify = FALSE))), simplify = FALSE)
  coefs_se <- sapply(distributions, function(fam) data.frame(do.call(rbind, sapply(surv_fits, function(df) sqrt(diag(df[[fam]]$cov)), simplify = FALSE))), simplify = FALSE)
  colnames(coefs$exp)    <- "rate"
  colnames(coefs_se$exp) <- "rate"

  # fit meta-analytic priors
  coefs_meta <- list()
  for(fam in distributions){
    for(par in names(coefs[[fam]])){

      temp_fit <- .random_effects_metaanalysis(
        y   = coefs[[fam]][[par]],
        se  = coefs_se[[fam]][[par]],
        prior_mu  = prior_mu,
        prior_tau = prior_tau,
        ...)


      # deal with the negative parametrization for exp and gamma family
      if(fam %in% c("exp", "gamma") & par == "rate"){

        priors_int[[.flex2RoBSA_distribution(fam)]] <- prior(
          distribution = "normal",
          parameters   = list(
            mean = - temp_fit["mu", "Mean"],
            sd   = sqrt(temp_fit["mu", "SD"]^2 + temp_fit["tau", "Mean"]^2)
          ))

      }else{

        if(par %in% c("rate", "scale", "meanlog")){

          priors_int[[.flex2RoBSA_distribution(fam)]] <- prior(
            distribution = "normal",
            parameters   = list(
              mean = temp_fit["mu", "Mean"],
              sd   = sqrt(temp_fit["mu", "SD"]^2 + temp_fit["tau", "Mean"]^2)
          ))

        }else{

          priors_aux[[.flex2RoBSA_distribution(fam)]] <- prior(
            distribution = "lognormal",
            parameters   = list(
              meanlog = temp_fit["mu", "Mean"],
              sdlog   = sqrt(temp_fit["mu", "SD"]^2 + temp_fit["tau", "Mean"]^2)
          ))

        }

      }
    }
  }

  return(list(
    intercept = priors_int,
    aux       = priors_aux
  ))
}

.random_effects_metaanalysis <- function(y, se, prior_mu, prior_tau, ...){

  model_syntax <- "model{
        for(i in 1:K){
          y[i] ~ dnorm(mu, 1/(pow(tau, 2) + pow(se[i], 2)))
        }
        }"

  fit <- BayesTools::JAGS_fit(
    model_syntax = model_syntax,
    data         = list(y = y, se = se, K = length(y)),
    prior_list   = list(mu = prior_mu, tau = prior_tau),
    ...)

  fit_summary <- BayesTools::runjags_estimates_table(fit)

  return(fit_summary)
}

.RoBSA2flex_distribution <- function(distribution){
  return(switch(
    distribution,
    "exp-aft"     = "exp",
    "weibull-aft" = "weibull",
    "lnorm-aft"   = "lognormal",
    "llogis-aft"  = "llogis",
    "gamma-aft"   = "gamma"
  ))
}
.flex2RoBSA_distribution <- function(distribution){
  return(switch(
    distribution,
    "exp"       = "exp-aft",
    "weibull"   = "weibull-aft",
    "lognormal" = "lnorm-aft",
    "llogis"    = "llogis-aft",
    "gamma"     = "gamma-aft"
  ))
}

#' @title Plots a fitted RoBSA object
#'
#' @description \code{plot.RoBSA} allows to visualize
#' posterior distribution of different \code{"RoBSA"} object
#' parameters. See \code{plot_survival} for plotting the survival
#' ways. See \code{type} for the different model types.
#'
#' @param x a fitted RoBSA object
#' @param parameter a name of parameter to be plotted. Defaults to
#' the first regression parameter if left unspecified. Use
#' \code{"intercept"} and \code{"aux"} to plot the intercepts and
#' auxiliary parameters of each distribution family.
#' @param conditional whether conditional estimates should be
#' plotted. Defaults to \code{FALSE} which plots the model-averaged
#' estimates.
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting. Defaults to
#' \code{"base"}.
#' @param prior whether prior distribution should be added to
#' figure. Defaults to \code{FALSE}.
#' @param dots_prior list of additional graphical arguments
#' to be passed to the plotting function of the prior
#' distribution. Supported arguments are \code{lwd},
#' \code{lty}, \code{col}, and \code{col.fill}, to adjust
#' the line thickness, line type, line color, and fill color
#' of the prior distribution respectively.
#' @param ... list of additional graphical arguments
#' to be passed to the plotting function. Supported arguments
#' are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
#' \code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim}
#' to adjust the line thickness, line type, line color, fill color,
#' x-label, y-label, title, x-axis range, and y-axis range
#' respectively.
#'
#' @examples \dontrun{
#' # (execution of the example takes several minutes)
#' # example from the README (more details and explanation therein)
#' data(cancer, package = "survival")
#' priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
#' df <- data.frame(
#'   time         = veteran$time / 12,
#'   status       = veteran$status,
#'   treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
#'   karno_scaled = veteran$karno / 100
#' )
#' RoBSA.options(check_scaling = FALSE)
#' fit <- RoBSA(
#'   Surv(time, status) ~ treatment + karno_scaled,
#'   data   = df,
#'   priors = list(
#'     treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
#'                                 truncation = list(0, Inf), contrast = "treatment"),
#'     karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
#'   ),
#'   test_predictors = "treatment",
#'   prior_intercept = priors[["intercept"]],
#'   prior_aux       = priors[["aux"]],
#'   parallel = TRUE, seed = 1
#' )
#'
#' # plot posterior distribution of the treatment effect
#' plot(fit, parameter = "treatment")
#'
#' }
#'
#'
#' @return \code{plot.RoBSA} returns either \code{NULL} if \code{plot_type = "base"}
#' or an object object of class 'ggplot2' if \code{plot_type = "ggplot2"}.
#'
#' @seealso [RoBSA()]
#' @export
plot.RoBSA  <- function(x, parameter = NULL, conditional = FALSE, plot_type = "base", prior = FALSE, dots_prior = NULL, ...){

  # check whether plotting is possible
  if(sum(.get_model_convergence(x)) == 0)
    stop("There is no converged model in the ensemble.")

  #check settings
  BayesTools::check_char(parameter, "parameter", allow_NULL = TRUE)
  BayesTools::check_bool(conditional, "conditional")
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(prior, "prior")

  # verify whether given parameter exists and set a default parameter for plotting
  if(is.null(parameter)){
    parameter <- x$add_info[["predictors"]][1]
  }else if(!parameter %in% c("intercept", "aux", x$add_info[["predictors"]])){
    stop(paste0("The passed parameter does not correspond to any of the specified predictors: ", paste0("'", x$add_info[["predictors"]], "'", collapse = ", ")))
  }


  # choose the samples
  if(parameter == "intercept"){
    samples <- x[["RoBSA"]][["posteriors_intercept"]]
  }else if(parameter == "aux"){
    samples <- x[["RoBSA"]][["posteriors_aux"]]
  }else{
    if(conditional){
      samples <- x[["RoBSA"]][["posteriors_conditional"]]
    }else{
      samples <- x[["RoBSA"]][["posteriors"]]
    }
  }


  dots       <- .set_dots_plot(...)
  dots_prior <- .set_dots_prior(dots_prior)

  # prepare the argument call
  args                          <- dots
  args$samples                  <- samples
  args$plot_type                <- plot_type
  args$prior                    <- prior
  args$n_points                 <- 1000
  args$n_samples                <- 10000
  args$force_samples            <- FALSE
  args$transformation           <- NULL
  args$transformation_arguments <- NULL
  args$transformation_settings  <- FALSE
  args$rescale_x                <- FALSE
  args$dots_prior               <- dots_prior

  if(parameter %in% c("intercept", "aux")){
    plot <- list()
    for(i in seq_along(samples)){
      args$parameter <- names(samples)[i]
      args$par_name  <- names(samples)[i]
      plot[[names(samples)[i]]] <- do.call(BayesTools::plot_posterior, args)
    }
  }else{
    args$parameter <- .BayesTools_parameter_name(parameter)
    args$par_name  <- parameter
    plot           <- do.call(BayesTools::plot_posterior, args)
  }


  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}


.set_dots_plot        <- function(...){

  dots <- list(...)
  if(is.null(dots[["col"]])){
    dots[["col"]]      <- "black"
  }
  if(is.null(dots[["col.fill"]])){
    dots[["col.fill"]] <- "#4D4D4D4C" # scales::alpha("grey30", .30)
  }

  return(dots)
}
.set_dots_prior       <- function(dots_prior){

  if(is.null(dots_prior)){
    dots_prior <- list()
  }

  if(is.null(dots_prior[["col"]])){
    dots_prior[["col"]]      <- "grey60"
  }
  if(is.null(dots_prior[["lty"]])){
    dots_prior[["lty"]]      <- 1
  }
  if(is.null(dots_prior[["col.fill"]])){
    dots_prior[["col.fill"]] <- "#B3B3B34C" # scales::alpha("grey70", .30)
  }

  return(dots_prior)
}

.normalize_fit_summary_quantile_names <- function(fit_summary){

  if(is.null(fit_summary))
    return(fit_summary)

  columns <- colnames(fit_summary)
  rename_map  <- c(
    "lCI"    = "0.025",
    "Median" = "0.5",
    "uCI"    = "0.975"
  )

  for(old_name in names(rename_map)){
    new_name <- rename_map[[old_name]]
    if(old_name %in% columns && !(new_name %in% columns)){
      columns[columns == old_name] <- new_name
    }
  }

  colnames(fit_summary) <- columns
  return(fit_summary)
}


#' @title Models plot for a RoBSA object
#'
#' @description \code{plot_models} plots individual models'
#' estimates for a \code{"RoBSA"} object.
#'
#' @param parameter a name of parameter to be plotted. Defaults to
#' the first regression parameter if left unspecified.
#' @param order how the models should be ordered.
#' Defaults to \code{"decreasing"} which orders them in decreasing
#' order in accordance to \code{order_by} argument. The alternative is
#' \code{"increasing"}.
#' @param order_by what feature should be use to order the models.
#' Defaults to \code{"model"} which orders the models according to
#' their number. The alternatives are \code{"estimate"} (for the effect
#' size estimates), \code{"probability"} (for the posterior model probability),
#' and \code{"BF"} (for the inclusion Bayes factor).
#' @inheritParams plot.RoBSA
#'
#' @examples \dontrun{
#' # (execution of the example takes several minutes)
#' # example from the README (more details and explanation therein)
#' data(cancer, package = "survival")
#' priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
#' df <- data.frame(
#'   time         = veteran$time / 12,
#'   status       = veteran$status,
#'   treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
#'   karno_scaled = veteran$karno / 100
#' )
#' RoBSA.options(check_scaling = FALSE)
#' fit <- RoBSA(
#'   Surv(time, status) ~ treatment + karno_scaled,
#'   data   = df,
#'   priors = list(
#'     treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
#'                                 truncation = list(0, Inf), contrast = "treatment"),
#'     karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
#'   ),
#'   test_predictors = "treatment",
#'   prior_intercept = priors[["intercept"]],
#'   prior_aux       = priors[["aux"]],
#'   parallel = TRUE, seed = 1
#' )
#'
#' # plot posterior distribution of the treatment effect from each model
#' plot_models(fit, parameter = "treatment")
#'
#' }
#'
#'
#' @return \code{plot_models} returns either \code{NULL} if \code{plot_type = "base"}
#' or an object object of class 'ggplot2' if \code{plot_type = "ggplot2"}.
#'
#' @export
plot_models <- function(x, parameter = NULL, conditional = FALSE, plot_type = "base", order = "decreasing", order_by = "model", ...){

  if(!is.RoBSA(x))
    stop("'plot_models' function supports only 'RoBSA' type objects.")
  if(sum(.get_model_convergence(x)) == 0)
    stop("There is no converged model in the ensemble.")

  #check settings
  BayesTools::check_char(parameter, "parameter", allow_NULL = TRUE)
  BayesTools::check_bool(conditional, "conditional")
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_char(order, "order", allow_NULL = TRUE, allow_values = c("increasing", "decreasing"))
  BayesTools::check_char(order, "order", allow_NULL = TRUE, allow_values = c("increasing", "decreasing"))
  BayesTools::check_char(order_by, "order_by", allow_NULL = TRUE, allow_values = c("model", "estimate", "probability", "BF"))


  # verify whether given parameter exists and set a default parameter for plotting
  if(is.null(parameter)){
    parameter <- x$add_info[["predictors"]][1]
  }else if(!parameter %in% x$add_info[["predictors"]]){
    stop(paste0("The passed parameter does not correspond to any of the specified predictors: ", paste0("'", x$add_info[["predictors"]], "'", collapse = ", ")))
  }


  ### prepare input
  if(conditional){

    model_list <- x[["models"]]
    samples    <- x[["RoBSA"]][["posteriors_conditional"]]
    inference  <- x[["RoBSA"]][["inference_conditional"]]

  }else{

    model_list <- x[["models"]]
    samples    <- x[["RoBSA"]][["posteriors"]]
    inference  <- x[["RoBSA"]][["inference"]]

  }

  model_list <- lapply(model_list, function(model){
    model[["fit_summary"]] <- .normalize_fit_summary_quantile_names(model[["fit_summary"]])
    return(model)
  })

  dots <- list(...)

  # prepare the argument call
  args                          <- dots
  args$model_list               <- model_list
  args$samples                  <- samples
  args$inference                <- inference
  args$parameter                <- .BayesTools_parameter_name(parameter)
  args$par_name                 <- NULL
  args$plot_type                <- plot_type
  args$prior                    <- FALSE
  args$conditional              <- conditional
  args$order                    <- list(order, order_by)
  args$transformation           <- NULL
  args$transformation_arguments <- NULL
  args$transformation_settings  <- FALSE
  args$formula_prefix           <- FALSE

  plot <- do.call(BayesTools::plot_models, args)


  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}


#' @title Survival plots for a RoBSA object
#'
#' @param x a fitted RoBSA object.
#' @param time_range a numeric of length two specifying the range for the
#' survival prediction. Defaults to \code{NULL} which uses the range of
#' observed times.
#'
#' @inheritParams predict.RoBSA
#' @inheritParams plot.RoBSA
#' @param ... additional arguments.
#'
#' @examples \dontrun{
#' # (execution of the example takes several minutes)
#' # example from the README (more details and explanation therein)
#' data(cancer, package = "survival")
#' priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
#' df <- data.frame(
#'   time         = veteran$time / 12,
#'   status       = veteran$status,
#'   treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
#'   karno_scaled = veteran$karno / 100
#' )
#' RoBSA.options(check_scaling = FALSE)
#' fit <- RoBSA(
#'   Surv(time, status) ~ treatment + karno_scaled,
#'   data   = df,
#'   priors = list(
#'     treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
#'                                 truncation = list(0, Inf), contrast = "treatment"),
#'     karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
#'   ),
#'   test_predictors = "treatment",
#'   prior_intercept = priors[["intercept"]],
#'   prior_aux       = priors[["aux"]],
#'   parallel = TRUE, seed = 1
#' )
#'
#' # plot survival for each level the treatment
#' plot_survival(fit, parameter = "treatment")
#'
#' # plot hazard for each level the treatment
#' plot_hazard(fit, parameter = "treatment")
#'
#' # plot density for each level the treatment
#' plot_density(fit, parameter = "treatment")
#' }
#'
#' @return returns either \code{NULL} if \code{plot_type = "base"}
#' or an object object of class 'ggplot2' if \code{plot_type = "ggplot2"}.
#'
#' @export plot_prediction
#' @export plot_survival
#' @export plot_hazard
#' @export plot_density
#' @name plot_prediction
#' @aliases plot_survival plot_hazard plot_density
NULL

#' @rdname plot_prediction
plot_prediction <- function(x, type = "survival", time_range = NULL, new_data = NULL, predictor = NULL, covariates_data = NULL,
                            conditional = FALSE, plot_type = "base", samples = 10000, ...){

  if(!is.RoBSA(x))
    stop("'plot_models' function supports only 'RoBSA' type objects.")
  BayesTools::check_real(time_range, "time_range", lower = 0, allow_NULL = TRUE, check_length = TRUE)
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  # the other input is checked within predict.RoBSA

  # set a default time_range
  if(is.null(time_range)){
    time_range <- c(0, max(c(
      x[["data"]][["survival"]][["t_event"]],
      x[["data"]][["survival"]][["t_cens_r"]],
      x[["data"]][["survival"]][["t_cens_ir"]]
    )))
    time_range <- range(pretty(time_range))
  }
  time_range <- seq(time_range[1], time_range[2], length.out = 101)
  if(type == "density"){
    time_range <- time_range[-1]
  }

  # obtain predictions for the survival
  plot_data  <- predict.RoBSA(x, time = time_range, new_data = new_data, predictor = predictor, covariates_data = covariates_data,
                              type = type, summarize = TRUE, averaged = TRUE, conditional = conditional, samples = samples)

  # generate plotting settings
  dots <- .set_dots_plot_prediction(plot_data, ...)


  # create times for the prediction
  if(plot_type == "base"){

    plot <- NULL
    graphics::plot(NA, type = "n", bty  = "n", las = 1, xlab = dots[["xlab"]], ylab = dots[["ylab"]], main = dots[["main"]],
                   xlim = dots[["xlim"]], ylim = dots[["ylim"]],
                   cex.axis = dots[["cex.axis"]], cex.lab = dots[["cex.lab"]], cex.main = dots[["cex.main"]],
                   col.axis = dots[["col.axis"]], col.lab = dots[["col.lab"]], col.main = dots[["col.main"]])

    for(i in seq_along(plot_data)){
      graphics::polygon(
        x   = c(plot_data[[i]][,"time"], rev(plot_data[[i]][,"time"])),
        y   = c(plot_data[[i]][,"lCI"],  rev(plot_data[[i]][,"uCI"])),
        col = dots[["col.fill"]][i], border = NA
      )
    }

    for(i in seq_along(plot_data)){
      graphics::lines(plot_data[[i]][,"time"], plot_data[[i]][,"mean"], lwd = dots[["lwd"]][i], lty = dots[["lty"]][i], col = dots[["col"]][i])
    }

    if(dots[["legend"]]){
      graphics::legend(
        dots[["legend.position"]],
        legend = dots[["legend.text"]],
        col    = dots[["col"]],
        lty    = dots[["lty"]],
        lwd    = dots[["lwd"]],
        bty    = "n")
    }

  }else if(plot_type == "ggplot"){

    plot <- ggplot2::ggplot()+
      ggplot2::ggtitle(dots[["main"]]) +
      ggplot2::scale_x_continuous(name = dots[["xlab"]], limits = dots[["xlim"]], oob = scales::oob_keep) +
      ggplot2::scale_y_continuous(name = dots[["ylab"]], limits = dots[["ylim"]], oob = scales::oob_keep)

    for(i in seq_along(plot_data)){
      plot <- plot + ggplot2::geom_polygon(
        data    = data.frame(
          x = c(plot_data[[i]][,"time"], rev(plot_data[[i]][,"time"])),
          y = c(plot_data[[i]][,"lCI"],  rev(plot_data[[i]][,"uCI"]))),
        mapping = ggplot2::aes(
          x = .data[["x"]],
          y = .data[["y"]]),
        fill    = dots[["col.fill"]][i]
      )
    }

    for(i in seq_along(plot_data)){
      plot <- plot + ggplot2::geom_line(
        data    = data.frame(
          x     = plot_data[[i]][,"time"],
          y     = plot_data[[i]][,"mean"],
          level = if(dots[["legend"]]) dots[["legend.text"]][i] else ""),
        mapping = ggplot2::aes(
          x         = .data[["x"]],
          y         = .data[["y"]],
          color     = if(length(dots[["legend.text"]]) > 1) .data[["level"]],
          linetype  = if(length(dots[["legend.text"]]) > 1) .data[["level"]],
          group     = if(length(dots[["legend.text"]]) > 1) .data[["level"]]),
        show.legend = dots[["legend"]])
    }

    if(dots[["legend"]]){
      names(dots[["lty"]]) <- dots[["legend.text"]]
      names(dots[["col"]]) <- dots[["legend.text"]]
      names(dots[["lwd"]]) <- dots[["legend.text"]]
      plot <- plot +
        ggplot2::scale_linetype_manual(name = "level", values = dots[["lty"]]) +
        ggplot2::scale_color_manual(name = "level", values = dots[["col"]]) +
        ggplot2::scale_size_manual(name = "level", values = dots[["lwd"]]) +
        ggplot2::theme(
        legend.title    = ggplot2::element_blank(),
        legend.position = dots[["legend_position"]])
    }
  }

  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
#' @rdname plot_prediction
plot_survival   <- function(x, time_range = NULL, new_data = NULL, predictor = NULL, covariates_data = NULL,
                            conditional = FALSE, plot_type = "base", samples = 10000, ...){
  plot_prediction(x, type = "survival", time_range = time_range, new_data = new_data, predictor = predictor, covariates_data = covariates_data,
                  conditional = conditional, plot_type = plot_type, samples = samples, ...)
}
#' @rdname plot_prediction
plot_hazard     <- function(x, time_range = NULL, new_data = NULL, predictor = NULL, covariates_data = NULL,
                            conditional = FALSE, plot_type = "base", samples = 10000, ...){
  plot_prediction(x, type = "hazard", time_range = time_range, new_data = new_data, predictor = predictor, covariates_data = covariates_data,
                  conditional = conditional, plot_type = plot_type, samples = samples, ...)
}
#' @rdname plot_prediction
plot_density    <- function(x, time_range = NULL, new_data = NULL, predictor = NULL, covariates_data = NULL,
                            conditional = FALSE, plot_type = "base", samples = 10000, ...){
  plot_prediction(x, type = "density", time_range = time_range, new_data = new_data, predictor = predictor, covariates_data = covariates_data,
                  conditional = conditional, plot_type = plot_type, samples = samples, ...)
}

.set_dots_plot_prediction <- function(plot_data, outcome, ...){

  dots   <- list(...)
  data   <- attr(plot_data, "data")
  levels <- nrow(data)

  if(is.null(dots[["xlab"]])){
    dots[["xlab"]] <- "Time"
  }

  if(is.null(dots[["ylab"]])){
    dots[["ylab"]] <- switch(
      attr(plot_data, "outcome"),
      "survival" = "Survival",
      "hazard"   = "Hazard",
      "density"  = "Density"
    )
  }

  if(is.null(dots[["main"]])){
    dots[["main"]] <- NULL
  }

  if(is.null(dots[["xlim"]])){
    dots[["xlim"]] <- range(attr(plot_data, "time"))
  }

  if(is.null(dots[["ylim"]])){
    dots[["ylim"]] <- switch(
      attr(plot_data, "outcome"),
      "survival" = c(0, 1),
      "hazard"   = c(0, 1),
      "density"  = range(pretty(range(do.call(c, lapply(plot_data, function(d) d[,-1])))))
    )
  }

  if(is.null(dots[["col"]])){
    if(levels == 1){
      dots[["col"]] <- "black"
    }else{
      dots[["col"]] <- scales::viridis_pal()(levels)
    }
  }else if(length(dots[["col"]]) == 1){
    dots[["col"]] <- rep(dots[["col"]], levels)
  }else if(dots[["col"]] != levels){
    stop("The number of specified colors does not match the number of predicted variable levels.")
  }

  if(is.null(dots[["col.fill"]])){
    if(levels == 1){
      dots[["col.fill"]] <- scales::alpha("grey30", alpha = .30)
    }else{
      dots[["col.fill"]] <- scales::alpha(dots[["col"]], alpha = .30)
    }
  }else if(length(dots[["col.fill"]]) == 1){
    dots[["col.fill"]] <- rep(dots[["col.fill"]], levels)
  }else if(dots[["col.fill"]] != levels){
    stop("The number of specified filling colors does not match the number of predicted variable levels.")
  }

  if(is.null(dots[["lwd"]])){
    dots[["lwd"]] <- rep(1, levels)
  }else if(length(dots[["lwd"]]) == 1){
    dots[["lwd"]] <- rep(dots[["lwd"]], levels)
  }else if(dots[["lwd"]] != levels){
    stop("The number of specified line widths (lwd) colors does not match the number of predicted variable levels.")
  }

  if(is.null(dots[["lty"]])){
    dots[["lty"]] <- rep(1, levels)
  }else if(length(dots[["lty"]]) == 1){
    dots[["lty"]] <- rep(dots[["lty"]], levels)
  }else if(dots[["lty"]] != levels){
    stop("The number of specified line types (lty) colors does not match the number of predicted variable levels.")
  }

  if(is.null(dots[["legend"]])){
    dots[["legend"]] <- TRUE
  }

  if(is.null(dots[["legend.text"]])){
    # remove constant predictors for defaults legend
    for(i in ncol(data):1){
      if(length(unique(data[,i])) == 1){
        data <- data[,-i, drop = FALSE]
      }
    }
    if(ncol(data) >= 1){
      dots[["legend.text"]] <- sapply(1:nrow(data), function(i) paste0(
        colnames(data), " = ", sapply(data[i,], function(x) if(is.numeric(x)) sprintf("%.2f", x) else x),
        collapse = "; "))
    }
  }

  if(is.null(dots[["legend.position"]])){
    dots[["legend.position"]] <- "topright"
  }

  if(length(dots[["legend.text"]]) == 0){
    dots[["legend"]] <- FALSE
  }


  return(dots)
}

# the main functions
.fit_RoBSA_model <- function(object, i){

  model              <- object[["models"]][[i]]
  priors             <- model[["priors"]]
  fit_control        <- object[["fit_control"]]
  autofit_control    <- object[["autofit_control"]]
  convergence_checks <- object[["convergence_checks"]]
  formula            <- object[["formula"]]
  data               <- object[["data"]]

  errors   <- NULL
  warnings <- NULL

  if(!fit_control[["silent"]]){
    cat(paste0("\nFitting model [", i, "]\n"))
  }

  # don't sample the complete null model
  if(!.is_model_constant(priors)){

    # generate the model syntax
    model_syntax <- .generate_model_syntax(model[["distribution"]], data)

    formula_list        <- .generate_model_formula_list(formula)
    formula_data_list   <- .generate_model_formula_data_list(data)
    formula_prior_list  <- .generate_model_formula_prior_list(priors)

    # remove unnecessary objects from data to mitigate warnings
    fit_data   <- .fit_data(data)
    fit_priors <- .fit_priors(priors)

    # fit the model
    fit <- BayesTools::JAGS_fit(
      model_syntax       = model_syntax,
      data               = fit_data,
      prior_list         = fit_priors,
      formula_list       = formula_list,
      formula_data_list  = formula_data_list,
      formula_prior_list = formula_prior_list,
      chains             = fit_control[["chains"]],
      adapt              = fit_control[["adapt"]],
      burnin             = fit_control[["burnin"]],
      sample             = fit_control[["sample"]],
      thin               = fit_control[["thin"]],
      autofit            = fit_control[["autofit"]],
      autofit_control    = autofit_control,
      parallel           = fit_control[["parallel"]],
      cores              = fit_control[["cores"]],
      silent             = fit_control[["silent"]],
      seed               = fit_control[["seed"]],
      required_packages  = "RoBSA"
    )

    # assess the model fit and deal with errors
    if(inherits(fit, "error")){

      if(grepl("Unknown function", fit$message))
        stop("The RoBSA module could not be loaded. Check whether the RoBSA package was installed correctly and whether 'RoBSA::RoBSA.private$module_location' contains path to the RoBSA JAGS module.")

      fit            <- list()
      converged      <- FALSE
      has_posterior  <- FALSE
      errors         <- c(errors, fit$message)
      # deal with failed models
      marglik        <- list()
      marglik$logml  <- NA
      class(marglik) <- "bridge"

    }else{

      has_posterior <- TRUE
      check_fit     <- BayesTools::JAGS_check_convergence(
        fit          = fit,
        prior_list   = attr(fit, "prior_list"),
        max_Rhat     = convergence_checks[["max_Rhat"]],
        min_ESS      = convergence_checks[["min_ESS"]],
        max_error        = convergence_checks[["max_error"]],
        max_SD_error     = convergence_checks[["max_SD_error"]],
        check_indicators = convergence_checks[["check_indicators"]]
      )
      warnings    <- c(warnings, attr(fit, "warnings"), attr(check_fit, "errors"))
      if(convergence_checks[["remove_failed"]] && !check_fit){
        converged <- FALSE
      }else{
        converged <- TRUE
      }

    }

    # compute marginal likelihood
    if(length(fit) != 0){

      marglik <- BayesTools::JAGS_bridgesampling(
        fit                = fit,
        log_posterior      = .marglik_function,
        data               = data,
        prior_list         = fit_priors,
        formula_list       = formula_list,
        formula_data_list  = formula_data_list,
        formula_prior_list = formula_prior_list,
        maxiter            = 50000,
        silent             = fit_control[["silent"]],
        distribution       = model[["distribution"]]
      )

      # deal with failed marginal likelihoods
      if(inherits(marglik, "error")){

        errors         <- c(errors, marglik$message)
        converged      <- FALSE
        marglik        <- list()
        marglik$logml  <- NA
        class(marglik) <- "bridge"

      }else{

        # forward warnings if present
        warnings <- c(warnings, attr(marglik, "warnings"))

      }
    }


  }else{

    fit_data                <- .fit_data(object[["data"]])
    converged               <- TRUE
    has_posterior           <- FALSE
    fit                     <- list()
    attr(fit, "prior_list") <- priors
    class(fit)              <- "null_model"
    marglik                 <- list()
    marglik$logml           <- .marglik_function_null(priors, data, model[["distribution"]])
    class(marglik)          <- "bridge"

  }

  # add model summaries
  if(has_posterior){
    fit_summary <- BayesTools::runjags_estimates_table(fit = fit, warnings = warnings, transform_factors = TRUE, formula_prefix = FALSE)
  }else{
    fit_summary <- BayesTools::runjags_estimates_empty_table()
  }

  model <- c(
    model,
    fit           = list(fit),
    fit_summary   = list(fit_summary),
    marglik       = list(marglik),
    errors        = list(errors),
    warnings      = list(warnings),
    converged     = converged,
    has_posterior = has_posterior
  )

  return(model)
}

# model parameter tools
.fit_data    <- function(data){

  types <- attr(data[["survival"]], "type")

  fit_data <- list("time" = unname(do.call(c, data[["survival"]][c(
    if(any(types == "event"))  "t_event",
    if(any(types == "cens_r")) "t_cens_r",
    if(any(types == "cens_l")) "t_cens_l",
    if(any(types == "cens_i")) c("t_cens_ir", "t_cens_il")
  )])))

  return(fit_data)
}
.fit_priors  <- function(priors){
  if(is.null(priors[["aux"]])){
    return(NULL)
  }else{
    return(priors["aux"])
  }
}

# formula tools
.generate_model_syntax             <- function(distribution, data){

  types <- attr(data[["survival"]], "type")

  model_syntax <- "model{\n"

  for(i in seq_along(types)){

    if(i == 1){
      from <- 1
      to   <- data[["survival"]][[paste0("n_", types[i])]]
    }else{
      from <- from + data[["survival"]][[paste0("n_", types[i-1])]]
      to   <- to   + data[["survival"]][[paste0("n_", types[i])]]
    }

    model_syntax <- paste0(model_syntax, .JAGS_survival_likelihood(distribution, types[i], from, to))
  }

  model_syntax <- paste0(model_syntax, "}")

  return(model_syntax)
}
.generate_model_formula_list       <- function(formula){

  # remove the left hand side
  formula[2] <- NULL
  formula    <- list("mu" = formula)

  return(formula)
}
.generate_model_formula_data_list  <- function(data){

  if(length(data[["predictors"]]) == 0){
    data <- list("mu" = data.frame(matrix(ncol = 0, nrow = sum(unlist(data[["survival"]][c("n_event", "n_cens_r", "n_cens_l", "n_cens_i")])))))
  }else{
    data <- list("mu" = data.frame(data[["predictors"]]))
  }

  return(data)
}
.generate_model_formula_prior_list <- function(priors){

  priors <- list("mu" = c(priors["intercept"], priors[["terms"]]))

  return(priors)
}

# marglik tools
.marglik_function      <- function(parameters, data, distribution){

  types <- attr(data[["survival"]], "type")

  log_lik <- 0

  for(i in seq_along(types)){

    if(i == 1){
      from <- 1
      to   <- data[["survival"]][[paste0("n_", types[i])]]
    }else{
      from <- from + data[["survival"]][[paste0("n_", types[i-1])]]
      to   <- to   + data[["survival"]][[paste0("n_", types[i])]]
    }

    log_lik <- log_lik + sum(.marglik_survival_likelihood(
      distribution = distribution,
      type         = types[i],
      x            = data[["survival"]][[paste0("t_", types[i])]],
      mu           = parameters[["mu"]][from:to],
      aux          = parameters[["aux"]]))
  }


  return(log_lik)
}
.marglik_function_null <- function(priors, data, distribution){

  types <- attr(data[["survival"]], "type")

  log_lik <- 0

  for(i in seq_along(types)){

    if(i == 1){
      from <- 1
      to   <- data[["survival"]][[paste0("n_", types[i])]]
    }else{
      from <- from + data[["survival"]][[paste0("n_", types[i-1])]]
      to   <- to   + data[["survival"]][[paste0("n_", types[i])]]
    }

    log_lik <- log_lik + .marglik_survival_likelihood(
      distribution = distribution,
      type         = types[i],
      x            = data[["time"]][from:to],
      mu           = priors[["intercept"]]$parameters[["location"]],
      aux          = priors[["aux"]]$parameters[["location"]])
  }


  return(log_lik)
}

# additional tools
.JAGS_survival_likelihood    <- function(distribution, type, from, to){
  paste0(
    "for(i in ", from,":", to, "){\n",
    "   time[i]", " ~ ", gsub("-", "_", distribution), "_", type, "(", "mu[i]", if(.has_aux(distribution)){", aux"}, ")\n",
    "}\n")
}
.marglik_survival_likelihood <- function(distribution, type, x, mu, aux){

  # TODO: add other types of censoring
  survival_likelihood <- .get_marginal_distribution_function(distribution, type)
  args <- list(
    t   = x,
    eta = mu
  )
  if(.has_aux(distribution)){
    args <- c(args, list(aux))
  }

  log_lik <- do.call(survival_likelihood, args)

  return(log_lik)
}
.fitting_priority            <- function(models){

  terms         <- sapply(models, function(m)sum(!sapply(m[["priors"]][["terms"]], BayesTools::is.prior.point)))
  distributions <- sapply(models, function(m)switch(
    as.character(m$distribution),
    "exp-aft"     = 0,
    "weibull-aft" = 1.5,
    "lnorm-aft"   = 1,
    "llogis-aft"  = 1,
    "gamma-aft"   = 1.5
  ))

  fitting_difficulty <- terms + distributions

  return(order(fitting_difficulty, decreasing = TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{RoBSA}
\alias{RoBSA}
\title{Fit Robust Bayesian Survival Analysis}
\usage{
RoBSA(
  formula,
  data,
  priors = NULL,
  test_predictors = NULL,
  distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
  distributions_weights = rep(1, length(distributions)),
  prior_beta_null = get_default_prior_beta_null(),
  prior_beta_alt = get_default_prior_beta_alt(),
  prior_factor_null = get_default_prior_factor_null(),
  prior_factor_alt = get_default_prior_factor_alt(),
  prior_intercept = get_default_prior_intercept(),
  prior_aux = get_default_prior_aux(),
  chains = 3,
  sample = 5000,
  burnin = 2000,
  adapt = 500,
  thin = 1,
  parallel = FALSE,
  autofit = TRUE,
  autofit_control = set_autofit_control(),
  convergence_checks = set_convergence_checks(),
  save = "all",
  seed = NULL,
  silent = TRUE,
  rescale_data = FALSE,
  ...
)
}
\arguments{
\item{formula}{formula for the survival model}

\item{data}{data frame containing the data}

\item{priors}{names list of prior distributions for each
predictor. It allows users to specify both the null and alternative
hypothesis prior distributions by assigning a named list
(with \code{"null"} and \code{"alt"} object) to the predictor}

\item{test_predictors}{vector of predictor names
to be tested with Bayesian model-averaged testing.
Defaults to \code{NULL}, no parameters are tested.}

\item{distributions}{distributions of parametric
survival models}

\item{distributions_weights}{prior odds for the competing
distributions}

\item{prior_beta_null}{default prior distribution for the
null hypotheses of continuous predictors}

\item{prior_beta_alt}{default prior distribution for the
alternative hypotheses of continuous predictors}

\item{prior_factor_null}{default prior distribution for the
null hypotheses of categorical predictors}

\item{prior_factor_alt}{default prior distribution for the
alternative hypotheses of categorical predictors}

\item{prior_intercept}{named list containing prior
distribution for the intercepts (with names corresponding
to the distributions)}

\item{prior_aux}{named list containing prior
distribution for the auxiliary parameters (with names corresponding
to the distributions)}

\item{chains}{a number of chains of the MCMC algorithm.}

\item{sample}{a number of sampling iterations of the MCMC algorithm.
Defaults to \code{5000}.}

\item{burnin}{a number of burnin iterations of the MCMC algorithm.
Defaults to \code{2000}.}

\item{adapt}{a number of adaptation iterations of the MCMC algorithm.
Defaults to \code{500}.}

\item{thin}{a thinning of the chains of the MCMC algorithm. Defaults to
\code{1}.}

\item{parallel}{whether the individual models should be fitted in parallel.
Defaults to \code{FALSE}. The implementation is not completely stable
and might cause a connection error.}

\item{autofit}{whether the model should be fitted until the convergence
criteria (specified in \code{autofit_control}) are satisfied. Defaults to
\code{TRUE}.}

\item{autofit_control}{allows to pass autofit control settings with the
\code{\link[=set_autofit_control]{set_autofit_control()}} function. See \code{?set_autofit_control} for
options and default settings.}

\item{convergence_checks}{automatic convergence checks to assess the fitted
models, passed with \code{\link[=set_convergence_checks]{set_convergence_checks()}} function. See
\code{?set_convergence_checks} for options and default settings.}

\item{save}{whether all models posterior distributions should be kept
after obtaining a model-averaged result. Defaults to \code{"all"} which
does not remove anything. Set to \code{"min"} to significantly reduce
the size of final object, however, some model diagnostics and further
manipulation with the object will not be possible.}

\item{seed}{a seed to be set before model fitting, marginal likelihood
computation, and posterior mixing for reproducibility of results. Defaults
to \code{NULL} - no seed is set.}

\item{silent}{whether all print messages regarding the fitting process
should be suppressed. Defaults to \code{TRUE}. Note that \code{parallel = TRUE}
also suppresses all messages.}

\item{rescale_data}{whether continuous predictors should be rescaled prior to
estimating the model. Defaults to \code{FALSE}.}

\item{...}{additional arguments.}
}
\value{
\code{RoBSA} returns an object of class 'RoBSA'.
}
\description{
\code{RoBSA} is used to estimate a robust Bayesian
survival analysis. The interface allows a complete customization of
the ensemble with different prior distributions for the null and
alternative hypothesis of each parameter.
(See README for an example.)
}
\examples{
\dontrun{
# (execution of the example takes several minutes)
# example from the README (more details and explanation therein)
data(cancer, package = "survival")
priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
df <- data.frame(
  time         = veteran$time / 12,
  status       = veteran$status,
  treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
  karno_scaled = veteran$karno / 100
)
RoBSA.options(check_scaling = FALSE)
fit <- RoBSA(
  Surv(time, status) ~ treatment + karno_scaled,
  data   = df,
  priors = list(
    treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
                                truncation = list(0, Inf), contrast = "treatment"),
    karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
  ),
  test_predictors = "treatment",
  prior_intercept = priors[["intercept"]],
  prior_aux       = priors[["aux"]],
  parallel = TRUE, seed = 1
)
summary(fit)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_prediction}
\alias{plot_prediction}
\alias{plot_survival}
\alias{plot_hazard}
\alias{plot_density}
\title{Survival plots for a RoBSA object}
\usage{
plot_prediction(
  x,
  type = "survival",
  time_range = NULL,
  new_data = NULL,
  predictor = NULL,
  covariates_data = NULL,
  conditional = FALSE,
  plot_type = "base",
  samples = 10000,
  ...
)

plot_survival(
  x,
  time_range = NULL,
  new_data = NULL,
  predictor = NULL,
  covariates_data = NULL,
  conditional = FALSE,
  plot_type = "base",
  samples = 10000,
  ...
)

plot_hazard(
  x,
  time_range = NULL,
  new_data = NULL,
  predictor = NULL,
  covariates_data = NULL,
  conditional = FALSE,
  plot_type = "base",
  samples = 10000,
  ...
)

plot_density(
  x,
  time_range = NULL,
  new_data = NULL,
  predictor = NULL,
  covariates_data = NULL,
  conditional = FALSE,
  plot_type = "base",
  samples = 10000,
  ...
)
}
\arguments{
\item{x}{a fitted RoBSA object.}

\item{type}{what type of prediction should be created}

\item{time_range}{a numeric of length two specifying the range for the
survival prediction. Defaults to \code{NULL} which uses the range of
observed times.}

\item{new_data}{a data.frame containing fully specified predictors for which
predictions should be made}

\item{predictor}{an alternative input to \code{new_data} that automatically
generates predictions for each level of the predictor across all either across
levels of covariates specified by \code{covariates_data} or at the default values
of other predictors}

\item{covariates_data}{a supplementary input to \code{predictor} that specifies
levels of covariates for which predictions should be made}

\item{conditional}{whether only models assuming presence of the specified
\code{predictor} should be used}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting. Defaults to
\code{"base"}.}

\item{samples}{number of posterior samples to be evaluated}

\item{...}{additional arguments.}
}
\value{
returns either \code{NULL} if \code{plot_type = "base"}
or an object object of class 'ggplot2' if \code{plot_type = "ggplot2"}.
}
\description{
Survival plots for a RoBSA object
}
\examples{
\dontrun{
# (execution of the example takes several minutes)
# example from the README (more details and explanation therein)
data(cancer, package = "survival")
priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
df <- data.frame(
  time         = veteran$time / 12,
  status       = veteran$status,
  treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
  karno_scaled = veteran$karno / 100
)
RoBSA.options(check_scaling = FALSE)
fit <- RoBSA(
  Surv(time, status) ~ treatment + karno_scaled,
  data   = df,
  priors = list(
    treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
                                truncation = list(0, Inf), contrast = "treatment"),
    karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
  ),
  test_predictors = "treatment",
  prior_intercept = priors[["intercept"]],
  prior_aux       = priors[["aux"]],
  parallel = TRUE, seed = 1
)

# plot survival for each level the treatment
plot_survival(fit, parameter = "treatment")

# plot hazard for each level the treatment
plot_hazard(fit, parameter = "treatment")

# plot density for each level the treatment
plot_density(fit, parameter = "treatment")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_informed}
\alias{prior_informed}
\title{Creates an informed prior distribution based on research}
\usage{
prior_informed(name, parameter = NULL, type = "smd")
}
\arguments{
\item{name}{name of the prior distribution. There are many options based on prior psychological
or medical research.
For psychology, the possible options are
\describe{
\item{\code{"van Erp"}}{for an informed prior distribution for the heterogeneity parameter tau
of meta-analytic effect size estimates based on standardized mean differences
(van Erp et al. 2017),}
\item{\code{"Oosterwijk"}}{for an informed prior distribution for the effect sizes expected in
social psychology based on prior elicitation with dr. Oosterwijk
(Gronau et al. 2017).}
}
For medicine, the possible options are based on Bartoš et al. (2021)
who developed empirical prior distributions for the effect size and heterogeneity parameters of the
continuous standardized outcomes based on the Cochrane database of systematic reviews.
Use \code{"Cochrane"} for a prior distribution based on the whole database or call
\code{print(prior_informed_medicine_names)} to inspect the names of
all 46 subfields and set the appropriate \code{parameter} and \code{type}.}

\item{parameter}{parameter name describing what prior distribution is supposed to be produced in cases
where the \code{name} corresponds to multiple prior distributions. Relevant only for the empirical medical
prior distributions.}

\item{type}{prior type describing what prior distribution is supposed to be produced in cases
where the \code{name} and \code{parameter} correspond to multiple prior distributions. Relevant only for
the empirical medical prior distributions.}
}
\value{
\code{prior_informed} returns an object of class 'prior'.
}
\description{
\code{prior_informed} creates an informed prior distribution based on past
research. The prior can be visualized by the \code{plot} function.
}
\details{
Further details can be found in \insertCite{erp2017estimates;textual}{BayesTools},
\insertCite{gronau2017bayesian;textual}{BayesTools}, and
\insertCite{bartos2021bayesian;textual}{BayesTools}.
}
\examples{
# prior distribution representing expected effect sizes in social psychology
# based on prior elicitation with dr. Oosterwijk
p1 <- prior_informed("Oosterwijk")

# the prior distribution can be visualized using the plot function
# (see ?plot.prior for all options)
plot(p1)


# empirical prior distribution for the standardized mean differences from the oral health
# medical subfield based on meta-analytic effect size estimates from the
# Cochrane database of systematic reviews
p2 <- prior_informed("Oral Health", parameter = "effect", type = "smd")
print(p2)
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[BayesTools:prior]{prior()}}, \link[BayesTools]{prior_informed_medicine_names}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.RoBSA}
\alias{predict.RoBSA}
\title{Predict method for RoBSA objects.}
\usage{
\method{predict}{RoBSA}(
  object,
  time = NULL,
  new_data = NULL,
  predictor = NULL,
  covariates_data = NULL,
  type = c("survival", "hazard", "density", "mean", "sd"),
  summarize = TRUE,
  averaged = TRUE,
  conditional = FALSE,
  samples = 10000,
  ...
)
}
\arguments{
\item{object}{a fitted RoBSA object}

\item{time}{a vector of time values at which the survival/hazard/density
will be predicted (for each passed data point)}

\item{new_data}{a data.frame containing fully specified predictors for which
predictions should be made}

\item{predictor}{an alternative input to \code{new_data} that automatically
generates predictions for each level of the predictor across all either across
levels of covariates specified by \code{covariates_data} or at the default values
of other predictors}

\item{covariates_data}{a supplementary input to \code{predictor} that specifies
levels of covariates for which predictions should be made}

\item{type}{what type of prediction should be created}

\item{summarize}{whether the predictions should be aggregated as mean and sd.
Otherwise, prediction for for posterior samples is returned.}

\item{averaged}{whether predictions should be combined with Bayesian model-averaging
or whether predictions for each individual model should be returned.}

\item{conditional}{whether only models assuming presence of the specified
\code{predictor} should be used}

\item{samples}{number of posterior samples to be evaluated}

\item{...}{additional arguments (unused)}
}
\value{
a list with predictions (or a list of lists in case that predictions for each
individual model are requested \code{averaged = FALSE})
}
\description{
Predicts survival/hazard/density/mean/sd for a given
RoBSA object. Either predicts values for each row of a fully specified
\code{new_data} data.frame, or for all levels of a given \code{predictor}
at the mean of continuous covariate values and default factor levels or
covariate values specified as \code{covariates_data} data.frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors_calibration.R
\name{calibrate_quartiles}
\alias{calibrate_quartiles}
\title{Calibrate prior distributions based on quartiles}
\usage{
calibrate_quartiles(
  median_t,
  iq_range_t,
  prior_sd = 0.5,
  distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
  verbose = FALSE,
  search_bounds1 = c(-100, 100),
  search_bounds2 = c(0 + 0.01, 100)
)
}
\arguments{
\item{median_t}{median survival}

\item{iq_range_t}{interquartile range of the survival}

\item{prior_sd}{pre-specified standard deviation of the
prior distributions (either a single value that is used
for both the intercept and auxiliary parameter or a vector
where the first value corresponds to the sd for the prior
distribution on the intercept and the second value to the
sd for the prior distribution on the auxiliary parameter)}

\item{distributions}{vector of parametric families for which
prior distributions ought to be calibrated}

\item{verbose}{whether debug information be printed}

\item{search_bounds1}{search boundaries for the intercept
parameter}

\item{search_bounds2}{search boundaries for the auxiliary
parameter}
}
\value{
returns a list of prior distribution for the intercepts
and auxiliary parameters.
}
\description{
Calibrates prior distributions for parametric
survival analysis based on median survival and interquartile
range. Returns a list of prior distribution for the intercepts
and auxiliary parameters.
}
\examples{
priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{print.RoBSA}
\alias{print.RoBSA}
\title{Prints a fitted RoBSA object}
\usage{
\method{print}{RoBSA}(x, ...)
}
\arguments{
\item{x}{a fitted RoBSA object.}

\item{...}{additional arguments.}
}
\value{
\code{print.RoBSA} invisibly returns the print statement.
}
\description{
Prints a fitted RoBSA object
}
\seealso{
\code{\link[=RoBSA]{RoBSA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_none}
\alias{prior_none}
\title{Creates a prior distribution}
\usage{
prior_none(prior_weights = 1)
}
\arguments{
\item{prior_weights}{prior odds associated with a given distribution.
The value is passed into the model fitting function, which creates models
corresponding to all combinations of prior distributions for each of
the model parameters and sets the model priors odds to the product
of its prior distributions.}
}
\value{
\code{prior} and \code{prior_none} return an object of class 'prior'.
A named list containing the distribution name, parameters, and prior weights.
}
\description{
\code{prior} creates a prior distribution.
The prior can be visualized by the \code{plot} function.
}
\examples{
# create a standard normal prior distribution
p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))

# create a half-normal standard normal prior distribution
p2 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1),
truncation = list(lower = 0, upper = Inf))

# the prior distribution can be visualized using the plot function
# (see ?plot.prior for all options)
plot(p1)
}
\seealso{
\code{\link[BayesTools:plot.prior]{plot.prior()}}, \link[stats]{Normal}, \link[stats]{Lognormal}, \link[stats]{Cauchy},
\link[stats]{Beta}, \link[stats]{Exponential},
\link[extraDistr]{LocationScaleT}, \link[extraDistr]{InvGamma}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flexsurv-tools.R
\name{extract_flexsurv}
\alias{extract_flexsurv}
\title{Extract parameter estimates from \code{flexsurv} object}
\usage{
extract_flexsurv(fit)
}
\arguments{
\item{fit}{an object fitted with the \code{flexsurv::flexsurvreg}
function}
}
\value{
\code{extract_flexsurv} return list of estimates lists for each
parameter.
}
\description{
\code{extract_flexsurv} extracts estimates from a
\code{flexsurv} object in and transform them to match the
\code{RoBSA} output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors_calibration.R
\name{calibrate_meta_analytic}
\alias{calibrate_meta_analytic}
\title{Create meta-analytic predictive prior distributions}
\usage{
calibrate_meta_analytic(
  datasets,
  distributions = c("exp-aft", "weibull-aft", "lnorm-aft", "llogis-aft", "gamma-aft"),
  prior_mu = prior("cauchy", parameters = list(location = 0, scale = 100)),
  prior_tau = prior("cauchy", parameters = list(location = 0, scale = 10), truncation =
    list(0, Inf)),
  ...
)
}
\arguments{
\item{datasets}{list of data.frames containing the historical
data. Each data.frame must contain a column named \code{"time"}
with the survival times and a column named \code{"status"}
with the censoring status.}

\item{distributions}{vector of parametric families for which
prior distributions ought to be calibrated}

\item{prior_mu}{prior distribution for the the meta-analytic mean parameter}

\item{prior_tau}{prior distribution for the the meta-analytic heterogeneity parameter}

\item{...}{additional parameters to be passed to the meta-analytic function.
See \link[BayesTools:JAGS_fit]{BayesTools::JAGS_fit} for more details.}
}
\value{
returns a list of prior distribution for the intercepts
and auxiliary parameters.
}
\description{
Calibrates prior distributions for parametric
survival analysis based on historical data.
Returns a list of prior distribution for the intercepts
and auxiliary parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-input-and-settings.R
\name{RoBSA_control}
\alias{RoBSA_control}
\alias{set_autofit_control,}
\alias{set_convergence_checks}
\alias{set_autofit_control}
\title{Control MCMC fitting process}
\usage{
set_autofit_control(
  max_Rhat = 1.05,
  min_ESS = 500,
  max_error = NULL,
  max_SD_error = NULL,
  max_time = list(time = 60, unit = "mins"),
  sample_extend = 1000,
  restarts = 10,
  max_extend = 10,
  check_indicators = FALSE
)

set_convergence_checks(
  max_Rhat = 1.05,
  min_ESS = 500,
  max_error = NULL,
  max_SD_error = NULL,
  check_indicators = FALSE,
  remove_failed = FALSE,
  balance_probability = TRUE
)
}
\arguments{
\item{max_Rhat}{maximum value of the R-hat diagnostic.
Defaults to \code{1.05}.}

\item{min_ESS}{minimum estimated sample size.
Defaults to \code{500}.}

\item{max_error}{maximum value of the MCMC error.
Defaults to \code{NULL}. Be aware that PEESE publication bias
adjustment can have estimates on different scale than the rest of
the output, resulting in relatively large max MCMC error.}

\item{max_SD_error}{maximum value of the proportion of MCMC error
of the estimated SD of the parameter.
Defaults to \code{NULL}.}

\item{max_time}{list with the time and unit specifying the maximum
autofitting process per model. Passed to \link[base]{difftime} function
(possible units are \code{"secs", "mins", "hours", "days", "weeks", "years"}).
Defaults to \code{list(time = 60, unit = "mins")}.}

\item{sample_extend}{number of samples to extend the fitting process if
the criteria are not satisfied.
Defaults to \code{1000}.}

\item{restarts}{number of fitting restarts attempted when the initial JAGS
model fitting fails. Defaults to \code{10}.}

\item{max_extend}{maximum number of autofit extensions per model.
Defaults to \code{10}.}

\item{check_indicators}{whether convergence checks should include latent
prior inclusion indicators. Defaults to \code{FALSE}.}

\item{remove_failed}{whether models not satisfying the convergence checks should
be removed from the inference. Defaults to \code{FALSE} - only a warning is raised.}

\item{balance_probability}{whether prior model probability should be balanced
across the combinations of models with the same H0/H1 for effect / heterogeneity / bias
in the case of non-convergence. Defaults to \code{TRUE}.}
}
\value{
\code{set_autofit_control} returns a list of autofit control settings
and \code{set_convergence_checks} returns a list of convergence checks settings.
}
\description{
Controls settings for the autofit
process of the MCMC JAGS sampler (specifies termination
criteria), and values for the convergence checks.
}
\seealso{
\link{RoBSA}, \link{update.RoBSA}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{check_RoBSA}
\alias{check_RoBSA}
\title{Check fitted RoBSA object for errors and warnings}
\usage{
check_RoBSA(fit)
}
\arguments{
\item{fit}{a fitted RoBSA object.}
}
\value{
\code{check_RoBSA} returns a vector of error and
warning messages.
}
\description{
Checks fitted RoBSA object
for warnings and errors and prints them to the
console.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{is.RoBSA}
\alias{is.RoBSA}
\title{Reports whether x is a RoBSA object}
\usage{
is.RoBSA(x)
}
\arguments{
\item{x}{an object to test}
}
\value{
is.RoBSA returns a boolean.
}
\description{
Reports whether x is a RoBSA object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnostics}
\alias{diagnostics}
\alias{diagnostics_autocorrelation}
\alias{diagnostics_trace}
\alias{diagnostics_density}
\title{Visualizes MCMC diagnostics for a fitted RoBSA object}
\usage{
diagnostics(
  fit,
  parameter = NULL,
  type,
  plot_type = "base",
  show_models = NULL,
  lags = 30,
  title = is.null(show_models) | length(show_models) > 1,
  ...
)

diagnostics_autocorrelation(
  fit,
  parameter = NULL,
  plot_type = "base",
  show_models = NULL,
  lags = 30,
  title = is.null(show_models) | length(show_models) > 1,
  ...
)

diagnostics_trace(
  fit,
  parameter = NULL,
  plot_type = "base",
  show_models = NULL,
  title = is.null(show_models) | length(show_models) > 1,
  ...
)

diagnostics_density(
  fit,
  parameter = NULL,
  plot_type = "base",
  show_models = NULL,
  title = is.null(show_models) | length(show_models) > 1,
  ...
)
}
\arguments{
\item{fit}{a fitted RoBSA object}

\item{parameter}{a parameter to be plotted.}

\item{type}{type of MCMC diagnostic to be plotted.
Options are \code{"trace"} for the chains' trace plots,
\code{"autocorrelation"} for autocorrelation of the
chains, and \code{"densities"} for the overlaying
densities of the individual chains. Can be abbreviated to
first letters.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting. Defaults to
\code{"base"}.}

\item{show_models}{MCMC diagnostics of which models should be
plotted. Defaults to \code{NULL} which plots MCMC diagnostics
for a specified parameter for every model that is part of the
ensemble.}

\item{lags}{number of lags to be shown for
\code{type = "autocorrelation"}. Defaults to \code{30}.}

\item{title}{whether the model number should be displayed in title.
Defaults to \code{TRUE} when more than one model is selected.}

\item{...}{additional arguments to be passed to the plotting functions.}
}
\value{
\code{diagnostics} returns either \code{NULL} if \code{plot_type = "base"}
or an object/list of objects (depending on the number of parameters to be plotted)
of class 'ggplot2' if \code{plot_type = "ggplot2"}.
}
\description{
\code{diagnostics} creates visual
checks of individual models convergence. Numerical
overview of individual models can be obtained by
\code{summary(object, type = "diagnostics")},
or even more detailed information by
\code{summary(object, type = "individual")}.
}
\seealso{
\code{\link[=RoBSA]{RoBSA()}}, \code{\link[=summary.RoBSA]{summary.RoBSA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{print.summary.RoBSA}
\alias{print.summary.RoBSA}
\title{Prints summary object for RoBSA method}
\usage{
\method{print}{summary.RoBSA}(x, ...)
}
\arguments{
\item{x}{a summary of a RoBSA object}

\item{...}{additional arguments}
}
\value{
\code{print.summary.RoBSA} invisibly returns the print statement.
}
\description{
Prints summary object for RoBSA method
}
\seealso{
\code{\link[=RoBSA]{RoBSA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RoBSA-package.R
\docType{package}
\name{RoBSA-package}
\alias{RoBSA-package}
\alias{_PACKAGE}
\alias{RoBSA_package}
\alias{RoBSA.package}
\title{RoBSA: Robust Bayesian survival analysis}
\description{
Bayesian model-averaged parametric survival analysis
with ability to specify informed prior distributions and draw inference with
inclusion Bayes factors. See \insertCite{bartos2021informed;textual}{RoBSA}
for more details about the methodology.
}
\section{User guide}{
 See \insertCite{bartos2021informed;textual}{RoBSA},
for details regarding the RoBSA methodology.
}

\references{
\insertAllCited{}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fbartos.github.io/RoBSA/}
  \item Report bugs at \url{https://github.com/FBartos/RoBSA/issues}
}

}
\author{
František Bartoš \email{f.bartos96@gmail.com}
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_factor}
\alias{prior_factor}
\title{Creates a prior distribution for factors}
\usage{
prior_factor(
  distribution,
  parameters,
  truncation = list(lower = -Inf, upper = Inf),
  prior_weights = 1,
  contrast = "orthonormal"
)
}
\arguments{
\item{distribution}{name of the prior distribution. The
possible options are
\describe{
\item{\code{"point"}}{for a point density characterized by a
\code{location} parameter.}
\item{\code{"normal"}}{for a normal distribution characterized
by a \code{mean} and \code{sd} parameters.}
\item{\code{"lognormal"}}{for a lognormal distribution characterized
by a \code{meanlog} and \code{sdlog} parameters.}
\item{\code{"cauchy"}}{for a Cauchy distribution characterized
by a \code{location} and \code{scale} parameters. Internally
converted into a generalized t-distribution with \code{df = 1}.}
\item{\code{"t"}}{for a generalized t-distribution characterized
by a \code{location}, \code{scale}, and \code{df} parameters.}
\item{\code{"gamma"}}{for a gamma distribution characterized
by either \code{shape} and \code{rate}, or \code{shape} and
\code{scale} parameters. The later is internally converted to
the \code{shape} and \code{rate} parametrization}
\item{\code{"invgamma"}}{for an inverse-gamma distribution
characterized by a \code{shape} and \code{scale} parameters. The
JAGS part uses a 1/gamma distribution with a shape and rate
parameter.}
\item{\code{"beta"}}{for a beta distribution
characterized by an \code{alpha} and \code{beta} parameters.}
\item{\code{"exp"}}{for an exponential distribution
characterized by either \code{rate} or \code{scale}
parameter. The later is internally converted to
\code{rate}.}
\item{\code{"uniform"}}{for a uniform distribution defined on a
range from \code{a} to \code{b}}
}}

\item{parameters}{list of appropriate parameters for a given
\code{distribution}.}

\item{truncation}{list with two elements, \code{lower} and
\code{upper}, that define the lower and upper truncation of the
distribution. Defaults to \code{list(lower = -Inf, upper = Inf)}.
The truncation is automatically set to the bounds of the support.}

\item{prior_weights}{prior odds associated with a given distribution.
The value is passed into the model fitting function, which creates models
corresponding to all combinations of prior distributions for each of
the model parameters and sets the model priors odds to the product
of its prior distributions.}

\item{contrast}{type of contrast for the prior distribution. The possible options are
\describe{
\item{\code{"meandif"}}{for contrast centered around the grand mean
with equal marginal distributions, making the prior distribution exchangeable
across factor levels. In contrast to \code{"orthonormal"}, the marginal distributions
are identical regardless of the number of factor levels and the specified prior
distribution corresponds to the difference from grand mean for each factor level.
Only supports \code{distribution = "mnormal"} and \code{distribution = "mt"}
which generates the corresponding multivariate normal/t distributions.}
\item{\code{"orthonormal"}}{for contrast centered around the grand mean
with equal marginal distributions, making the prior distribution exchangeable
across factor levels. Only supports \code{distribution = "mnormal"} and
\code{distribution = "mt"} which generates the corresponding multivariate normal/t
distributions.}
\item{\code{"treatment"}}{for contrasts using the first level as a comparison
group and setting equal prior distribution on differences between the individual
factor levels and the comparison level.}
\item{\code{"independent"}}{for contrasts specifying dependent prior distribution
for each factor level (note that this leads to an overparameterized model if the
intercept is included).}
}}
}
\value{
return an object of class 'prior'.
}
\description{
\code{prior_factor} creates a prior distribution for fitting
models with factor predictors. (Note that results across different operating
systems might vary due to differences in JAGS numerical precision.)
}
\examples{
# create an orthonormal prior distribution
p1 <- prior_factor(distribution = "mnormal", contrast = "orthonormal",
                   parameters = list(mean = 0, sd = 1))
}
\seealso{
\code{\link[BayesTools:prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-input-and-settings.R
\name{default_prior}
\alias{default_prior}
\alias{get_default_prior_beta_null,}
\alias{get_default_prior_beta_alt,}
\alias{get_default_prior_factor_null,}
\alias{get_default_prior_factor_alt,}
\alias{get_default_prior_intercept,}
\alias{get_default_prior_aux}
\alias{get_default_prior_beta_null}
\alias{get_default_prior_beta_alt}
\alias{get_default_prior_factor_null}
\alias{get_default_prior_factor_alt}
\alias{get_default_prior_intercept}
\title{Default prior distributions}
\usage{
get_default_prior_beta_null()

get_default_prior_beta_alt()

get_default_prior_factor_null()

get_default_prior_factor_alt()

get_default_prior_intercept()

get_default_prior_aux()
}
\value{
\code{get_default_prior_beta_null} and \code{get_default_prior_beta_alt}
return a prior distribution and \code{get_default_prior_intercept} and
\code{get_default_prior_aux} return a list of prior distributions.
}
\description{
Functions for setting default prior distributions. Note that
these default prior distributions might (and probably won't) apply to your
specific data scenario.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.RoBSA}
\alias{plot.RoBSA}
\title{Plots a fitted RoBSA object}
\usage{
\method{plot}{RoBSA}(
  x,
  parameter = NULL,
  conditional = FALSE,
  plot_type = "base",
  prior = FALSE,
  dots_prior = NULL,
  ...
)
}
\arguments{
\item{x}{a fitted RoBSA object}

\item{parameter}{a name of parameter to be plotted. Defaults to
the first regression parameter if left unspecified. Use
\code{"intercept"} and \code{"aux"} to plot the intercepts and
auxiliary parameters of each distribution family.}

\item{conditional}{whether conditional estimates should be
plotted. Defaults to \code{FALSE} which plots the model-averaged
estimates.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting. Defaults to
\code{"base"}.}

\item{prior}{whether prior distribution should be added to
figure. Defaults to \code{FALSE}.}

\item{dots_prior}{list of additional graphical arguments
to be passed to the plotting function of the prior
distribution. Supported arguments are \code{lwd},
\code{lty}, \code{col}, and \code{col.fill}, to adjust
the line thickness, line type, line color, and fill color
of the prior distribution respectively.}

\item{...}{list of additional graphical arguments
to be passed to the plotting function. Supported arguments
are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
\code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim}
to adjust the line thickness, line type, line color, fill color,
x-label, y-label, title, x-axis range, and y-axis range
respectively.}
}
\value{
\code{plot.RoBSA} returns either \code{NULL} if \code{plot_type = "base"}
or an object object of class 'ggplot2' if \code{plot_type = "ggplot2"}.
}
\description{
\code{plot.RoBSA} allows to visualize
posterior distribution of different \code{"RoBSA"} object
parameters. See \code{plot_survival} for plotting the survival
ways. See \code{type} for the different model types.
}
\examples{
\dontrun{
# (execution of the example takes several minutes)
# example from the README (more details and explanation therein)
data(cancer, package = "survival")
priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
df <- data.frame(
  time         = veteran$time / 12,
  status       = veteran$status,
  treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
  karno_scaled = veteran$karno / 100
)
RoBSA.options(check_scaling = FALSE)
fit <- RoBSA(
  Surv(time, status) ~ treatment + karno_scaled,
  data   = df,
  priors = list(
    treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
                                truncation = list(0, Inf), contrast = "treatment"),
    karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
  ),
  test_predictors = "treatment",
  prior_intercept = priors[["intercept"]],
  prior_aux       = priors[["aux"]],
  parallel = TRUE, seed = 1
)

# plot posterior distribution of the treatment effect
plot(fit, parameter = "treatment")

}


}
\seealso{
\code{\link[=RoBSA]{RoBSA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{RoBSA_options}
\alias{RoBSA_options}
\alias{RoBSA.options}
\alias{RoBSA.get_option}
\title{Options for the RoBSA package}
\usage{
RoBSA.options(...)

RoBSA.get_option(name)
}
\arguments{
\item{...}{named option(s) to change - for a list of available options, see
details below.}

\item{name}{the name of the option to get the current value of - for a list of
available options, see details below.}
}
\value{
The current value of all available RoBSA options (after applying any
changes specified) is returned invisibly as a named list.
}
\description{
A placeholder object and functions for the RoBSA package.
(adapted from the runjags R package).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior}
\alias{prior}
\title{Creates a prior distribution}
\usage{
prior(
  distribution,
  parameters,
  truncation = list(lower = -Inf, upper = Inf),
  prior_weights = 1
)
}
\arguments{
\item{distribution}{name of the prior distribution. The
possible options are
\describe{
\item{\code{"point"}}{for a point density characterized by a
\code{location} parameter.}
\item{\code{"normal"}}{for a normal distribution characterized
by a \code{mean} and \code{sd} parameters.}
\item{\code{"lognormal"}}{for a lognormal distribution characterized
by a \code{meanlog} and \code{sdlog} parameters.}
\item{\code{"cauchy"}}{for a Cauchy distribution characterized
by a \code{location} and \code{scale} parameters. Internally
converted into a generalized t-distribution with \code{df = 1}.}
\item{\code{"t"}}{for a generalized t-distribution characterized
by a \code{location}, \code{scale}, and \code{df} parameters.}
\item{\code{"gamma"}}{for a gamma distribution characterized
by either \code{shape} and \code{rate}, or \code{shape} and
\code{scale} parameters. The later is internally converted to
the \code{shape} and \code{rate} parametrization}
\item{\code{"invgamma"}}{for an inverse-gamma distribution
characterized by a \code{shape} and \code{scale} parameters. The
JAGS part uses a 1/gamma distribution with a shape and rate
parameter.}
\item{\code{"beta"}}{for a beta distribution
characterized by an \code{alpha} and \code{beta} parameters.}
\item{\code{"exp"}}{for an exponential distribution
characterized by either \code{rate} or \code{scale}
parameter. The later is internally converted to
\code{rate}.}
\item{\code{"uniform"}}{for a uniform distribution defined on a
range from \code{a} to \code{b}}
}}

\item{parameters}{list of appropriate parameters for a given
\code{distribution}.}

\item{truncation}{list with two elements, \code{lower} and
\code{upper}, that define the lower and upper truncation of the
distribution. Defaults to \code{list(lower = -Inf, upper = Inf)}.
The truncation is automatically set to the bounds of the support.}

\item{prior_weights}{prior odds associated with a given distribution.
The value is passed into the model fitting function, which creates models
corresponding to all combinations of prior distributions for each of
the model parameters and sets the model priors odds to the product
of its prior distributions.}
}
\value{
\code{prior} and \code{prior_none} return an object of class 'prior'.
A named list containing the distribution name, parameters, and prior weights.
}
\description{
\code{prior} creates a prior distribution.
The prior can be visualized by the \code{plot} function.
}
\examples{
# create a standard normal prior distribution
p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))

# create a half-normal standard normal prior distribution
p2 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1),
truncation = list(lower = 0, upper = Inf))

# the prior distribution can be visualized using the plot function
# (see ?plot.prior for all options)
plot(p1)
}
\seealso{
\code{\link[BayesTools:plot.prior]{plot.prior()}}, \link[stats]{Normal}, \link[stats]{Lognormal}, \link[stats]{Cauchy},
\link[stats]{Beta}, \link[stats]{Exponential},
\link[extraDistr]{LocationScaleT}, \link[extraDistr]{InvGamma}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.RoBSA}
\alias{summary.RoBSA}
\title{Summarize fitted RoBSA object}
\usage{
\method{summary}{RoBSA}(
  object,
  type = "ensemble",
  conditional = FALSE,
  exp = FALSE,
  parameters = FALSE,
  probs = c(0.025, 0.975),
  logBF = FALSE,
  BF01 = FALSE,
  transform_factors = TRUE,
  short_name = FALSE,
  remove_spike_0 = FALSE,
  ...
)
}
\arguments{
\item{object}{a fitted RoBSA object.}

\item{type}{whether to show the overall RoBSA results (\code{"ensemble"}),
an overview of the individual models (\code{"models"}), or detailed summary
for the individual models (\code{"individual"}).}

\item{conditional}{show the conditional estimates (assuming that the
alternative is true). Defaults to \code{FALSE}. Only available for
\code{type == "conditional"}.}

\item{exp}{whether exponents of the regression estimates should be also presented}

\item{parameters}{character vector of parameters (or a
named list with of character vectors for summary and
diagnostics tables) specifying the parameters
(and their grouping) for the summary table}

\item{probs}{quantiles of the posterior samples to be displayed.
Defaults to \code{c(.025, .50, .975)}}

\item{logBF}{show log of the BFs. Defaults to \code{FALSE}.}

\item{BF01}{show BF in support of the null hypotheses. Defaults to
\code{FALSE}.}

\item{transform_factors}{Whether factors with orthonormal prior
distributions should be transformed to differences from the grand mean. Defaults
to \code{TRUE}.}

\item{short_name}{whether the prior distribution names should be
shortened. Defaults to \code{FALSE}.}

\item{remove_spike_0}{whether prior distributions equal to spike
at 0 should be removed from the \code{prior_list}}

\item{...}{additional arguments}
}
\value{
summary of a RoBSA object

\code{summary.RoBSA} returns a list of tables of class 'BayesTools_table'.
}
\description{
\code{summary.RoBSA} creates a numerical
summary of the RoBSA object.
}
\note{
See \code{\link[=diagnostics]{diagnostics()}} for visual convergence checks of the individual models.
}
\examples{
\dontrun{
# (execution of the example takes several minutes)
# example from the README (more details and explanation therein)
data(cancer, package = "survival")
priors <- calibrate_quartiles(median_t = 5, iq_range_t = 10, prior_sd = 0.5)
df <- data.frame(
  time         = veteran$time / 12,
  status       = veteran$status,
  treatment    = factor(ifelse(veteran$trt == 1, "standard", "new"), levels = c("standard", "new")),
  karno_scaled = veteran$karno / 100
)
RoBSA.options(check_scaling = FALSE)
fit <- RoBSA(
  Surv(time, status) ~ treatment + karno_scaled,
  data   = df,
  priors = list(
    treatment    = prior_factor("normal", parameters = list(mean = 0.30, sd = 0.15),
                                truncation = list(0, Inf), contrast = "treatment"),
    karno_scaled = prior("normal", parameters = list(mean = 0, sd = 1))
  ),
  test_predictors = "treatment",
  prior_intercept = priors[["intercept"]],
  prior_aux       = priors[["aux"]],
  parallel = TRUE, seed = 1
)

# summary can provide many details about the model
summary(fit)

# note that the summary function contains additional arguments
# that allow to obtain a specific output, i.e, the conditional estimates
# (assuming that the non-null models are true) can be obtained
summary(fit, conditional = TRUE)

# overview of the models and their prior and posterior probability, marginal likelihood,
# and inclusion Bayes factor:
summary(fit, type = "models")

# and the model diagnostics overview, containing maximum R-hat and minimum ESS across parameters
# but see '?diagnostics' for diagnostics plots for individual model parameters
summary(fit, type = "diagnostics")

# summary of individual models and their parameters can be further obtained by
summary(fit, type = "individual")

}

}
\seealso{
\code{\link[=RoBSA]{RoBSA()}}, \code{\link[=diagnostics]{diagnostics()}}, \code{\link[=check_RoBSA]{check_RoBSA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\docType{data}
\name{prior_informed_medicine_names}
\alias{prior_informed_medicine_names}
\title{Names of medical subfields from the Cochrane database of systematic reviews}
\format{
An object of class \code{character} of length 47.
}
\usage{
prior_informed_medicine_names
}
\description{
Contain names identifying the individual subfields
from the Cochrane database of systematic reviews. The individual
elements correspond to valid name arguments for the \code{\link[=prior_informed]{prior_informed()}}
function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{update.RoBSA}
\alias{update.RoBSA}
\title{Updates a fitted RoBSA object}
\usage{
\method{update}{RoBSA}(
  object,
  refit_failed = TRUE,
  formula = NULL,
  priors = NULL,
  test_predictors = "",
  distribution = NULL,
  model_weights = 1,
  prior_beta_null = get_default_prior_beta_null(),
  prior_beta_alt = get_default_prior_beta_alt(),
  prior_factor_null = get_default_prior_factor_null(),
  prior_factor_alt = get_default_prior_factor_alt(),
  prior_intercept = get_default_prior_intercept(),
  prior_aux = get_default_prior_aux(),
  chains = NULL,
  adapt = NULL,
  burnin = NULL,
  sample = NULL,
  thin = NULL,
  autofit = NULL,
  parallel = NULL,
  autofit_control = NULL,
  convergence_checks = NULL,
  save = "all",
  seed = NULL,
  silent = TRUE,
  ...
)
}
\arguments{
\item{object}{a fitted RoBSA object}

\item{refit_failed}{whether failed models should be refitted. Relevant only
if new priors or \code{prior_weights} are not supplied. Defaults to \code{TRUE}.}

\item{formula}{formula for the survival model}

\item{priors}{names list of prior distributions for each
predictor. It allows users to specify both the null and alternative
hypothesis prior distributions by assigning a named list
(with \code{"null"} and \code{"alt"} object) to the predictor}

\item{test_predictors}{vector of predictor names
to be tested with Bayesian model-averaged testing.
Defaults to \code{NULL}, no parameters are tested.}

\item{distribution}{a distribution of the new model.}

\item{model_weights}{either a single value specifying prior model weight
of a newly specified model using priors argument, or a vector of the
same length as already fitted models to update their prior weights.}

\item{prior_beta_null}{default prior distribution for the
null hypotheses of continuous predictors}

\item{prior_beta_alt}{default prior distribution for the
alternative hypotheses of continuous predictors}

\item{prior_factor_null}{default prior distribution for the
null hypotheses of categorical predictors}

\item{prior_factor_alt}{default prior distribution for the
alternative hypotheses of categorical predictors}

\item{prior_intercept}{named list containing prior
distribution for the intercepts (with names corresponding
to the distributions)}

\item{prior_aux}{named list containing prior
distribution for the auxiliary parameters (with names corresponding
to the distributions)}

\item{chains}{a number of chains of the MCMC algorithm.}

\item{adapt}{a number of adaptation iterations of the MCMC algorithm.
Defaults to \code{500}.}

\item{burnin}{a number of burnin iterations of the MCMC algorithm.
Defaults to \code{2000}.}

\item{sample}{a number of sampling iterations of the MCMC algorithm.
Defaults to \code{5000}.}

\item{thin}{a thinning of the chains of the MCMC algorithm. Defaults to
\code{1}.}

\item{autofit}{whether the model should be fitted until the convergence
criteria (specified in \code{autofit_control}) are satisfied. Defaults to
\code{TRUE}.}

\item{parallel}{whether the individual models should be fitted in parallel.
Defaults to \code{FALSE}. The implementation is not completely stable
and might cause a connection error.}

\item{autofit_control}{allows to pass autofit control settings with the
\code{\link[=set_autofit_control]{set_autofit_control()}} function. See \code{?set_autofit_control} for
options and default settings.}

\item{convergence_checks}{automatic convergence checks to assess the fitted
models, passed with \code{\link[=set_convergence_checks]{set_convergence_checks()}} function. See
\code{?set_convergence_checks} for options and default settings.}

\item{save}{whether all models posterior distributions should be kept
after obtaining a model-averaged result. Defaults to \code{"all"} which
does not remove anything. Set to \code{"min"} to significantly reduce
the size of final object, however, some model diagnostics and further
manipulation with the object will not be possible.}

\item{seed}{a seed to be set before model fitting, marginal likelihood
computation, and posterior mixing for reproducibility of results. Defaults
to \code{NULL} - no seed is set.}

\item{silent}{whether all print messages regarding the fitting process
should be suppressed. Defaults to \code{TRUE}. Note that \code{parallel = TRUE}
also suppresses all messages.}

\item{...}{additional arguments.}
}
\value{
\code{update.RoBSA} returns an object of class 'RoBSA'.
}
\description{
\code{update.RoBSA} can be used to
\enumerate{
\item{add an additional model to an existing \code{"RoBSA"} object by
specifying the distribution, and either null or alternative priors
for each parameter and prior weight of the model,}
\item{change the prior weights of fitted models by specifying a vector
\code{prior_weights} of the same length as the fitted models,}
\item{refitting models that failed to converge with updated settings
of control parameters,}
\item{or changing the convergence criteria and recalculating the ensemble
results by specifying new \code{control} argument and setting
\code{refit_failed == FALSE}.}
}
}
\details{
See \code{\link[=RoBSA]{RoBSA()}} for more details.
}
\seealso{
\code{\link[=RoBSA]{RoBSA()}}, \code{\link[=summary.RoBSA]{summary.RoBSA()}}, \code{\link[=prior]{prior()}}, \code{\link[=check_setup]{check_setup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{weibull-aft}
\alias{weibull-aft}
\alias{weibull_aft_log_density}
\alias{weibull_aft_log_hazard}
\alias{weibull_aft_log_survival}
\alias{weibull_aft_density}
\alias{weibull_aft_hazard}
\alias{weibull_aft_survival}
\alias{weibull_aft_mean}
\alias{weibull_aft_sd}
\alias{weibull_aft_r}
\alias{weibull_aft_q}
\alias{weibull_aft_p}
\title{Weibull AFT parametric family.}
\usage{
weibull_aft_log_density(t, eta, shape)

weibull_aft_log_hazard(t, eta, shape)

weibull_aft_log_survival(t, eta, shape)

weibull_aft_density(t, eta, shape)

weibull_aft_hazard(t, eta, shape)

weibull_aft_survival(t, eta, shape)

weibull_aft_mean(eta, shape)

weibull_aft_sd(eta, shape)

weibull_aft_r(n, eta, shape)

weibull_aft_q(p, eta, shape)

weibull_aft_p(q, eta, shape)
}
\arguments{
\item{t}{vector of survival times}

\item{eta}{linear predictor}

\item{shape}{auxiliary parameter}

\item{n}{number of observations}

\item{p}{vector of probabilities}

\item{q}{vector of quantiles}
}
\value{
\code{weibull_aft_density}, \code{weibull_aft_hazard}, and
\code{weibull_aft_survival} return the density, hazard, and survival
of the specified survival distribution. The \code{weibull_aft_log_density},
\code{weibull_aft_log_hazard}, \code{weibull_aft_log_survival} return log of
the corresponding qualities. \code{weibull_aft_mean} and \code{weibull_aft_sd}
return the mean and standard deviation of the specified survival distribution.
\code{weibull_aft_r}, \code{weibull_aft_q}, and \code{weibull_aft_p} return a random
generation, quantiles, and cumulative probabilities of the specified
survival distribution.
}
\description{
(log) density, hazard, and survival
functions for AFT Weibull parametric family.
}
